/*____________________________________________________________________________
		Copyright (C) 2002 PGP Corporation
        All rights reserved.

        $Id: CPGPdiskHandlerImp.cpp,v 1.12 2002/08/06 20:09:34 dallen Exp $
____________________________________________________________________________*/

#include "pgpClassesConfig.h"
#include <string.h>

#include "CString.h"
#include "UString.h"
#include "UTime.h"

#include "pgpClientErrors.h"
#include "CDriverSubsystems.h"
#include "CPGPdisk.h"
#include "CPGPdiskHandlerImp.h"

_USING_PGP

_UNNAMED_BEGIN

// Constants
const char	*kPGPdiskDeviceName	= "PGPdiskVolume";

_UNNAMED_END


// Types

enum CPGPdiskHandlerImp::HandlerTask
{
	kMountTask, 
	kUnmountAllTask, 
	kUnmountTask, 
	kEverySecondTask
};

struct CPGPdiskHandlerImp::TaskInfoHeader
{
	CPGPdiskHandlerImp	*origThis;

	PGPBoolean	trueIfSync;
	CComboError	error;
	HandlerTask	task;
};

struct CPGPdiskHandlerImp::MountTaskInfo
{
	TaskInfoHeader	header;

	const char *path;
	const char *root;
	PGPdiskEncryptionAlgorithm	algorithm;
	const void	*exportedContext;
	PGPUInt32	sizeContext;
	PGPUInt64	firstDataBlock;
	PGPUInt64	numDataBlocks;
	PGPUInt32	procId;
	char		*deviceName;
	PGPBoolean	isReadOnly;
};

struct CPGPdiskHandlerImp::UnmountTaskInfo
{
	TaskInfoHeader	header;

	const char	*root;
	PGPBoolean	isForced;
};

struct CPGPdiskHandlerImp::UnmountAllTaskInfo
{
	TaskInfoHeader	header;
	PGPBoolean		isForced;
};

struct CPGPdiskHandlerImp::EverySecondTaskInfo
{
	TaskInfoHeader	header;
};


// Class CPGPdiskHandlerImp member functions

CPGPdiskHandlerImp::CPGPdiskHandlerImp() : 
	mInUnmountAllMode(FALSE), mWorkerThreadCreated(FALSE)
{
	Status() = mPGPdiskAllocator.Status();

	if (Status().IsntError())
		Status() = mPGPdiskListLock.Status();

	if (Status().IsntError())
		Status() = mTaskThread.Status();

	if (Status().IsntError())
		Status() = mEverySecondTimer.Status();
}

const char * 
CPGPdiskHandlerImp::GetDiskDevNameByPath(const char *path) const
{
	CPGPdisk	*pDisk	= GetPGPdiskByPath(path);

	if (IsNull(pDisk))
		return NULL;

	return pDisk->DeviceName();
}

const char * 
CPGPdiskHandlerImp::GetDiskDevNameByRoot(const char *root) const
{
	CPGPdisk	*pDisk	= GetPGPdiskByRoot(root);

	if (IsNull(pDisk))
		return NULL;

	return pDisk->DeviceName();
}

CComboError 
CPGPdiskHandlerImp::ExportPGPdiskInfo(
	PGPdiskInfo	*pInfoArray, 
	PGPUInt32	numElems, 
	PGPUInt32	&elemsFilled) const
{
	pgpAssertAddrValid(pInfoArray, PGPdiskInfo);

	CComboError	error;;
	mPGPdiskListLock.StartReading();

	if (numElems < NumPGPdisks())
	{
		elemsFilled = NumPGPdisks();	// total number needed
		error.pgpErr = kPGPError_BufferTooSmall;
	}

	if (error.IsntError())
	{
		CPGPdisk	*pDisk	= mPGPdiskList.Head();
		PGPUInt32	i		= 0;

		while (IsntNull(pDisk))
		{
			strcpy(pInfoArray[i].path, pDisk->Path());
			strcpy(pInfoArray[i].root, pDisk->Root());
			strcpy(pInfoArray[i].deviceName, pDisk->DeviceName());

			pInfoArray[i].isReadOnly = pDisk->IsReadOnly();
			pInfoArray[i].hasTimedOut = pDisk->HasInactivityTimedOut();

			i++;
			pDisk = mPGPdiskList.Next(pDisk);
		}

		elemsFilled = i;
	}

	mPGPdiskListLock.StopReading();
	
	return error;
}

CComboError 
CPGPdiskHandlerImp::SetInactivityTimeout(const char *root, PGPUInt32 seconds)
{
	pgpAssertStrValid(root);

	CComboError	error;

	if (seconds > kPGPdiskMaxMinutesTimeout * 60)
		seconds = kPGPdiskMaxMinutesTimeout * 60;

	CPGPdisk	*pDisk	= GetPGPdiskByRoot(root);

	if (IsNull(pDisk))
		error.pgpErr = kPGPError_ItemNotFound;

	if (error.IsntError())
		error = pDisk->SetInactivityTimeout(seconds);

	return error;
}

PGPBoolean 
CPGPdiskHandlerImp::AreDisksTimedOut()
{
	PGPBoolean	areTimedOut	= FALSE;

	mPGPdiskListLock.StartReading();
	CPGPdisk	*pDisk	= mPGPdiskList.Head();

	while (IsntNull(pDisk))
	{
		if (pDisk->HasInactivityTimedOut())
			areTimedOut = TRUE;

		pDisk = mPGPdiskList.Next(pDisk);
	}

	mPGPdiskListLock.StopReading();

	return areTimedOut;
}

void 
CPGPdiskHandlerImp::ResetTimedOutDisk(const char *root)
{
	pgpAssertStrValid(root);

	CPGPdisk	*pDisk	= GetPGPdiskByRoot(root);
	pDisk->ResetTimedOut();
}

CComboError 
CPGPdiskHandlerImp::MountPGPdisk(
	const char *path, 
	const char *root, 
	PGPdiskEncryptionAlgorithm	algorithm, 
	const void	*exportedContext, 
	PGPUInt32	sizeContext, 
	PGPUInt64	firstDataBlock, 
	PGPUInt64	numDataBlocks, 
	char		*deviceName, 
	PGPBoolean	readOnly)
{
	pgpAssertStrValid(path);
	pgpAssertStrValid(root);
	pgpAssertAddrValid(exportedContext, VoidAlign);

	CComboError	error;

	// Execute this command in thread context.
	MountTaskInfo	MTI;

	MTI.path			= path;
	MTI.root			= root;
	MTI.algorithm		= algorithm;
	MTI.exportedContext	= exportedContext;
	MTI.sizeContext		= sizeContext;
	MTI.firstDataBlock	= firstDataBlock;
	MTI.numDataBlocks	= numDataBlocks;
	MTI.procId			= GetPlatformProcessId();
	MTI.deviceName		= deviceName;
	MTI.isReadOnly		= readOnly;

	error = PerformPGPdiskTask(&MTI.header, kMountTask);

	return error;
}

CComboError 
CPGPdiskHandlerImp::UnmountPGPdisk(const char *root, PGPBoolean isForced)
{
	pgpAssertStrValid(root);

	CComboError	error;

	// Execute this command in thread context.
	UnmountTaskInfo	UTI;

	UTI.root = root;
	UTI.isForced = isForced;

	error = PerformPGPdiskTask(&UTI.header, kUnmountTask);

	return error;
}

CComboError 
CPGPdiskHandlerImp::UnmountAllPGPdisks(PGPBoolean isForced)
{
	CComboError	error;

	// Execute this command in thread context.
	UnmountAllTaskInfo	UATI;
	UATI.isForced = isForced;

	error = PerformPGPdiskTask(&UATI.header, kUnmountAllTask);

	return error;
}

void 
CPGPdiskHandlerImp::Shutdown()
{
	if (mWorkerThreadCreated)
	{
		UnmountAllPGPdisks(TRUE);
		DestroyWorkerThread();
	}
}

CComboError 
CPGPdiskHandlerImp::Startup()
{
	// Can't create thread and timer during boot because 98 barfs.
	return CComboError();
}

CComboError 
CPGPdiskHandlerImp::CreateWorkerThread()
{
	pgpAssert(!mWorkerThreadCreated);

	CComboError	error;
	error = mTaskThread.StartThread(MaxQueuedTasks);

	if (error.IsntError())
	{
		mWorkerThreadCreated = TRUE;

		error = mEverySecondTimer.StartTimer(EverySecondTimerCallback, this, 
			1000);

		if (error.IsError())
		{
			mTaskThread.KillThread();
			mWorkerThreadCreated = FALSE;
		}
	}

	return error;
}

void 
CPGPdiskHandlerImp::DestroyWorkerThread()
{
	pgpAssert(mWorkerThreadCreated);

	mEverySecondTimer.StopTimer();
	mTaskThread.KillThread();

	mWorkerThreadCreated = FALSE;
}

CComboError 
CPGPdiskHandlerImp::ConstructDeviceName(CString& deviceName) const
{
	CComboError	error;
	PGPBoolean	foundSpace	= FALSE;

	for (PGPUInt32 i = 1; i <= kPGPdiskMaxMountedDisks; i++)
	{
		error = deviceName.Format("%s%d", kPGPdiskDeviceName, i);

		if (error.IsntError())
			foundSpace = IsNull(GetPGPdiskByDeviceName(deviceName));

		if (foundSpace || error.IsError())
			break;
	}

	if (error.IsntError())
	{
		if (!foundSpace)
			error.pgpErr = kPGPClientError_MaxDisksMounted;
	}

	return error;
}

CPGPdisk * 
CPGPdiskHandlerImp::GetPGPdiskByPath(const char *path) const
{
	pgpAssertStrValid(path);

	mPGPdiskListLock.StartReading();
	CPGPdisk	*pDisk	= mPGPdiskList.Head();

	while (IsntNull(pDisk))
	{
		CString	cpath(path);

		if (cpath.Status().IsntError())
		{
			if (cpath.CompareNoCase(pDisk->Path()))
				break;
		}

		pDisk = mPGPdiskList.Next(pDisk);
	}

	mPGPdiskListLock.StopReading();

	return pDisk;
}

CPGPdisk *  
CPGPdiskHandlerImp::GetPGPdiskByRoot(const char *root) const
{
	pgpAssertStrValid(root);

	mPGPdiskListLock.StartReading();
	CPGPdisk	*pDisk	= mPGPdiskList.Head();

	while (IsntNull(pDisk))
	{
		CString	croot(root);

		if (croot.Status().IsntError())
		{
			if (croot.CompareNoCase(pDisk->Root()))
				break;
		}

		pDisk = mPGPdiskList.Next(pDisk);
	}

	mPGPdiskListLock.StopReading();

	return pDisk;
}

CPGPdisk *  
CPGPdiskHandlerImp::GetPGPdiskByDeviceName(const char *deviceName) const
{
	pgpAssertStrValid(deviceName);

	CPGPdisk	*pDisk	= NULL;
	CString		cdeviceName(deviceName);

	if (cdeviceName.Status().IsntError())
	{
		mPGPdiskListLock.StartReading();
		pDisk = mPGPdiskList.Head();

		while (IsntNull(pDisk))
		{
			if (cdeviceName == pDisk->DeviceName())
				break;

			pDisk = mPGPdiskList.Next(pDisk);
		}

		mPGPdiskListLock.StopReading();
	}

	return pDisk;
}

CComboError 
CPGPdiskHandlerImp::UnmountPGPdiskAux(CPGPdisk *pDisk, PGPBoolean isForced)
{
	pgpAssertAddrValid(pDisk, CPGPdisk);
	pgpAssert(pDisk->IsMounted());

	CComboError	error;

	// unmount all PGPdisks stacked on this one
	CPGPdisk	*pClient	= mPGPdiskList.Head();

	while (IsntNull(pClient))
	{
		CString	path(pClient->Path());
		CString	pathRoot;
		CString	root(pDisk->Root());

		if (path.Status().IsntError() && 
			pathRoot.Status().IsntError() && 
			root.Status().IsntError() && 
			(pDisk != pClient) && 
			(path.Length() > root.Length()) && 
			path.Left(root.Length(), pathRoot).IsntError() && 
			pathRoot.CompareNoCase(root) && 
			UnmountPGPdiskAux(pClient, isForced).IsntError())
		{
			pClient = mPGPdiskList.Head();
		}
		else
		{
			pClient = mPGPdiskList.Next(pClient);
		}
	}

	// Peform the unmount.
	if (error.IsntError())
		error = pDisk->Unmount(isForced);

	if (error.IsntError())
	{
		// De-list and delete the PGPdisk object.
		mPGPdiskList.Remove(pDisk);
		mPGPdiskAllocator.Free(pDisk);
	}

	return error;
}

CComboError 
CPGPdiskHandlerImp::MountPGPdiskCallback(MountTaskInfo *pMTI)
{
	pgpAssertAddrValid(pMTI, MountTaskInfo);

	CComboError	error;
	
	// Check if we've already mounted this PGPdisk.
	if (IsFileAMountedPGPdisk(pMTI->path))
		error.pgpErr = kPGPClientError_DiskAlreadyMounted;

	if (error.IsntError())
	{
		// Create a new PGPdisk object.
		CPGPdisk	*pDisk;
		CString		deviceName;

		error = deviceName.Status();

		if (error.IsntError())
			error = ConstructDeviceName(deviceName);

		if (error.IsntError())
			error = mPGPdiskAllocator.Allocate(pDisk);

		if (error.IsntError())
		{
			error = pDisk->Status();

			if (error.IsntError())
			{
				// Mount the PGPdisk.
				error = pDisk->Mount(pMTI->path, pMTI->root, deviceName, 
					pMTI->algorithm, pMTI->exportedContext, 
					pMTI->sizeContext, pMTI->firstDataBlock, 
					pMTI->numDataBlocks, GetPlatformIoHandlerFunc(), 
					pMTI->procId, pMTI->isReadOnly);
			}

			if (error.IsntError())
			{
				// Add the PGPdisk to the list.
				mPGPdiskListLock.StartWriting();
				mPGPdiskList.AddHead(pDisk);
				mPGPdiskListLock.StopWriting();

				strcpy(pMTI->deviceName, deviceName);
			}

			if (error.IsError())
				mPGPdiskAllocator.Free(pDisk);
		}
	}

	pgpAssert(pMTI->header.trueIfSync);

	return error;
}

CComboError 
CPGPdiskHandlerImp::UnmountPGPdiskCallback(UnmountTaskInfo *pUTI)
{
	pgpAssertAddrValid(pUTI, UnmountTaskInfo);
	pgpAssert(pUTI->header.trueIfSync);

	CComboError	error;
	mPGPdiskListLock.StartWriting();

	// Find the PGPdisk in question.
	CPGPdisk	*pDisk	= mPGPdiskList.Head();

	while (IsntNull(pDisk))
	{
		CString	root(pDisk->Root());

		if (root.Status().IsntError())
		{
			if (root.CompareNoCase(pUTI->root))
				break;
		}

		pDisk = mPGPdiskList.Next(pDisk);
	}

	if (IsNull(pDisk))
		error.pgpErr = kPGPError_ItemNotFound;

	if (error.IsntError())
		error = UnmountPGPdiskAux(pDisk, pUTI->isForced);

	mPGPdiskListLock.StopWriting();

	return error;
}

CComboError 
CPGPdiskHandlerImp::UnmountAllPGPdisksCallback(UnmountAllTaskInfo *pUATI)
{
	pgpAssertAddrValid(pUATI, UnmountAllTaskInfo);
	pgpAssert(pUATI->header.trueIfSync);

	CComboError	error;
	mPGPdiskListLock.StartWriting();

	CPGPdisk	*pDisk	= mPGPdiskList.Head();

	// Unmount all PGPdisks, skipping those we can't unmount.
	while (IsntNull(pDisk))
	{
		CComboError	tempErr	= UnmountPGPdiskAux(pDisk, pUATI->isForced);

		if (tempErr.IsError())
		{
			error = tempErr;
			pDisk = mPGPdiskList.Next(pDisk);
		}
		else
		{
			pDisk = mPGPdiskList.Head();
		}
	}

	mPGPdiskListLock.StopWriting();
	
	if (NumPGPdisks() > 0)
		return error;
	else
		return CComboError();
}

void 
CPGPdiskHandlerImp::EverySecondThreadCallback(EverySecondTaskInfo *pESTI)
{
	pgpAssertAddrValid(pESTI, EverySecondTaskInfo);

	static PGPUInt32	lastToggleTime	= 0;
	static PGPUInt32	lastVerifyTime	= 0;

	// Set toggle/verify time flags.
	PGPBoolean	shouldToggleContexts	= FALSE;
	PGPBoolean	shouldVerifyContexts	= FALSE;

	if (lastToggleTime++ > ContextToggleSeconds)
	{
		lastToggleTime = 0;
		shouldToggleContexts = TRUE;
	}

	if (lastVerifyTime++ > ContextVerifySeconds)
	{
		lastVerifyTime = 0;
		shouldVerifyContexts = TRUE;
	}

	// If we are in 'unmount all mode', keep trying to unmount all PGPdisks.
	if (mInUnmountAllMode)
	{
		CComboError	error;

		// We're in thread context, fake an 'UnmountAllTask' callback.
		UnmountAllTaskInfo	UATI;

		UATI.header.trueIfSync = TRUE;
		UATI.isForced = FALSE;

		error = UnmountAllPGPdisksCallback(&UATI);

		if (error.IsntError())
			mInUnmountAllMode = FALSE;
	}

	// Perform various tasks on each PGPdisk.
	mPGPdiskListLock.StartReading();
	CPGPdisk	*pDisk	= mPGPdiskList.Head();

	while (IsntNull(pDisk))
	{
		if (shouldToggleContexts)
			pDisk->FlipContexts();
		
		if (shouldVerifyContexts)
		{
			CComboError	error	= pDisk->ValidateContexts();

			// Forcibly unmount PGPdisk if cipher context is invalid.
			if (error.IsError())
			{
				pgpAssert("PGPdisk cipher context corrupted!");
				UnmountPGPdiskAux(pDisk, TRUE);
			}
		}

		pDisk = mPGPdiskList.Next(pDisk);
	}

	mPGPdiskListLock.StopReading();
}

void 
CPGPdiskHandlerImp::PGPdiskTaskCallbackAux(TaskInfoHeader *pTIH)
{
	pgpAssertAddrValid(pTIH, TaskInfoHeader);

	CComboError	error;

	switch (pTIH->task)
	{
	case kMountTask:
		error = MountPGPdiskCallback(
			reinterpret_cast<MountTaskInfo *>(pTIH));
		break;

	case kUnmountTask:
		error = UnmountPGPdiskCallback(
			reinterpret_cast<UnmountTaskInfo *>(pTIH));
		break;

	case kUnmountAllTask:
		error = UnmountAllPGPdisksCallback(
			reinterpret_cast<UnmountAllTaskInfo *>(pTIH));
		break;

	case kEverySecondTask:
		EverySecondThreadCallback(
			reinterpret_cast<EverySecondTaskInfo *>(pTIH));
		break;

	default:
		pgpAssert(FALSE);
		error.pgpErr = kPGPError_BadParams;
		break;
	}

	if (pTIH->trueIfSync)
		pTIH->error = error;
	else
		delete pTIH;
}

void 
_cdecl 
CPGPdiskHandlerImp::PGPdiskTaskCallback(void *refPtr)
{
	TaskInfoHeader	*pTIH	= static_cast<TaskInfoHeader *>(refPtr);
	pgpAssertAddrValid(pTIH, TaskInfoHeader);

	CPGPdiskHandlerImp	*pImp	= static_cast<CPGPdiskHandlerImp *>(
		pTIH->origThis);
	pgpAssertAddrValid(pImp, CPGPdiskHandlerImp);

	pImp->PGPdiskTaskCallbackAux(pTIH);
}

CComboError 
CPGPdiskHandlerImp::PerformPGPdiskTask(
	TaskInfoHeader	*pTIH, 
	HandlerTask		task)
{
	CComboError	error;

	// Create worker thread on demand, but leave it running.
	if (!mWorkerThreadCreated)
		error = CreateWorkerThread();

	if (error.IsntError())
	{
		pgpAssertAddrValid(pTIH, TaskInfoHeader);

		pTIH->origThis		= this;
		pTIH->trueIfSync	= TRUE;
		pTIH->task			= task;

		error = mTaskThread.PerformSyncCallback(PGPdiskTaskCallback, pTIH);

		if (error.IsntError())
			error = pTIH->error;
	}

	return error;
}

CComboError 
CPGPdiskHandlerImp::SchedulePGPdiskTask(
	TaskInfoHeader	*pTIH, 
	HandlerTask		task)
{
	pgpAssertAddrValid(pTIH, TaskInfoHeader);

	CComboError	error;

	// Create worker thread on demand, but leave it running.
	if (!mWorkerThreadCreated)
		error = CreateWorkerThread();

	if (error.IsntError())
	{
		pTIH->origThis		= this;
		pTIH->trueIfSync	= FALSE;
		pTIH->task			= task;

		error = mTaskThread.PerformAsyncCallback(PGPdiskTaskCallback, pTIH);
	}

	return error;
}

void  
CPGPdiskHandlerImp::EverySecondTimerCallbackAux()
{
	// Execute this command in thread context.
	EverySecondTaskInfo	*pESTI	= new EverySecondTaskInfo;

	if (IsntNull(pESTI))
	{
		CComboError	error	= SchedulePGPdiskTask(&pESTI->header, 
			kEverySecondTask);

		if (error.IsError())
			delete pESTI;
	}
}

void 
_cdecl 
CPGPdiskHandlerImp::EverySecondTimerCallback(void *refPtr)
{
	CPGPdiskHandlerImp	*pImp	= static_cast<CPGPdiskHandlerImp *>(refPtr);
	pgpAssertAddrValid(pImp, CPGPdiskHandlerImp);

	pImp->EverySecondTimerCallbackAux();
}
