/*____________________________________________________________________________
	Copyright (C) 2002 PGP Corporation
	All rights reserved.

	$Id: pgpDiskEngineAppWorker.c,v 1.62 2002/11/26 07:07:45 dallen Exp $
____________________________________________________________________________*/

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <pthread.h>
#include <syslog.h>
#include "pgpPFLPriv.h"
#include "pgpErrors.h"
#include "pgpMem.h"
#include "pgpUtilities.h"
#include "pgpKeys.h"
#include "pgpDiskEngine.h"
#include "pgpRPCPacker.h"
#include "pgpDiskEngineKernelWorker.h"
#include "pgpDiskEngineAppWorker.h"
#include "pgpDiskEngineTypes.h"
#include "pgpDiskEngineDisk.h"

/* Some neat macros */
#define CKERR		if( IsPGPError( err ) ) goto done
#define CKNULL(x)	do { if( IsNull( x ) ) { err = kPGPError_OutOfMemory; goto done; } } while(0)
#define RETERR(x)	do { err = x; goto done; } while(0)

extern PGPBoolean		gbRunning;

typedef struct PGPdiskAppWorkerThreadData
{
	PGPContextRef					context;
	PGPMemoryMgrRef					mgr;		/* A global memory manager */
	PGPInt32						asock;
	PGPdiskKernelWorkerContextRef	kernCon;

	pthread_mutex_t *				numConnectionsMutex;
	PGPUInt32 *						pNumConnections;
} PGPdiskAppWorkerThreadData;

typedef struct PGPdiskAppWorkerContext
{
	PGPMemoryMgrRef					mgr;
	PGPdiskKernelWorkerContextRef	kernCon;
} PGPdiskAppWorkerContext;

	static PGPError
sCreateIPCSocket(
	char *				sockPath,
	PGPInt32 *			outSock )
{
	PGPError			err			= kPGPError_NoErr;
	struct sockaddr_un	sun;
	PGPInt32			sock		= 0;
	PGPInt32			retval		= 0;

	if( PGPdiskIsEngineRunning() )
	{
#if PGP_DEBUG
		printf( "PGPdiskEngine is already running\n" );
#endif
		return kPGPError_SocketsAlreadyConnected;
	}
	
	sock = socket( AF_LOCAL, SOCK_STREAM, 0 );
	if( sock < 0 )
		return kPGPError_UnknownError;
		
	(void) PGPdiskGetSockPath( sockPath );
	
	/* In case the sock is sitting around from a previous run */
	(void) unlink( sockPath );
	
	sun.sun_family = AF_LOCAL;
	pgpCopyMemory( sockPath, sun.sun_path, strlen( sockPath ) + 1 );
	
	retval = bind( sock, (struct sockaddr *) &sun, SUN_LEN( &sun ) );
	if( retval < 0 )
		return kPGPError_UnknownError;
		
	retval = listen( sock, 10 );
	if( retval < 0 )
		return kPGPError_UnknownError;
		
	*outSock = sock;

	return err;
}

	static PGPError
sDestroyIPCSocket(
	PGPInt32		sock,
	char *			sockPath )
{
	PGPError		err			= kPGPError_NoErr;
	
	if( sock > 0 )
		(void) close( sock );
		
	(void) unlink( sockPath );
	
	return err;
}

	static PGPError
sConnectToClient(
	const char *		sockPath,
	PGPInt32 *			outSock )
{
	PGPInt32			sock		= 0;
	struct sockaddr_un	sun;
	PGPInt32			retval		= 0;
	
	PGPValidatePtr( sockPath );
	PGPValidatePtr( outSock );
		
	sock = socket( AF_LOCAL, SOCK_STREAM, 0 );
	if( sock < 0 )
		goto done;
		
	sun.sun_family = AF_LOCAL;
	memcpy( sun.sun_path, sockPath, strlen( sockPath ) + 1 );
	
	retval = connect( sock, (struct sockaddr *) &sun, sizeof( sun ) );
	if( retval < 0 )
		goto done;
		
	*outSock = sock;
	
done:
	if( retval < 0 )
	{
		if( sock > 0 )
			(void) close( sock );
		return kPGPError_UnknownError;
	}
	return kPGPError_NoErr;
}

	static PGPError
sNewBEUser(
	PGPdiskBEContextRef			diskcon,
	PGPdiskBEUserRef *			outUser )
{
	PGPError					err			= kPGPError_NoErr;
	PGPdiskBEUserRef			user		= kInvalidPGPdiskBEUserRef;
	PGPdiskBEUserRef			walk		= kInvalidPGPdiskBEUserRef;
	
	user = PGPNewData( diskcon->mgr,
		sizeof( PGPdiskBEUser ),
		kPGPMemoryMgrFlags_Clear );
	CKNULL( user );
	
	pgpClearMemory( user, sizeof( PGPdiskBEUser ) );
	
	user->diskcon = diskcon;
	
	/* Add to diskcon's list of users */
	if( IsNull( diskcon->userList ) )
		diskcon->userList = user;
	else
	{
		for( walk = diskcon->userList; IsntNull( walk->next ); walk = walk->next );
			/* nothing */ ;
		walk->next = user;
	}
	
	*outUser = user;
	
done:
	return err;
}

	static PGPError
sFreeBEUser(
	PGPdiskBEUserRef		user )
{
	PGPdiskBEUserRef		walk		= kInvalidPGPdiskBEUserRef;
	
	if( IsntNull( user->passphrase ) )
	{
		pgpClearMemory( user->passphrase, strlen( user->passphrase ) );
		(void) PGPFreeData( user->passphrase );
	}
	
	/* Remember to remove user from the context's list of valid users */
	if( user == user->diskcon->userList )
	{
		user->diskcon->userList = user->diskcon->userList->next;
		(void) PGPFreeData( user );
	}
	else
	{
		for( walk = user->diskcon->userList; IsntNull( walk ); walk = walk->next )
		{
			if( walk->next == user )
			{
				walk->next = walk->next->next;
				(void) PGPFreeData( user );
				break;
			}
		}
	}
	return kPGPError_NoErr;
}

	static PGPError
sNewBEDisk(
	PGPdiskBEContextRef			diskcon,
	PGPdiskBEDiskRef	*		outDisk )
{
	PGPError					err			= kPGPError_NoErr;
	PGPdiskBEDiskRef			disk		= kInvalidPGPdiskBEDiskRef;
	PGPdiskBEDiskRef			walk		= kInvalidPGPdiskBEDiskRef;
	
	disk = PGPNewData( PGPPeekContextMemoryMgr( diskcon->context ),
		sizeof( PGPdiskBEDisk ),
		kPGPMemoryMgrFlags_Clear );
	CKNULL( disk );
	
	pgpClearMemory( disk, sizeof( PGPdiskBEDisk ) );
	
	disk->diskcon	= diskcon;
	
	/* Add to diskcon's list of disks */
	if( IsNull( diskcon->diskList ) )
		diskcon->diskList = disk;
	else
	{
		for( walk = diskcon->diskList; IsntNull( walk->next ); walk = walk->next );
			/* nothing */ ;
		walk->next = disk;
	}
	
	*outDisk = disk;
	
done:
	return err;
}

	static PGPError
sFreeBEDisk(
	PGPdiskBEDiskRef		disk )
{
	PGPdiskBEDiskRef		walk		= kInvalidPGPdiskBEDiskRef;
	PGPdiskBEDiskRef		diskToFree	= kInvalidPGPdiskBEDiskRef;
	PGPdiskBEUserRef		user		= kInvalidPGPdiskBEUserRef;
	PGPdiskBEUserRef		nextuser	= kInvalidPGPdiskBEUserRef;
	
	/*
	 * Walk through our list of known BEUsers.  If the BEUser is assocated with the same
	 * PGPdiskIODisk as the BEDisk we are freeing, free the BEUser.
	 */
	for( user = disk->diskcon->userList; IsntNull( user ); user = nextuser )
	{
		nextuser = user->next;
		if( IsntNull( user->ioDiskUser ) )
		{
			if( user->ioDiskUser->disk == disk->ioDisk )
				(void) sFreeBEUser( user );
		}
	}

	if( PGPdiskIODiskRefIsValid( disk->ioDisk ) )
		(void) PGPdiskEngineCloseDisk( disk->ioDisk );
		
	/* Remember to remove disk from the context's list of valid disks */
	if( disk == disk->diskcon->diskList )
	{
		disk->diskcon->diskList = disk->diskcon->diskList->next;
		diskToFree = disk;
	}
	else
	{
		for( walk = disk->diskcon->diskList; IsntNull( walk ); walk = walk->next )
		{
			if( walk->next == disk )
			{
				diskToFree = walk->next;
				walk->next = walk->next->next;
				break;
			}
		}
	}

	if( PGPdiskBEDiskRefIsValid( diskToFree ) )
		(void) PGPFreeData( diskToFree );

	return kPGPError_NoErr;
}

	static PGPBoolean
sIsDiskRefValid(
	PGPdiskBEContextRef	diskcon,
	PGPdiskBEDiskRef	disk )
{
	PGPdiskBEDiskRef	walk		= kInvalidPGPdiskBEDiskRef;
	
	for( walk = diskcon->diskList; IsntNull( walk ); walk = walk->next )
	{
		if( walk == disk )
			return TRUE;
	}
	/* Not a valid disk */
	return FALSE;
}

	static PGPBoolean
sIsUserRefValid(
	PGPdiskBEContextRef	diskcon,
	PGPdiskBEUserRef	user )
{
	PGPdiskBEUserRef	walk		= kInvalidPGPdiskBEUserRef;
	
	for( walk = diskcon->userList; IsntNull( walk ); walk = walk->next )
	{
		if( walk == user )
			return TRUE;
	}
	/* Not a valid user */
	return FALSE;
}

	static PGPError
sNewBEContext(
	PGPContextRef				context,
	PGPdiskBEContextRef *		outDiskcon )
{
	PGPError					err			= kPGPError_NoErr;
	PGPdiskBEContextRef			diskcon		= kInvalidPGPdiskBEContextRef;

	diskcon = PGPNewData( PGPPeekContextMemoryMgr( context ),
		sizeof( PGPdiskBEContext ),
		kPGPMemoryMgrFlags_Clear );
	CKNULL( diskcon );

	diskcon->context = context;

	*outDiskcon = diskcon;

done:
	return err;
}

	static PGPError
sFreeBEContext(
	PGPdiskBEContextRef			diskcon )
{
	PGPdiskBEDiskRef			disk		= kInvalidPGPdiskBEDiskRef;
	PGPdiskBEDiskRef			nextdisk	= kInvalidPGPdiskBEDiskRef;
	PGPdiskBEUserRef			user		= kInvalidPGPdiskBEUserRef;
	PGPdiskBEUserRef			nextuser	= kInvalidPGPdiskBEUserRef;
	
	/* Free all items used by context */
	for( disk = diskcon->diskList; IsntNull( disk ); disk = nextdisk )
	{
		nextdisk = disk->next;
		(void) sFreeBEDisk( disk );
	}
	for( user = diskcon->userList; IsntNull( user ); user = nextuser )
	{
		nextuser = user->next;
		(void) sFreeBEUser( user );
	}
	
	if( PGPKeyDBRefIsValid( diskcon->keydb ) )
		(void) PGPFreeKeyDB( diskcon->keydb );

	(void) PGPFreeData( diskcon );

	return kPGPError_NoErr;
}

	static PGPError
sAcceptSocket(
	PGPInt32		sock,
	PGPInt32 *		outASock)
{
	PGPError		err			= kPGPError_NoErr;
	PGPInt32		asock		= 0;
	struct sockaddr_un	sun;
	fd_set			sockSet;
	struct timeval	tv;
	PGPInt32		retval		= 0;
	char			buf[4];
	PGPSize			len			= 0;

	len = sizeof( sun );
	asock = accept( sock, (struct sockaddr *) &sun, (int *) &len );
	if( asock < 0 )
		RETERR( kPGPError_UnknownError );

	FD_ZERO( &sockSet );
	FD_SET( asock, &sockSet );

	tv.tv_sec = 30;
	tv.tv_usec = 0;

	/*
	 * Little hack - In order to minimize our calling of PGPsdkInit, we'll try to peek the
	 * first 4 bytes from the client connecting to us.  If we don't find 4 bytes, assume
	 * it's just somebody calling PGPdiskIsEngineRunning, and let's not call PGPsdkInit this
	 * time.
	 */
	retval = select( asock + 1, &sockSet, NULL, NULL, &tv );
	if( retval != 1 )
	{
		(void) close( asock );
		RETERR( kPGPError_UnknownError );
	}

	retval = recv( asock, buf, sizeof( buf ), MSG_PEEK );
	if( retval != sizeof( buf ) )
	{
		(void) close( asock );
		RETERR( kPGPError_UnknownError );
	}

	*outASock = asock;

done:
	return err;
}

	static PGPError
sGetNewContextMessage(
	PGPContextRef	context,
	PGPInt32		sock,
	PGPByte **		outBuf,
	PGPSize *		outLen )
{
	PGPError		err			= kPGPError_NoErr;
	PGPByte *		buf			= NULL;
	PGPSize			len			= 0;

	err = PGPrpcRead( context, sock, &buf, &len ); CKERR;

	*outBuf = buf;
	*outLen = len;

done:
	if( IsPGPError( err ) )
	{
		if( IsntNull( buf ) )
			(void) PGPFreeData( buf );
	}

	return err;
}

	static PGPError
sLoadKeyrings(
	PGPdiskBEContextRef	diskcon,
	char *				pszPublicFilename,
	char *				pszPrivateFilename )
{
	PGPError			err			= kPGPError_NoErr;
	PGPFileSpecRef		pubFile		= kInvalidPGPFileSpecRef;
	PGPFileSpecRef		privFile	= kInvalidPGPFileSpecRef;
	PGPKeyDBRef			keydb		= kInvalidPGPKeyDBRef;
	
	PGPdiskValidateBEContextRef( diskcon );
	PGPValidatePtr( pszPublicFilename );
	PGPValidatePtr( pszPrivateFilename );
	
	err = PGPNewFileSpecFromFullPath( diskcon->context, pszPublicFilename, &pubFile ); CKERR;
	err = PGPNewFileSpecFromFullPath( diskcon->context, pszPrivateFilename, &privFile ); CKERR;
	
	err = PGPOpenKeyDBFile( diskcon->context, kPGPOpenKeyDBFileOptions_Create,
		pubFile, privFile, &keydb ); CKERR;

	if( PGPKeyDBRefIsValid( diskcon->keydb ) )
	{
		/* We're re-reading the keydb after a change */
		(void) PGPFreeKeyDB( diskcon->keydb );
	}
	diskcon->keydb = keydb;
		
done:
	if( PGPFileSpecRefIsValid( pubFile ) )
		(void) PGPFreeFileSpec( pubFile );
	if( PGPFileSpecRefIsValid( privFile ) )
		(void) PGPFreeFileSpec( privFile );
	return err;
}

	static PGPError
sSetADK(
	PGPdiskBEContextRef	diskcon,
	PGPKeyID *			pKeyID )
{
	PGPError			err			= kPGPError_NoErr;
	PGPKeyDBObjRef		key			= kInvalidPGPKeyDBObjRef;
	
	PGPValidatePtr( pKeyID );
	
	err = PGPFindKeyByKeyID( diskcon->keydb, pKeyID, &key ); CKERR;
	diskcon->adkKey = key;
	
done:
	return err;
}

	static PGPError
sSetGlobalDiskTimeout(
	PGPdiskBEContextRef	diskcon,
	PGPUInt32			uTimeout )
{
	PGPError			err			= kPGPError_NoErr;

	diskcon->globalTimeout = uTimeout;
	err = PGPdiskEngineUpdateGlobalDiskTimeout( uTimeout ); CKERR;

done:
	return err;
}

	static PGPError
sSetEjectOnLogout(
	PGPdiskBEContextRef	diskcon,
	PGPBoolean			bEjectOnLogout )
{
	diskcon->ejectOnLogout = bEjectOnLogout;
	return kPGPError_NoErr;
}

	static PGPError
sSetPassphraseCaching(
	PGPdiskBEContextRef	diskcon,
	PGPBoolean			bCachePassphrases,
	PGPUInt32			uTimeout,
	PGPBoolean			bShareCache )
{
	diskcon->cachePassphrases = bCachePassphrases;
	diskcon->cacheTimeout = uTimeout;
	diskcon->shareCache = bShareCache;
	return kPGPError_NoErr;
}

	static PGPError
sSetUnmountOnSleep(
	PGPdiskBEContextRef	diskcon,
	PGPBoolean			bUnmountOnSleep )
{
	diskcon->kernCon->unmountOnSleep = bUnmountOnSleep;
	return kPGPError_NoErr;
}

	static PGPError
sSetForcedAutoDiskUnmount(
	PGPdiskBEContextRef	diskcon,
	PGPBoolean			bForced )
{
	diskcon->forcedUnmount = bForced;
	diskcon->kernCon->forcedUnmount = bForced;
	return kPGPError_NoErr;
}

	static PGPError
sSetPreventSleep(
	PGPdiskBEContextRef	diskcon,
	PGPBoolean			bPreventSleep )
{
	diskcon->kernCon->preventSleep = bPreventSleep;
	return kPGPError_NoErr;
}

	static PGPBoolean
sIsDiskBeingReEncrypted(
	PGPdiskBEDiskRef	disk )
{
	PGPError			err			= kPGPError_NoErr;
	PGPBoolean			b		= FALSE;
	
	err = PGPdiskEngineGetDiskBooleanProperty( disk->ioDisk, kPGPdiskDiskProp_IsBeingReEncrypted, &b );
	if( IsPGPError( err ) )
		b = FALSE;
	
	return b; 
}

	static PGPError
sNewUserWithSym(
	PGPdiskBEContextRef	diskcon,
	char *				pszUsername,
	char *				pszPassphrase,
	PGPBoolean			bReadOnly,
	PGPdiskBEUserRef *	outUser )
{
	PGPError			err			= kPGPError_NoErr;
	PGPdiskBEUserRef	user		= kInvalidPGPdiskBEUserRef;

	PGPValidatePtr( pszUsername );
	PGPValidatePtr( pszPassphrase );
	PGPValidatePtr( outUser );

	err = sNewBEUser( diskcon, &user ); CKERR;
	
	user->bSymUser = TRUE;

	user->passphrase = PGPNewSecureData( PGPPeekContextMemoryMgr( diskcon->context ),
		strlen( pszPassphrase ) + 1,
		kPGPMemoryMgrFlags_Clear );
	CKNULL( user->passphrase );
	strcpy( user->passphrase, pszPassphrase );
	
	strcpy( user->username, pszUsername );
	user->bReadOnly = bReadOnly;

	*outUser = user;

done:
	return err;
}

	static PGPError
sNewUserWithPub(
	PGPdiskBEContextRef	diskcon,
	PGPKeyID *			pKeyID,
	PGPBoolean			bReadOnly,
	PGPBoolean			bIsLocked,
	PGPdiskBEUserRef *	outUser )
{
	PGPError			err			= kPGPError_NoErr;
	PGPdiskBEUserRef	user		= kInvalidPGPdiskBEUserRef;
	PGPKeyDBObjRef		key			= kInvalidPGPKeyDBObjRef;
	char				szUserID[kPGPMaxUserIDSize];
	PGPSize				userIDSize	= 0;

	PGPValidatePtr( pKeyID );
	PGPValidatePtr( outUser );

	err = sNewBEUser( diskcon, &user ); CKERR;
	
	/* Make sure we have the key */
	err = PGPFindKeyByKeyID( diskcon->keydb, pKeyID, &key ); CKERR;
	
	/* Extra check to catch sdk frontend-backend inconsistencies */
	err = PGPGetPrimaryUserIDName( key, szUserID, sizeof( szUserID ), &userIDSize ); CKERR;

	user->bSymUser	= FALSE;
	user->bReadOnly	= bReadOnly;
	pgpCopyMemory( pKeyID, &user->keyid, sizeof( PGPKeyID ) );
	user->bLocked	= bIsLocked;

	*outUser = user;

done:
	return err;
}

	static PGPError
sGetUserPGPKeyID(
	PGPdiskBEContextRef	diskcon,
	PGPdiskBEUserRef	user,
	PGPKeyID *			pKeyID )
{
	PGPError			err			= kPGPError_NoErr;
	
	PGPValidatePtr( pKeyID );

	pgpCopyMemory( &user->keyid, pKeyID, sizeof( PGPKeyID ) );
	
	return err;
}

	static PGPError
sIsUserPassphraseCorrect(
	PGPdiskBEContextRef	diskcon,
	PGPdiskBEUserRef	user,
	const char *		pszPassphrase,
	PGPBoolean *		bCorrect )
{
	PGPError			err			= kPGPError_NoErr;

	if( !sIsUserRefValid( diskcon, user ) )
		RETERR( kPGPError_BadParams );
	
	err = PGPdiskEngineIsPassphraseCorrect( diskcon, user->ioDiskUser, pszPassphrase, bCorrect ); CKERR;
	
done:
	return err;
}

	static PGPError
sUserChangePassphrase(
	PGPdiskBEContextRef	diskcon,
	PGPdiskBEUserRef	user,
	const char *		pszPassphrase,
	const char *		pszNewPassphrase )
{
	PGPError			err			= kPGPError_NoErr;

	if( !sIsUserRefValid( diskcon, user ) )
		RETERR( kPGPError_BadParams );
		
	if( user->bDisabled )
		RETERR( kPGPdiskError_IsDisabled );
	
	err = PGPdiskEngineChangeUserPassphrase( diskcon, user->ioDiskUser, pszPassphrase, pszNewPassphrase ); CKERR;

done:
	return err;
}

	static PGPError
sGetUserBooleanProp(
	PGPdiskBEContextRef	diskcon,
	PGPdiskBEUserRef	user,
	PGPUInt32			prop,
	PGPBoolean *		bProperty )
{
	PGPError			err			= kPGPError_NoErr;
	
	PGPValidatePtr( bProperty );
	
	if( !sIsUserRefValid( diskcon, user ) )
		RETERR( kPGPError_BadParams );
		
	switch( prop )
	{
		case kPGPdiskUserProp_HasPublicKey:
			*bProperty = !user->bSymUser;
			break;
		case kPGPdiskUserProp_HasSymmetricKey:
			*bProperty = user->bSymUser;
			break;
		case kPGPdiskUserProp_IsReadOnly:
			*bProperty = user->bReadOnly;
			break;
		case kPGPdiskUserProp_IsLocked:
			*bProperty = user->bLocked;
			break;
		case kPGPdiskUserProp_IsAdminUser:
			*bProperty = user->bAdminUser;
			break;
		case kPGPdiskUserProp_IsDisabled:
			*bProperty = user->bDisabled;
			break;
		case kPGPdiskUserProp_CanSetReadOnly:
			*bProperty = !user->bAdminUser;
			break;
		case kPGPdiskUserProp_CanSetLocked:
			*bProperty = !user->bSymUser;
			break;
		case kPGPdiskUserProp_CanSetUserName:
			*bProperty = user->bSymUser;
			break;
		default:
			RETERR( kPGPError_InvalidProperty );
			break;
	}
	
done:
	return err;
}

	static PGPError
sSetUserBooleanProp(
	PGPdiskBEContextRef	diskcon,
	PGPdiskBEUserRef	user,
	PGPUInt32			prop,
	PGPBoolean			bProperty )
{
	PGPError			err			= kPGPError_NoErr;
	PGPBoolean			bCanSet		= TRUE;
	PGPBoolean			bIsAdmin	= TRUE;
	
	if( !sIsUserRefValid( diskcon, user ) )
		RETERR( kPGPError_BadParams );
		
	switch( prop )
	{
		case kPGPdiskUserProp_IsReadOnly:
			err = sGetUserBooleanProp( diskcon, user, kPGPdiskUserProp_CanSetReadOnly, &bCanSet ); CKERR;
			if( bCanSet )
				user->bReadOnly = bProperty;
			break;
			
		case kPGPdiskUserProp_IsLocked:
			err = sGetUserBooleanProp( diskcon, user, kPGPdiskUserProp_CanSetLocked, &bCanSet ); CKERR;
			if( bCanSet )
				user->bReadOnly = bProperty;
			break;
			
		case kPGPdiskUserProp_IsDisabled:
			err = sGetUserBooleanProp( diskcon, user, kPGPdiskUserProp_IsAdminUser, &bIsAdmin ); CKERR;
			if( !bIsAdmin )
				user->bDisabled = bProperty;
			else
				RETERR( kPGPError_FeatureNotAvailable );
			break;
			
		default:
			RETERR( kPGPError_InvalidProperty );
			break;
	}
	
	if( !bCanSet )
		RETERR( kPGPError_FeatureNotAvailable );
	
	if( IsntNull( user->ioDiskUser ) )
	{
		err = PGPdiskEngineBEUserModified( user );
		CKERR;
	}
	
done:
	return err;
}

	static PGPError
sGetUserDataProp(
	PGPdiskBEContextRef	diskcon,
	PGPdiskBEUserRef	user,
	PGPUInt32			prop,
	char **				pszProperty )
{
	PGPError			err			= kPGPError_NoErr;
	PGPSize				size		= 0;
	char				szKeyID[kPGPMaxKeyIDStringSize];
	PGPKeyDBObjRef		key			= kInvalidPGPKeyDBObjRef;
	
	if( !sIsUserRefValid( diskcon, user ) )
		RETERR( kPGPError_BadParams );
		
	switch( prop )
	{
		case kPGPdiskUserProp_UserName:
			if( user->bSymUser )
			{
				*pszProperty = PGPNewData( PGPPeekContextMemoryMgr( diskcon->context ),
					strlen( user->username ) + 1,
					kPGPMemoryMgrFlags_Clear );
				CKNULL( *pszProperty );
				strcpy( *pszProperty, user->username );
			}
			else
			{
				*pszProperty = PGPNewData( PGPPeekContextMemoryMgr( diskcon->context ),
					kPGPMaxUserIDSize + 1,
					kPGPMemoryMgrFlags_Clear );
				CKNULL( *pszProperty );
				
				err = PGPFindKeyByKeyID( diskcon->keydb, &user->keyid, &key );
				if( IsntPGPError( err ) )
				{
					err = PGPGetPrimaryUserIDName( key, *pszProperty, kPGPMaxUserIDSize + 1, &size );
					CKERR;
				}
				else
				{
					err = PGPGetKeyIDString( &user->ioDiskUser->user.pubUser.keyID, kPGPKeyIDString_Abbreviated, szKeyID ); CKERR;
					sprintf( *pszProperty, "<Unknown User> (Key ID is %s)", szKeyID );
				}
			}
			break;
		default:
			RETERR( kPGPError_InvalidProperty );
			break;
	}
	
done:
	return err;
}

	static PGPError
sSetUserDataProp(
	PGPdiskBEContextRef	diskcon,
	PGPdiskBEUserRef	user,
	PGPUInt32			prop,
	char *				pszProperty )
{
	PGPError			err			= kPGPError_NoErr;
	
	if( !sIsUserRefValid( diskcon, user ) )
		RETERR( kPGPError_BadParams );
		
	switch( prop )
	{
		case kPGPdiskUserProp_UserName:
			if( user->bSymUser )
				strcpy( user->username, pszProperty );
			else
				RETERR( kPGPError_BadParams );
			break;
		default:
			RETERR( kPGPError_InvalidProperty );
			break;
	}
	if( IsntNull( user->ioDiskUser ) )
	{
		err = PGPdiskEngineBEUserModified( user );
		CKERR;
	}
	
done:
	return err;
}

	static PGPError
sGetDiskBooleanProp(
	PGPdiskBEContextRef	diskcon,
	PGPdiskBEDiskRef	disk,
	PGPUInt32			prop,
	PGPBoolean *		bProperty )
{
	PGPError			err			= kPGPError_NoErr;
	
	PGPValidatePtr( bProperty );
	
	if( !sIsDiskRefValid( diskcon, disk ) )
		RETERR( kPGPError_BadParams );
		
	err = PGPdiskEngineGetDiskBooleanProperty( disk->ioDisk, prop, bProperty ); CKERR;
	
done:
	return err;
}

	static PGPError
sSetDiskBooleanProp(
	PGPdiskBEContextRef	diskcon,
	PGPdiskBEDiskRef	disk,
	PGPUInt32			prop,
	PGPBoolean			bProperty )
{
	PGPError			err			= kPGPError_NoErr;
	
	if( !sIsDiskRefValid( diskcon, disk ) )
		RETERR( kPGPError_BadParams );
	if( sIsDiskBeingReEncrypted( disk ) )
		RETERR( kPGPdiskError_DiskNeedsReEncryption );
	if( disk->readonly )
		RETERR( kPGPdiskError_DiskIsReadOnly );
		
	err = PGPdiskEngineSetDiskBooleanProperty( disk->ioDisk, prop, bProperty ); CKERR;
	
done:
	return err;
}

	static PGPError
sGetDiskNumericProp(
	PGPdiskBEContextRef	diskcon,
	PGPdiskBEDiskRef	disk,
	PGPUInt32			prop,
	PGPInt32 *			iProperty )
{
	PGPError			err			= kPGPError_NoErr;
	
	PGPValidatePtr( iProperty );
	
	if( !sIsDiskRefValid( diskcon, disk ) )
		RETERR( kPGPError_BadParams );
		
	err = PGPdiskEngineGetDiskNumericProperty( disk->ioDisk, prop, iProperty ); CKERR;
	
done:
	return err;
}

	static PGPError
sSetDiskNumericProp(
	PGPdiskBEContextRef	diskcon,
	PGPdiskBEDiskRef	disk,
	PGPUInt32			prop,
	PGPInt32			iProperty )
{
	PGPError			err			= kPGPError_NoErr;
	
	if( !sIsDiskRefValid( diskcon, disk ) )
		RETERR( kPGPError_BadParams );
	if( sIsDiskBeingReEncrypted( disk ) )
		RETERR( kPGPdiskError_DiskNeedsReEncryption );
	if( disk->readonly )
		RETERR( kPGPdiskError_DiskIsReadOnly );
		
	err = PGPdiskEngineSetDiskNumericProperty( disk->ioDisk, prop, iProperty ); CKERR;
	
done:
	return err;
}

	static PGPError
sGetDiskDataProp(
	PGPdiskBEContextRef	diskcon,
	PGPdiskBEDiskRef	disk,
	PGPUInt32			prop,
	char **				pszProperty )
{
	PGPError			err			= kPGPError_NoErr;
	
	PGPValidatePtr( pszProperty );

	if( !sIsDiskRefValid( diskcon, disk ) )
		RETERR( kPGPError_BadParams );
		
	err = PGPdiskEngineGetDiskDataProperty( disk->ioDisk, prop, pszProperty ); CKERR;

done:
	return err;
}

	static PGPError
sSetDiskDataProp(
	PGPdiskBEContextRef	diskcon,
	PGPdiskBEDiskRef	disk,
	PGPUInt32			prop,
	char *				pszProperty )
{
	PGPError			err			= kPGPError_NoErr;
	
	PGPValidatePtr( pszProperty );

	if( !sIsDiskRefValid( diskcon, disk ) )
		RETERR( kPGPError_BadParams );
	if( sIsDiskBeingReEncrypted( disk ) )
		RETERR( kPGPdiskError_DiskNeedsReEncryption );
	if( disk->readonly )
		RETERR( kPGPdiskError_DiskIsReadOnly );
		
	err = PGPdiskEngineSetDiskDataProperty( disk->ioDisk, prop, pszProperty ); CKERR;

done:
	return err;
}

	static PGPError
sOpenDisk(
	PGPdiskBEContextRef	diskcon,
	char *				pszPath,
	PGPBoolean			bReadOnly,
	PGPdiskBEDiskRef *	outDisk )
{
	PGPError			err			= kPGPError_NoErr;
	PGPdiskBEDiskRef	disk		= kInvalidPGPdiskBEDiskRef;

	err = sNewBEDisk( diskcon, &disk ); CKERR;

	strcpy( disk->filename, pszPath );

	disk->readonly = bReadOnly;
	err = PGPdiskEngineOpenDisk( diskcon, pszPath, disk->readonly, &disk->ioDisk );
	
	if( IsPGPError( err ) )
	{
		/* Maybe the disk is on a read only file system */
		disk->readonly = TRUE;
		err = PGPdiskEngineOpenDisk( diskcon, pszPath, disk->readonly, &disk->ioDisk );
	}
	CKERR;
	
	*outDisk = disk;
	
done:
	if( IsPGPError( err ) )
		(void) sFreeBEDisk( disk );
		
	return err;
}

	static PGPError
sCreateDisk(
	PGPdiskBEContextRef	diskcon,
	const char *		pszPath,
	PGPUInt64			ullSize,
	const char *		pszDefaultMountPoint,
	PGPdiskEncryptionAlgorithm	alg,
	PGPBoolean			bPubKeyUser,
	PGPKeyID *			pKeyID,
	const char *		pszUsername,
	const char *		pszPassphrase,
	const char *		pszSockPath,
	PGPdiskBEDiskRef *	outDisk )
{
	PGPError			err				=  kPGPError_NoErr;
	PGPdiskBEDiskRef	disk			= kInvalidPGPdiskBEDiskRef;
	PGPKeyDBObjRef		key				= kInvalidPGPKeyDBObjRef;
	PGPInt32			callbackSock	= 0;

	err = sNewBEDisk( diskcon, &disk ); CKERR;

	strcpy( disk->filename, pszPath );
	disk->readonly = FALSE;
	
	if( bPubKeyUser )
	{
		err = PGPFindKeyByKeyID( diskcon->keydb, pKeyID, &key );
		CKERR;
	}
	
	err = sConnectToClient( pszSockPath, &callbackSock ); CKERR;
	
	err = PGPdiskEngineCreateDisk( diskcon,
		pszPath,
		ullSize,
		pszDefaultMountPoint,
		alg,
		bPubKeyUser,
		key,
		pszUsername,
		pszPassphrase,
		callbackSock,
		&disk->ioDisk ); CKERR;
	
	*outDisk = disk;
	
done:
	if( callbackSock > 0 )
		(void) close( callbackSock );
	if( IsPGPError( err ) )
	{
		if( PGPdiskBEDiskRefIsValid( disk ) )
			(void) sFreeBEDisk( disk );
	}
	return err;
}

	static PGPError
sConvertDisk(
	PGPdiskBEContextRef	diskcon,
	PGPdiskDiskFormat	format,
	const char *		originalFilename,
	const char *		adminPassphrase,
	const char *		newFilename,
	const char *		pszSockPath )
{
	PGPError			err				= kPGPError_NoErr;
	PGPInt32			callbackSock	= 0;


	PGPValidatePtr( originalFilename );
	PGPValidatePtr( adminPassphrase );
	PGPValidatePtr( newFilename );
	PGPValidatePtr( pszSockPath );

	err = sConnectToClient( pszSockPath, &callbackSock ); CKERR;

	switch( format )
	{
		case kPGPdiskDiskFormat_OS9:
			err = PGPdiskEngineConvertOS9Disk( diskcon,
				originalFilename,
				adminPassphrase,
				newFilename,
				callbackSock ); CKERR;
			break;
		default:
			RETERR( kPGPError_BadParams );
			break;
	}

done:
	if( IsPGPError( err ) )
		(void) PGPrpcPackAndWrite( diskcon->context, callbackSock, kPGPdiskEngineMsg_ConvertDiskCallback, "u", 0 );
	if( callbackSock > 0 )
		(void) close( callbackSock );
	return err;
}
	

	static PGPError
sReEncryptDisk(
	PGPdiskBEContextRef	diskcon,
	PGPdiskBEDiskRef	disk,
	char *				pszPassphrase,
	PGPdiskEncryptionAlgorithm newAlg,
	const char *		pszSockPath )
{
	PGPError			err				= kPGPError_NoErr;
	PGPInt32			callbackSock	= 0;
	
	if( !sIsDiskRefValid( diskcon, disk ) )
		RETERR( kPGPError_BadParams );

	err = sConnectToClient( pszSockPath, &callbackSock ); CKERR;

	err = PGPdiskEngineReEncryptDisk( diskcon,
		disk->ioDisk,
		pszPassphrase,
		newAlg,
		callbackSock ); CKERR;

done:
	if( IsPGPError( err ) )
		(void) PGPrpcPackAndWrite( diskcon->context, callbackSock, kPGPdiskEngineMsg_ReEncryptDiskCallback, "u", 0 );
	if( callbackSock > 0 )
		(void) close( callbackSock );
	return err;
}

	static PGPError
sCloseDisk(
	PGPdiskBEContextRef	diskcon,
	PGPdiskBEDiskRef	disk )
{
	PGPError			err			= kPGPError_NoErr;
	
	if( sIsDiskRefValid( diskcon, disk ) )
	{
		err = sFreeBEDisk( disk );
		CKERR;
	}
	else
		err = kPGPError_ItemNotFound;

done:
	return err;
}

	static PGPError
sFormatDisk(
	PGPdiskBEContextRef		diskcon,
	PGPdiskBEDiskRef		disk,
	const char *			pszPassphrase,
	const char *			pszVolumeName,
	PGPdiskFilesystemType	fsType )
{
	PGPError				err				= kPGPError_NoErr;
	PGPdiskIODiskUser *		adminUser		= NULL;

	adminUser = PGPdiskEngineGetAdminUser( disk->ioDisk );
	
	err = PGPdiskEngineFormatDisk( diskcon, disk->ioDisk, adminUser, pszPassphrase, pszVolumeName, fsType ); CKERR;
	
done:
	return err;
}

	static PGPError
sCanPassphraseMountDisk(
	PGPdiskBEContextRef	diskcon,
	PGPdiskBEDiskRef	disk,
	const char *		pszPassphrase,
	PGPBoolean *		bProperty )
{
	PGPError			err					= kPGPError_NoErr;
	PGPUInt32			numUsersOnDisk		= 0;
	PGPdiskBEUserRef	user				= kInvalidPGPdiskBEUserRef;
	PGPUInt32			i					= 0;
	PGPBoolean			bCorrect			= FALSE;

	if( !sIsDiskRefValid( diskcon, disk ) )
		RETERR( kPGPError_BadParams );

	if( sIsDiskBeingReEncrypted( disk ) )
		RETERR( kPGPdiskError_DiskNeedsReEncryption );

	err = PGPdiskEngineCountUsers( disk->ioDisk, &numUsersOnDisk ); CKERR;
	for( i = 0; i < numUsersOnDisk; i++ )
	{
		err = sNewBEUser( diskcon, &user ); CKERR;
		err = PGPdiskEngineGetIODiskUser( diskcon, disk->ioDisk, i, user ); CKERR;
		
		err = sIsUserPassphraseCorrect( diskcon, user, pszPassphrase, &bCorrect ); CKERR;

		/* Don't allow disabled users to mount disks */
		if( user->bDisabled )
			bCorrect = FALSE;

		(void) sFreeBEUser( user );
		user = kInvalidPGPdiskBEUserRef;
		
		if( bCorrect )
			break;
	}

	*bProperty = bCorrect;
	
done:
	if( PGPdiskBEUserRefIsValid( user ) )
		(void) sFreeBEUser( user );
	return err;
}

	static PGPError
sMountDisk(
	PGPdiskBEContextRef	diskcon,
	PGPdiskBEDiskRef	disk,
	const char *		pszRoot,
	PGPdiskBEUserRef	user,
	const char *		pszPassphrase,
	PGPBoolean			bReadOnly )
{
	PGPError			err							= kPGPError_NoErr;
	PGPBoolean			usesCustomEjectOnLogout		= FALSE;
	PGPBoolean			ejectOnLogout				= FALSE;
	
	if( !sIsDiskRefValid( diskcon, disk ) )
		RETERR( kPGPError_BadParams );
	if( !sIsUserRefValid( diskcon, user ) )
		RETERR( kPGPError_BadParams );

	if( sIsDiskBeingReEncrypted( disk ) )
		RETERR( kPGPdiskError_DiskNeedsReEncryption );
		
	if( user->bDisabled )
		RETERR( kPGPdiskError_IsDisabled );
		
	err = PGPdiskEngineGetDiskBooleanProperty( disk->ioDisk, kPGPdiskDiskProp_UsesCustomEjectOnLogout, &usesCustomEjectOnLogout ); CKERR;
	err = PGPdiskEngineGetDiskBooleanProperty( disk->ioDisk, kPGPdiskDiskProp_EjectOnLogout, &ejectOnLogout ); CKERR;
	
	/*
	 * Mount the disk readonly if we opened it readonly, or if we specifically
	 * request it to be mounted readonly
	 */
	err = PGPdiskEngineMountDisk( diskcon,
		disk->ioDisk,
		user->ioDiskUser,
		disk->readonly || user->bReadOnly || bReadOnly,
		usesCustomEjectOnLogout ? ejectOnLogout : diskcon->ejectOnLogout,
		TRUE,	/* pszRoot[0] == '\0' ? TRUE : FALSE, */
		pszRoot,
		pszPassphrase ); CKERR;

done:
	return err;
}

	static PGPError
sUnmountDisk(
	PGPdiskBEContextRef	diskcon,
	PGPdiskBEDiskRef	disk,
	PGPBoolean			bForcedUnmount )
{
	PGPError			err			= kPGPError_NoErr;
	
	if( !sIsDiskRefValid( diskcon, disk ) )
		RETERR( kPGPError_BadParams );

	err = PGPdiskEngineUnmountDisk( diskcon->kernCon,
		disk->ioDisk,
		bForcedUnmount ); CKERR;

done:
	return err;
}

	static PGPError
sDiskExportUserSet(
	PGPdiskBEContextRef	diskcon,
	PGPdiskBEDiskRef	disk,
	PGPdiskBEUserRef **	outUsers,
	PGPUInt32 *			outNumUsers )
{
	PGPError			err			= kPGPError_NoErr;
	PGPUInt32			numUsers	= 0;
	PGPUInt32			i			= 0;
	PGPdiskBEUserRef	user		= kInvalidPGPdiskBEUserRef;
	PGPdiskBEUserRef *	users		= NULL;
	
	PGPValidatePtr( outUsers );
	PGPValidatePtr( outNumUsers );
	
	if( !sIsDiskRefValid( diskcon, disk ) )
		RETERR( kPGPError_BadParams );
		
	/*
	 * Even if the disk is partially re-encrypted, we're going to allow this
	 * so that the app can use the userset to verify user passphrases and things
	 * like that.
	 */

	/* Count users */
	err = PGPdiskEngineCountUsers( disk->ioDisk, &numUsers ); CKERR;
	
	users = PGPNewData( PGPPeekContextMemoryMgr( diskcon->context ),
		numUsers * sizeof( PGPdiskBEUserRef ),
		kPGPMemoryMgrFlags_Clear );
	CKNULL( users );
	
	/* For each user */
	for( i = 0; i < numUsers; i++ )
	{
		err = sNewBEUser( diskcon, &user ); CKERR;
		err = PGPdiskEngineGetIODiskUser( diskcon, disk->ioDisk, i, user ); CKERR;
		
		users[i] = user;
	}
	
	*outUsers = users;
	*outNumUsers = numUsers;
	
done:
	return err;
}

	static PGPError
sDiskImportUserSet(
	PGPdiskBEContextRef	diskcon,
	PGPdiskBEDiskRef	disk,
	const char *		adminPassphrase,
	PGPdiskBEUserRef *	users,
	PGPUInt32			numUsers,
	PGPdiskBEUserRef	newAdminUser )
{
	PGPError				err				= kPGPError_NoErr;
	PGPUInt32				i				= 0;
	PGPUInt32				j				= 0;
	PGPUInt32				numUsersOnDisk	= 0;
	PGPBoolean				bUserIsOK		= FALSE;
	PGPdiskIODiskUser *		adminUser		= NULL;
	PGPdiskBEUserRef		user			= kInvalidPGPdiskBEUserRef;
	PGPKeyDBObjRef			key				= kInvalidPGPKeyDBObjRef;
	PGPBoolean				bHavePrivate	= FALSE;
	
	if( !sIsDiskRefValid( diskcon, disk ) )
		RETERR( kPGPError_BadParams );
	if( !sIsUserRefValid( diskcon, newAdminUser ) )
		RETERR( kPGPError_BadParams );
	for( i = 0; i < numUsers; i++ )
	{
		if( !sIsUserRefValid( diskcon, users[i] ) )
			RETERR( kPGPError_BadParams );
	}

	if( sIsDiskBeingReEncrypted( disk ) )
		RETERR( kPGPdiskError_DiskNeedsReEncryption );
		
	if( newAdminUser->bDisabled )
		RETERR( kPGPdiskError_IsDisabled );
		
	if( disk->readonly )
		RETERR( kPGPdiskError_DiskIsReadOnly );
		
	if( !newAdminUser->bSymUser )
	{
		err = PGPFindKeyByKeyID( diskcon->keydb, &newAdminUser->keyid, &key ); CKERR;
		err = PGPGetKeyDBObjBooleanProperty( key, kPGPKeyProperty_IsSecret, &bHavePrivate ); CKERR;
		if( !bHavePrivate )
			RETERR( kPGPError_SecretKeyNotFound );
	}
	
	adminUser = PGPdiskEngineGetAdminUser( disk->ioDisk );
	
	/*
	 * We can't just remove every user and re-add them because we don't know all the
	 * required information to add each user (such as symmetric key user passphrases).
	 * What we'll do instead is add any new users, then go back and remove any old
	 * users that aren't in the new array of users.
	 */

	for( i = 0; i < numUsers; i++ )
	{
		if( !PGPdiskEngineIsUserOnDisk( disk->ioDisk, users[i] ) )
		{
			err = PGPdiskEngineAddUser( diskcon, disk->ioDisk, adminUser, adminPassphrase, users[i] );
			CKERR;
		}
	}
	
	err = PGPdiskEngineCountUsers( disk->ioDisk, &numUsersOnDisk ); CKERR;
	for( i = 0; i < numUsersOnDisk; i++ )
	{
		bUserIsOK = FALSE;

		err = sNewBEUser( diskcon, &user ); CKERR;
		err = PGPdiskEngineGetIODiskUser( diskcon, disk->ioDisk, i, user ); CKERR;

		for( j = 0; j < numUsers; j++ )
		{
			if( users[j]->ioDiskUser == user->ioDiskUser )
				bUserIsOK = TRUE;
		}
		
		if( !bUserIsOK )
		{
			err = PGPdiskEngineRemoveUser( disk->ioDisk, user->ioDiskUser ); CKERR;
			/* User i was just deleted, so the next user is going to be user i again */
			i--;
			numUsersOnDisk--;
		}
		(void) sFreeBEUser( user );
	}
	err = PGPdiskEngineSetAdminUser( disk->ioDisk, newAdminUser->ioDiskUser ); CKERR;

done:
	return err;
}

	static PGPError
sDispatchMessage(
	PGPdiskBEContextRef	diskcon,
	PGPByte *			buf,
	PGPSize				len )
{
	PGPError			err					= kPGPError_NoErr;
	PGPrpcMessage *		msg					= NULL;
	char *				pszPath				= NULL;
	char *				pszNewPath			= NULL;
	char *				pszRoot				= NULL;
	PGPBoolean			bReadOnly			= FALSE;
	PGPBoolean			bIsLocked			= FALSE;
	PGPUInt64			ullSize				= 0;
	char *				pszMountPoint		= NULL;
	PGPdiskEncryptionAlgorithm	alg;
	PGPBoolean			bPubKeyUser			= FALSE;
	PGPByte *			pKeyIDBuf			= NULL;
	PGPUInt32			uKeyIDSize			= 0;
	char *				pszUsername			= NULL;
	char *				pszPassphrase		= NULL;
	char *				pszNewPassphrase	= NULL;
	PGPdiskBEDiskRef	disk				= kInvalidPGPdiskBEDiskRef;
	PGPdiskBEUserRef	user				= kInvalidPGPdiskBEUserRef;
	PGPdiskBEUserRef *	users				= NULL;
	PGPUInt32			size				= 0;
	PGPUInt32			numUsers			= 0;
	PGPUInt32			prop				= 0;
	char *				pszProperty			= NULL;
	PGPBoolean			bProperty			= FALSE;
	PGPInt32			iProperty			= 0;
	PGPUInt32			uTimeout			= 0;
	char *				pszPublicFilename	= NULL;
	char *				pszPrivateFilename	= NULL;
	PGPBoolean			bForcedUnmount		= FALSE;
	PGPBoolean			bEjectOnLogout		= FALSE;
	PGPBoolean			bCachePassphrases	= FALSE;
	PGPBoolean			bShareCache			= FALSE;
	PGPKeyID			keyid;
	char *				pszCallbackSockPath	= NULL;
	PGPdiskDiskFormat	format				= kPGPdiskDiskFormat_Invalid;
	char *				pszVolumeName		= NULL;
	PGPdiskFilesystemType	fsType;
	PGPBoolean			updated				= FALSE;
	
	/*
	 * We don't get sdk notifications anymore since we moved our call to PGPsdkInit
	 * into a different thread from the one we have our CFRunLoop in, so here's the
	 * ugly polling loop to watch for KeyDB updates:
	 */
	if( diskcon->keydb )
		(void) PGPKeyDBIsUpdated( diskcon->keydb, &updated );
	
	msg = (PGPrpcMessage *) buf;
	syslog( LOG_DEBUG, "Got Message %d\n", msg->msgType );
	switch( msg->msgType )
	{
		/* PGPdiskContextRef messages */
		case kPGPdiskEngineMsg_LoadKeyrings:
			err = PGPrpcUnpack( diskcon->context, buf, len, "ss", &pszPublicFilename, &pszPrivateFilename ); CKERR;
			err = sLoadKeyrings( diskcon, pszPublicFilename, pszPrivateFilename ); /* CKERR; */
			err = PGPrpcPackAndWrite( diskcon->context, diskcon->sock, msg->msgType, "u", err ); CKERR;
			break;
			
		case kPGPdiskEngineMsg_SetADK:
			err = PGPrpcUnpack( diskcon->context, buf, len, "B", &pKeyIDBuf, &uKeyIDSize ); CKERR;
			pgpCopyMemory( pKeyIDBuf, &keyid, pgpMin( uKeyIDSize, sizeof( PGPKeyID ) ) );
			err = sSetADK( diskcon, &keyid ); /* CKERR; */
			err = PGPrpcPackAndWrite( diskcon->context, diskcon->sock, msg->msgType, "u", err ); CKERR;
			break;
			
		case kPGPdiskEngineMsg_SetGlobalDiskTimeout:
			err = PGPrpcUnpack( diskcon->context, buf, len, "u", &uTimeout ); CKERR;
			err = sSetGlobalDiskTimeout( diskcon, uTimeout ); /* CKERR; */
			err = PGPrpcPackAndWrite( diskcon->context, diskcon->sock, msg->msgType, "u", err ); CKERR;
			break;
		case kPGPdiskEngineMsg_SetGlobalEjectOnLogout:
			err = PGPrpcUnpack( diskcon->context, buf, len, "b", &bEjectOnLogout ); CKERR;
			err = sSetEjectOnLogout( diskcon, bEjectOnLogout ); /* CKERR; */
			err = PGPrpcPackAndWrite( diskcon->context, diskcon->sock, msg->msgType, "u", err ); CKERR;
			break;
		
		case kPGPdiskEngineMsg_SetPassphraseCaching:
			err = PGPrpcUnpack( diskcon->context, buf, len, "bub", &bCachePassphrases, &uTimeout, &bShareCache ); CKERR;
			err = sSetPassphraseCaching( diskcon, bCachePassphrases, uTimeout, bShareCache ); /* CKERR; */
			err = PGPrpcPackAndWrite( diskcon->context, diskcon->sock, msg->msgType, "u", err ); CKERR;
			break;
			
		case kPGPdiskEngineMsg_SetUnmountOnSleep:
			err = PGPrpcUnpack( diskcon->context, buf, len, "b", &bProperty ); CKERR;
			err = sSetUnmountOnSleep( diskcon, bProperty ); /* CKERR; */
			err = PGPrpcPackAndWrite( diskcon->context, diskcon->sock, msg->msgType, "u", err ); CKERR;
			break;

		case kPGPdiskEngineMsg_SetForcedAutoDiskUnmount:
			err = PGPrpcUnpack( diskcon->context, buf, len, "b", &bProperty ); CKERR;
			err = sSetForcedAutoDiskUnmount( diskcon, bProperty ); /* CKERR; */
			err = PGPrpcPackAndWrite( diskcon->context, diskcon->sock, msg->msgType, "u", err ); CKERR;
			break;
			
		case kPGPdiskEngineMsg_SetPreventSleep:
			err = PGPrpcUnpack( diskcon->context, buf, len, "b", &bProperty ); CKERR;
			err = sSetPreventSleep( diskcon, bProperty ); /* CKERR; */
			err = PGPrpcPackAndWrite( diskcon->context, diskcon->sock, msg->msgType, "u", err ); CKERR;
			break;
			
		case kPGPdiskEngineMsg_GetEnginePID:
			err = PGPrpcPackAndWrite( diskcon->context, diskcon->sock, msg->msgType, "iu", (PGPInt32) getpid(), err ); CKERR;
			break;
			
		/* PGPdiskUserRef messages */
		
        case kPGPdiskEngineMsg_NewUserWithSymmetricKey:
  			err = PGPrpcUnpack( diskcon->context, buf, len, "ssb",
				&pszUsername, &pszPassphrase, &bReadOnly ); CKERR;
			err = sNewUserWithSym( diskcon, pszUsername, pszPassphrase, bReadOnly, &user ); /* CKERR; */
			err = PGPrpcPackAndWrite( diskcon->context, diskcon->sock, msg->msgType,
				"uu", user, err ); CKERR;
			break;
		  
        case kPGPdiskEngineMsg_NewUserWithPublicKey:
			err = PGPrpcUnpack( diskcon->context, buf, len, "Bbb",
				&pKeyIDBuf, &uKeyIDSize, &bReadOnly, &bIsLocked ); CKERR;
			pgpCopyMemory( pKeyIDBuf, &keyid, pgpMin( uKeyIDSize, sizeof( PGPKeyID ) ) );
			err = sNewUserWithPub( diskcon, &keyid, bReadOnly, bIsLocked, &user ); /* CKERR; */
			err = PGPrpcPackAndWrite( diskcon->context, diskcon->sock, msg->msgType,
				"uu", user, err ); CKERR;
			break;
		 
        case kPGPdiskEngineMsg_UserGetPGPKeyID:
  			err = PGPrpcUnpack( diskcon->context, buf, len, "u", &user ); CKERR;
			err = sGetUserPGPKeyID( diskcon, user, &keyid ); /* CKERR; */
			err = PGPrpcPackAndWrite( diskcon->context, diskcon->sock, msg->msgType,
				"Bu", &keyid, sizeof( PGPKeyID ), err ); CKERR;
            break;
			
		case kPGPdiskEngineMsg_IsUserPassphraseCorrect:
  			err = PGPrpcUnpack( diskcon->context, buf, len,
				"us", &user, &pszPassphrase ); CKERR;
			err = sIsUserPassphraseCorrect( diskcon, user, pszPassphrase, &bProperty ); /* CKERR; */
			err = PGPrpcPackAndWrite( diskcon->context, diskcon->sock, msg->msgType,
				"ub", err, bProperty ); CKERR;
			break;
			
        case kPGPdiskEngineMsg_UserChangePassphrase:
  			err = PGPrpcUnpack( diskcon->context, buf, len,
				"uss", &user, &pszPassphrase, &pszNewPassphrase ); CKERR;
			err = sUserChangePassphrase( diskcon, user, pszPassphrase, pszNewPassphrase ); /* CKERR; */
			err = PGPrpcPackAndWrite( diskcon->context, diskcon->sock, msg->msgType,
				"u", err ); CKERR;
            break;
			
        case kPGPdiskEngineMsg_GetUserBooleanProp:
  			err = PGPrpcUnpack( diskcon->context, buf, len, "uu", &user, &prop ); CKERR;
			err = sGetUserBooleanProp( diskcon, user, prop, &bProperty ); /* CKERR; */
			err = PGPrpcPackAndWrite( diskcon->context, diskcon->sock, msg->msgType,
				"bu", bProperty, err ); CKERR;
            break;
			
        case kPGPdiskEngineMsg_GetUserNumericProp:
            break;
			
        case kPGPdiskEngineMsg_GetUserDataProp:
  			err = PGPrpcUnpack( diskcon->context, buf, len, "uu", &user, &prop ); CKERR;
			err = sGetUserDataProp( diskcon, user, prop, &pszProperty ); /* CKERR; */
			err = PGPrpcPackAndWrite( diskcon->context, diskcon->sock, msg->msgType,
				"su", pszProperty, err ); CKERR;
          break;
		  
        case kPGPdiskEngineMsg_SetUserBooleanProp:
  			err = PGPrpcUnpack( diskcon->context, buf, len, "uub", &user, &prop, &bProperty ); CKERR;
			err = sSetUserBooleanProp( diskcon, user, prop, bProperty ); /* CKERR; */
			err = PGPrpcPackAndWrite( diskcon->context, diskcon->sock, msg->msgType,
				"u", err ); CKERR;
            break;
			
        case kPGPdiskEngineMsg_SetUserNumericProp:
            break;
			
        case kPGPdiskEngineMsg_SetUserDataProp:
   			err = PGPrpcUnpack( diskcon->context, buf, len, "uus", &user, &prop, &pszProperty ); CKERR;
			err = sSetUserDataProp( diskcon, user, prop, pszProperty ); /* CKERR; */
			err = PGPrpcPackAndWrite( diskcon->context, diskcon->sock, msg->msgType,
				"u", err ); CKERR;
           break;

		/* PGPdiskDiskRef messages */
		
        case kPGPdiskEngineMsg_OpenDisk:
			err = PGPrpcUnpack( diskcon->context, buf, len, "su", &pszPath, &bReadOnly ); CKERR;
			err = sOpenDisk( diskcon, pszPath, bReadOnly, &disk ); /* CKERR; */
			err = PGPrpcPackAndWrite( diskcon->context, diskcon->sock, msg->msgType,
				"uu", disk, err ); CKERR;
			break;
			
        case kPGPdiskEngineMsg_CreateDisk:
			err = PGPrpcUnpack( diskcon->context, buf, len, "sUsubBsss",
				&pszPath,
				&ullSize,
				&pszMountPoint,
				&alg,
				&bPubKeyUser,
				&pKeyIDBuf, &uKeyIDSize,
				&pszUsername,
				&pszPassphrase,
				&pszCallbackSockPath ); CKERR;
			err = sCreateDisk( diskcon,
				pszPath,
				ullSize, 
				pszMountPoint,
				alg,
				bPubKeyUser,
				(PGPKeyID *) pKeyIDBuf,
				pszUsername,
				pszPassphrase,
				pszCallbackSockPath,
				&disk ); /* CKERR; */
			err = PGPrpcPackAndWrite( diskcon->context, diskcon->sock, msg->msgType,
				"uu", disk, err ); CKERR;
			break;
			
        case kPGPdiskEngineMsg_CloseDisk:
			err = PGPrpcUnpack( diskcon->context, buf, len, "u", &disk ); CKERR;
			err = sCloseDisk( diskcon, disk ); /* CKERR; */
			err = PGPrpcPackAndWrite( diskcon->context, diskcon->sock, msg->msgType,
				"u", err ); CKERR;
            break;
		
		case kPGPdiskEngineMsg_FormatDisk:
			err = PGPrpcUnpack( diskcon->context, buf, len, "ussu", &disk, &pszPassphrase, &pszVolumeName, &fsType ); CKERR;
			err = sFormatDisk( diskcon, disk, pszPassphrase, pszVolumeName, fsType ); /* CKERR; */
			err = PGPrpcPackAndWrite( diskcon->context, diskcon->sock, msg->msgType,
				"u", err ); CKERR;
			break;
		
		case kPGPdiskEngineMsg_CanPassphraseMountDisk:
			err = PGPrpcUnpack( diskcon->context, buf, len,
				"us", &disk, &pszPassphrase ); CKERR;
			err = sCanPassphraseMountDisk( diskcon, disk, pszPassphrase, &bProperty ); /* CKERR; */
			err = PGPrpcPackAndWrite( diskcon->context, diskcon->sock, msg->msgType,
				"bu", bProperty, err ); CKERR;
			break;
			
        case kPGPdiskEngineMsg_MountDisk:
			err = PGPrpcUnpack( diskcon->context, buf, len,
				"ususb", &disk, &pszRoot, &user, &pszPassphrase, &bReadOnly ); CKERR;
			err = sMountDisk( diskcon, disk, pszRoot, user, pszPassphrase, bReadOnly ); /* CKERR; */
			err = PGPrpcPackAndWrite( diskcon->context, diskcon->sock, msg->msgType,
				"u", err ); CKERR;
            break;
			
        case kPGPdiskEngineMsg_UnmountDisk:
 			err = PGPrpcUnpack( diskcon->context, buf, len, "ub", &disk, &bForcedUnmount ); CKERR;
			err = sUnmountDisk( diskcon, disk, bForcedUnmount ); /* CKERR; */
			err = PGPrpcPackAndWrite( diskcon->context, diskcon->sock, msg->msgType, "u", err ); CKERR;
           break;
			
        case kPGPdiskEngineMsg_DiskExportUserSet:
			err = PGPrpcUnpack( diskcon->context, buf, len, "u", &disk ); CKERR;
			err = sDiskExportUserSet( diskcon, disk, &users, &numUsers ); /* CKERR; */
			err = PGPrpcPackAndWrite( diskcon->context, diskcon->sock, msg->msgType,
				"uB", err, users, numUsers * sizeof( PGPdiskBEUserRef ) ); CKERR;
            break;
			
        case kPGPdiskEngineMsg_DiskImportUserSet:
			err = PGPrpcUnpack( diskcon->context, buf, len, "usBu",
				&disk, &pszPassphrase, &users, &size, &user ); CKERR;
			err = sDiskImportUserSet( diskcon, disk, pszPassphrase, users, size / sizeof( PGPUInt32 ), user ); /* CKERR */
			err = PGPrpcPackAndWrite( diskcon->context, diskcon->sock, msg->msgType,
				"u", err ); CKERR;
            break;
			
        case kPGPdiskEngineMsg_ReEncryptDisk:
			err = PGPrpcUnpack( diskcon->context, buf, len, "usus", &disk, &pszPassphrase, &alg, &pszCallbackSockPath ); CKERR;
			err = sReEncryptDisk( diskcon, disk, pszPassphrase, alg, pszCallbackSockPath ); /* CKERR; */
			err = PGPrpcPackAndWrite( diskcon->context, diskcon->sock, msg->msgType,
				"u", err ); CKERR;
            break;

		case kPGPdiskEngineMsg_ConvertDisk:
			err = PGPrpcUnpack( diskcon->context, buf, len,
				"ussss", &format, &pszPath, &pszPassphrase, &pszNewPath, &pszCallbackSockPath ); CKERR;
			err = sConvertDisk( diskcon, format, pszPath, pszPassphrase, pszNewPath, pszCallbackSockPath ); /* CKERR; */
			err = PGPrpcPackAndWrite( diskcon->context, diskcon->sock, msg->msgType, "u", err ); CKERR;
			break;
			
        case kPGPdiskEngineMsg_GetDiskBooleanProp:
   			err = PGPrpcUnpack( diskcon->context, buf, len, "uu", &disk, &prop ); CKERR;
			err = sGetDiskBooleanProp( diskcon, disk, prop, &bProperty ); /* CKERR; */
			err = PGPrpcPackAndWrite( diskcon->context, diskcon->sock, msg->msgType,
				"bu", bProperty, err ); CKERR;
			break;
			
        case kPGPdiskEngineMsg_GetDiskNumericProp:
   			err = PGPrpcUnpack( diskcon->context, buf, len, "uu", &disk, &prop ); CKERR;
			err = sGetDiskNumericProp( diskcon, disk, prop, &iProperty ); /* CKERR; */
			err = PGPrpcPackAndWrite( diskcon->context, diskcon->sock, msg->msgType,
				"iu", iProperty, err ); CKERR;
			break;
			
        case kPGPdiskEngineMsg_GetDiskDataProp:
   			err = PGPrpcUnpack( diskcon->context, buf, len, "uu", &disk, &prop ); CKERR;
			err = sGetDiskDataProp( diskcon, disk, prop, &pszProperty ); /* CKERR; */
			err = PGPrpcPackAndWrite( diskcon->context, diskcon->sock, msg->msgType,
				"su", pszProperty, err ); CKERR;
            break;
			
        case kPGPdiskEngineMsg_SetDiskBooleanProp:
   			err = PGPrpcUnpack( diskcon->context, buf, len, "uub", &disk, &prop, &bProperty ); CKERR;
			err = sSetDiskBooleanProp( diskcon, disk, prop, bProperty ); /* CKERR; */
			err = PGPrpcPackAndWrite( diskcon->context, diskcon->sock, msg->msgType,
				"u", err ); CKERR;
           break;
		   
        case kPGPdiskEngineMsg_SetDiskNumericProp:
   			err = PGPrpcUnpack( diskcon->context, buf, len, "uui", &disk, &prop, &iProperty ); CKERR;
			err = sSetDiskNumericProp( diskcon, disk, prop, iProperty ); /* CKERR; */
			err = PGPrpcPackAndWrite( diskcon->context, diskcon->sock, msg->msgType,
				"u", err ); CKERR;
            break;
			
        case kPGPdiskEngineMsg_SetDiskDataProp:
   			err = PGPrpcUnpack( diskcon->context, buf, len, "uus", &disk, &prop, &pszProperty ); CKERR;
			err = sSetDiskDataProp( diskcon, disk, prop, pszProperty ); /* CKERR; */
			err = PGPrpcPackAndWrite( diskcon->context, diskcon->sock, msg->msgType,
				"u", err ); CKERR;
            break;
			
		default:
			syslog( LOG_WARNING, "Invalid RPC message received (%d).  Ignoring.\n", msg->msgType );
			break;
	}
	
done:
	if( IsntNull( pszVolumeName ) )
		(void) PGPFreeData( pszVolumeName );
	if( IsntNull( pszProperty ) )
		(void) PGPFreeData( pszProperty );
	if( IsntNull( pszCallbackSockPath ) )
		(void) PGPFreeData( pszCallbackSockPath );
	if( IsntNull( pszPublicFilename ) )
		(void) PGPFreeData( pszPublicFilename );
	if( IsntNull( pszPrivateFilename ) )
		(void) PGPFreeData( pszPrivateFilename );
	if( IsntNull( pszPath ) )
		(void) PGPFreeData( pszPath );
	if( IsntNull( pszNewPath ) )
		(void) PGPFreeData( pszNewPath );
	if( IsntNull( pszRoot ) )
		(void) PGPFreeData( pszRoot );
	if( IsntNull( pszMountPoint ) )
		(void) PGPFreeData( pszMountPoint );
	if( IsntNull( pszUsername ) )
		(void) PGPFreeData( pszUsername );
	if( IsntNull( pszPassphrase ) )
		(void) PGPFreeData( pszPassphrase );
	if( IsntNull( pszNewPassphrase ) )
		(void) PGPFreeData( pszNewPassphrase );
	if( IsntNull( pKeyIDBuf ) )
		(void) PGPFreeData( pKeyIDBuf );
	if( IsntNull( users ) )
		(void) PGPFreeData( users );
	return err;
}

	static void *
sCreateNewWorkerThread(
	void *				arg )
{
	PGPError			err			= kPGPError_NoErr;
	PGPInt32			retval		= 0;
	fd_set				sockSet;
	PGPByte *			buf			= NULL;
	PGPSize				len			= 0;
	PGPContextRef		context		= kInvalidPGPContextRef;
	PGPdiskBEContextRef	diskcon		= kInvalidPGPdiskBEContextRef;
	PGPrpcMessage *		msg			= NULL;
	PGPdiskAppWorkerThreadData *	awtd	= arg;

	context = awtd->context;
	err = sNewBEContext( context, &diskcon ); CKERR;

	diskcon->mgr = awtd->mgr;
	diskcon->sock = awtd->asock;
	diskcon->kernCon = awtd->kernCon;
	
	FD_ZERO( &sockSet );
	FD_SET( diskcon->sock, &sockSet );

	syslog( LOG_DEBUG, "Waiting for client messages...\n" );
	while( ( retval = select( diskcon->sock + 1, &sockSet, NULL, NULL, NULL ) ) > 0 )
	{
		if( IsntNull( buf ) )
			(void) PGPFreeData( buf );
		buf = NULL;
		len = 0;

		/*
		 * If we get an error from PGPrpcRead, there was some problem
		 * reading from the socket, like the app died.  There's really no
		 * graceful way to handle it, so we'll cleanup and exit.
		 */
		err = PGPrpcRead( context, diskcon->sock, &buf, &len ); CKERR;

		msg = (PGPrpcMessage *) buf;
		if( msg->msgType == kPGPdiskEngineMsg_FreeContext )
			break;
		
		/* Do message operation */
		err = sDispatchMessage( diskcon, buf, len );
		if( IsPGPError( err ) )
			continue;
	}
	
done:
	pthread_mutex_lock( awtd->numConnectionsMutex );
	(*awtd->pNumConnections)--;
	pthread_mutex_unlock( awtd->numConnectionsMutex );

	if( IsntNull( buf ) )
		(void) PGPFreeData( buf );
	if( IsntNull( diskcon ) )
		(void) sFreeBEContext( diskcon );
	if( IsntNull( awtd ) )
		(void) PGPFreeData( awtd );
	if( PGPContextRefIsValid( context ) )
	{
		(void) PGPFreeContext( context );
		(void) PGPsdkCleanup();
	}
		
	syslog( LOG_DEBUG, "Thread exiting\n" );
	return NULL;
}

	static PGPError
sGetAppMessages(
	PGPMemoryMgrRef					mgr,
	PGPInt32						sock,
	PGPdiskKernelWorkerContextRef	kernCon )
{
	PGPError		err				= kPGPError_NoErr;
	PGPContextRef	context			= kInvalidPGPContextRef;
	fd_set			sockSet;
	PGPByte *		buf				= NULL;
	PGPSize			len				= 0;
	PGPInt32		asock			= 0;
	PGPInt32		retval			= 0;
	struct timeval	tv;
	PGPUInt32		numConnections	= 0;
	pthread_mutex_t	mutex;
	pthread_t		thread;
	PGPrpcMessage *	msg				= NULL;
	PGPdiskAppWorkerThreadData *	awtd	= NULL;

	pthread_mutex_init( &mutex, 0 );

	while( gbRunning )
	{
		FD_ZERO( &sockSet );
		FD_SET( sock, &sockSet );

#if PGP_DEBUG
		/* A really long timeout to help with debugging */
		tv.tv_sec = 60;
#else
		/* A shorter timeout to prevent diskengine from hanging around for too long after the app quits */
		tv.tv_sec = 10;
#endif
		tv.tv_usec = 0;

		retval = select( sock + 1, &sockSet, NULL, NULL, &tv );
		if( retval == 0 )
		{
			/* Timeout */
			pthread_mutex_lock( &mutex );
			if( ( numConnections == 0 ) && ( PGPdiskEngineNumMountedDisks() == 0 ) )
				gbRunning = FALSE;
			pthread_mutex_unlock( &mutex );
			continue;
		}
		else if( retval < 0 )
		{
			if( errno == EINTR )
				continue;
			else
				break;
		}
		
		/* Message is available */
		
		if( IsntNull( buf ) )
			(void) PGPFreeData( buf );
		buf = NULL;
		len = 0;
		
		/*
		 * We only handle one kind of message here: the NewContext message.
		 * This means that someone has just started a copy of the app and
		 * had called PGPdiskNewContext.  This is only called once in the
		 * client.
		 */
		 
		err = sAcceptSocket( sock, &asock );
		if( IsPGPError( err ) )
		{
			/* Client connecting, but not sending any data - ignore this connection */
			continue;
		}
		 
		/*
		 * We're keeping the sdk un-inited so that I don't have any connections
		 * to the sdk service, which allows me to survive a user logging out.
		 */
		err = PGPsdkInit( 0 ); CKERR;
		err = PGPNewContext( kPGPsdkAPIVersion, &context ); CKERR;
		
		err = sGetNewContextMessage( context, asock, &buf, &len );
		if( IsPGPError( err ) )
		{
			err = kPGPError_NoErr;
			if( PGPContextRefIsValid( context ) )
			{
				(void) PGPFreeContext( context );
				(void) PGPsdkCleanup();
			}
			continue;
		}
		else
		{
			msg = (PGPrpcMessage *) buf;
			if( msg->msgType == kPGPdiskEngineMsg_NewContext )
			{
				/* Perform some self checks */
				if( VerifyCastSBoxesValid() )
				{
					err = PGPrpcPackAndWrite( context, asock, kPGPdiskEngineMsg_NewContext,
						"u", kPGPError_NoErr ); CKERR;
				}
				else
				{
					err = PGPrpcPackAndWrite( context, asock, kPGPdiskEngineMsg_NewContext,
						"u", kPGPError_SelfTestFailed ); CKERR;
				}
				
				if( IsntNull( buf ) )
				{
					(void) PGPFreeData( buf );
					buf = NULL;
					len = 0;
					msg = NULL;
				}

				/* Create worker thread */
				awtd = PGPNewData( PGPPeekContextMemoryMgr( context ),
					sizeof( PGPdiskAppWorkerThreadData ),
					kPGPMemoryMgrFlags_Clear );
				CKNULL( awtd );

				awtd->context = context;
				awtd->mgr = mgr;
				awtd->asock = asock;
				awtd->kernCon = kernCon;

				awtd->pNumConnections = &numConnections;
				awtd->numConnectionsMutex = &mutex;
				
				pthread_create( &thread, NULL, sCreateNewWorkerThread, (void *) awtd );
				pthread_detach( thread );
				
				context = kInvalidPGPContextRef;
				
				pthread_mutex_lock( &mutex );
				numConnections++;
				pthread_mutex_unlock( &mutex );
			}
		}
	}
	
	/* Nicely unmount all mounted disks */
	(void) PGPdiskEngineUnmountAllDisks( FALSE );
	pthread_mutex_destroy( &mutex );
	
done:
	syslog( LOG_NOTICE, "Shutting down...\n" );

	if( PGPContextRefIsValid( context ) )
	{
		(void) PGPFreeContext( context );
		(void) PGPsdkCleanup();
	}

	if( IsntNull( buf ) )
		(void) PGPFreeData( buf );

	return err;
}

	void *
sAppWorkerThread(
	void *						arg )
{
	PGPError					err			= kPGPError_NoErr;
	PGPInt32					sock		= 0;
	char						sockPath[PATH_MAX];
	PGPdiskAppWorkerContext *	con			= (PGPdiskAppWorkerContext *) arg;
	
	syslog( LOG_DEBUG, "Starting App worker\n" );
	
	/* Create IPC socket */
	err = sCreateIPCSocket( sockPath, &sock ); CKERR;
	
	/* Loop, reading messages from App */
	(void) sGetAppMessages( con->mgr, sock, con->kernCon );
		
	err = sDestroyIPCSocket( sock, sockPath ); CKERR;
	
done:
	if( IsntNull( arg ) )
		(void) PGPFreeData( arg );
	return NULL;
}

	PGPError
PGPdiskEngineStartAppWorker(
	PGPMemoryMgrRef					mgr,
	PGPdiskKernelWorkerContextRef	kernCon )
{
	PGPError					err			= kPGPError_NoErr;
	PGPdiskAppWorkerContext *	con			= NULL;
	pthread_t					thread;

	con = PGPNewData( mgr,
		sizeof( PGPdiskAppWorkerContext ),
		kPGPMemoryMgrFlags_Clear );
	CKNULL( con );
	
	con->mgr = mgr;
	con->kernCon = kernCon;

	/* Start the thread that sits around waiting for the computer to go to sleep */
	pthread_create( &thread, NULL, sAppWorkerThread, con );
	pthread_detach( thread );
	
done:
	if( IsPGPError( err ) )
		(void) PGPFreeData( con );

	return err;
}

/*__Editor_settings____

	Local Variables:
	tab-width: 4
	End:
	vi: ts=4 sw=4
	vim: si
_____________________*/
