/*____________________________________________________________________________
	Copyright (C) 2002 PGP Corporation
	All rights reserved.

	$Id: pgpDiskEngineKernelWorker.c,v 1.50 2002/11/13 21:01:58 dallen Exp $
____________________________________________________________________________*/

#include "pgpErrors.h"
#include "pgpMem.h"
#include "pgpUtilities.h"
#include "pgpDiskEngine.h"
#include "pgpRPCPacker.h"

#include <errno.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <syslog.h>
#include <sys/param.h>
#include <sys/ucred.h>
#include <sys/mount.h>
#include <sys/uio.h>
#include <unistd.h>
#include <fcntl.h>
#include <hfs/hfs_mount.h>

#include <mach/mach_error.h>

#include <CoreFoundation/CoreFoundation.h>
#include <Carbon/Carbon.h>
#include <IOKit/IOKitLib.h>
#include <IOKit/iokitmig.h>
#include <IOKit/pwr_mgt/IOPMLib.h>
#include <IOKit/IOMessage.h>

#include "PGPdiskControllerUserClient.h"
#include "PGPdiskDriverUserClient.h"
#include "pgpDiskEngineKernelWorker.h"
#include "pgpDiskEngineAppWorker.h"
#include "pgpDiskEngineDisk.h"
#include "pgpDiskGetPath.h"

/* MACROS */
#define CKERR				if( IsPGPError( err ) ) goto done
#define CKNULL(x)			do { if( IsNull( x ) ) { err = kPGPError_OutOfMemory; goto done; } } while(0)
#define RETERR(x)			do { err = x; goto done; } while(0)

#define CKMACHERR(x,msg)	do { if( (x) != KERN_SUCCESS ) { \
								syslog( LOG_ERR, "Error: %s: %s\n", (msg), mach_error_string(x) ); \
								RETERR( kPGPError_UnknownError ); } } while(0)

extern PGPBoolean		gbRunning;

/* STRUCTS */
typedef struct PGPdiskEngineKernelWorkerThreadData
{
	PGPMemoryMgrRef					mgr;
	PGPdiskKernelWorkerContextRef	kernCon;

	PGPdiskIODiskRef				ioDisk;
	PGPUInt32						diskid;
	PGPBoolean						readonly;
	PGPBoolean						ejectOnLogout;
	PGPBoolean						autoDiskMount;
} PGPdiskEngineKernelWorkerThreadData;


/* FUNCTIONS */
	static io_service_t
sGetDevice(
	mach_port_t			masterPort,
	char *				className )
{
	PGPError			err			= kPGPError_NoErr;
	io_service_t		device		= 0;
	io_service_t		foundDevice	= 0;
	io_iterator_t		iter		= 0;
	CFDictionaryRef		dict		= NULL;
	kern_return_t		kr			= 0;
	
	dict = IOServiceMatching( className );
	if( dict == 0 )
	{
		syslog( LOG_ERR, "Error: Couldn't create a %s matching dictionary\n", className );
		RETERR( kPGPError_UnknownError );
	}
	
	kr = IOServiceGetMatchingServices( masterPort, dict, &iter );
	CKMACHERR( kr, "Couldn't create iterator" );
	
	/* Find the last device in the iterator */
	while( ( foundDevice = IOIteratorNext( iter ) ) )
		device = foundDevice;

done:
	if( iter )
		IOObjectRelease( iter );

	if( IsPGPError( err ) )
		device = 0;
	return device;
}

	static PGPError
sLoadDriver( void )
{
	PGPError	err			= kPGPError_NoErr;
	pid_t		pid;
	PGPInt32	status		= 0;
	FILE *		file		= NULL;
	char		currentPath[PATH_MAX];
	char		load_pgpdiskPath[PATH_MAX];

	err = PGPdiskGetCurrentExecutablePath( currentPath ); CKERR;
	
	snprintf( load_pgpdiskPath, sizeof( load_pgpdiskPath ), "%s/load_pgpdisk", currentPath );
	
	pid = fork();
	if( pid == 0 )
	{
		execl( load_pgpdiskPath, "load_pgpdisk", NULL );
			
		/* We'll only get here if there was an error running load_pgpdisk */
		exit( EXIT_FAILURE );
	}
	else
		waitpid( pid, &status, 0 );
		
	if( WEXITSTATUS( status ) != EXIT_SUCCESS )
		return kPGPdiskError_NotInstalledCorrectly;
done:
	if( IsntNull( file ) )
		(void) fclose( file );
	return err;
}

	static PGPError
sInitializeController(
	PGPdiskKernelWorkerContextRef	con )
{
	PGPError						err			= kPGPError_NoErr;
	kern_return_t					kr			= 0;
	
	if( !con->controllerDevice )
	{
		/* Initialize controller */
		kr = IOMasterPort( MACH_PORT_NULL, &con->masterPort );
		CKMACHERR( kr, "Couldn't create a master IOKit Port" );

		con->controllerDevice = sGetDevice( con->masterPort, kPGPdiskController_ClassName );
		if( !con->controllerDevice )
			RETERR( kPGPError_UnknownError );
	}

done:
	return err;
}

	static void
sComputerSleepCallback(
	void *			refcon,
	io_service_t	service,
	natural_t		messageType,
	void *			messageArgument )
{
	PGPError						err			= kPGPError_NoErr;
	PGPdiskKernelWorkerContextRef	kernCon		= (PGPdiskKernelWorkerContextRef) refcon;
	
	switch( messageType )
	{
		case kIOMessageSystemWillSleep:		/* System going to sleep NOW! */
		case kIOMessageCanSystemSleep:		/* System will go to sleep soon */
			if( kernCon->unmountOnSleep )
			{
				err = PGPdiskEngineUnmountAllDisks( kernCon->forcedUnmount );

				if( IsPGPError( err ) && kernCon->preventSleep )
				{
					syslog( LOG_WARNING, "Couldn't unmount all mounted disks.  Preventing system sleep.\n" );
					IOCancelPowerChange( kernCon->port, (long) messageArgument );
				}
				else
					IOAllowPowerChange( kernCon->port, (long) messageArgument );
			}
			else
				IOAllowPowerChange( kernCon->port, (long) messageArgument );
			break;

		default:
			break;
	}
	return;
}

	PGPError
PGPdiskEngineRun(
	PGPdiskKernelWorkerContextRef	kernCon )
{
	PGPError						err			= kPGPError_NoErr;
	IONotificationPortRef			notify;
	io_object_t						iter;
	io_connect_t					port;

	/*
	 * We need to basically just sit around in our CFRunLoop, waiting
	 * for the message that says the system is going to sleep.
	 */
	 
	port = IORegisterForSystemPower( kernCon, &notify, sComputerSleepCallback, &iter );
	if( IsNull( port ) )
	{
		syslog( LOG_WARNING, "IORegisterForSystemPower failed.  Not unmounting disks on computer sleep\n" );
		return NULL;
	}
	
	kernCon->port = port;
	
	CFRunLoopAddSource( CFRunLoopGetCurrent(),
		IONotificationPortGetRunLoopSource( notify ),
		kCFRunLoopDefaultMode );
	
	while( gbRunning )
		CFRunLoopRunInMode( kCFRunLoopDefaultMode, 10, FALSE );

	return err;
}

	PGPError
PGPdiskEngineStartKernelWorker(
	PGPMemoryMgrRef					mgr,
	PGPdiskKernelWorkerContextRef *	outCon )
{
	PGPError						err			= kPGPError_NoErr;
	PGPdiskKernelWorkerContext *	con			= NULL;

	con = PGPNewData( mgr,
		sizeof( PGPdiskKernelWorkerContext ),
		kPGPMemoryMgrFlags_Clear );
	CKNULL( con );

	*outCon = con;

done:
	if( IsPGPError( err ) )
		(void) PGPFreeData( con );

	return err;
}

	PGPError
PGPdiskEngineLoadDriver( void )
{
	PGPError						err			= kPGPError_NoErr;
	mach_port_t						masterPort;
	io_object_t						controllerDevice;
	kern_return_t					kr			= 0;
	time_t							loadTime	= 0;
	
	/* Initialize controller */
	kr = IOMasterPort( MACH_PORT_NULL, &masterPort );
	CKMACHERR( kr, "Couldn't create a master IOKit Port" );

	controllerDevice = sGetDevice( masterPort, kPGPdiskController_ClassName );
	if( !controllerDevice )
	{
		/* Driver not installed.  Load it and try again */
		err = sLoadDriver(); CKERR;
		
		/* Loop for 10 seconds, waiting for the driver to load */
		loadTime = time( NULL );
		do
		{
			/* At least _try_ to be nice in our ugly polling loop */
			usleep( 250000 );

			controllerDevice = sGetDevice( masterPort, kPGPdiskController_ClassName );
		} while( !controllerDevice && ( time(NULL) < loadTime + 10 ) );

		if( !controllerDevice )
			RETERR( kPGPdiskError_NotInstalledCorrectly );
	}
	
done:
	return err;
}

	PGPError
PGPdiskEngineStopKernelWorker(
	PGPdiskKernelWorkerContextRef		con )
{
	if( IsNull( con ) )
		return kPGPError_BadParams;
		
	if( con->controllerDevice )
		IOObjectRelease( con->controllerDevice );
		
	(void) PGPFreeData( con );
	
	return kPGPError_NoErr;
}

	static PGPError
sGetBSDName(
	PGPdiskKernelWorkerContextRef	con,
	PGPdiskIODiskRef				ioDisk,
	PGPUInt32						diskid )
{
	PGPError						err				= kPGPError_NoErr;
	kern_return_t					kr				= 0;
	io_iterator_t					iter			= NULL;
	io_object_t     				obj         	= NULL;
	io_iterator_t					childIter		= NULL;
	io_object_t						childObj		= NULL;
	PGPUInt32						thisDiskID		= 0;
	CFTypeRef						cfDiskID		= NULL;
	char							szName[128];
	CFTypeRef						cfBSDName		= NULL;
	char *							pszBSDName		= NULL;

	/*
	 * Now we need to find the BSD name of this disk (like /dev/disk1).  We do this
	 * by iterating through the IORegistry, first looking for devices called
	 * "PGPdiskDriver".  There could be several of these, so we know we have the right
	 * one when the registry key "diskid" matches the disk ID we got back from
	 * ActivateDrive.  We then need to iterate down until we find the IOMedia object.
	 * This object has a string property "BSD Name" which contains the BSD name of
	 * the device.
	 */
	kr = IORegistryEntryCreateIterator( con->controllerDevice, kIOServicePlane, kIORegistryIterateRecursively, &iter );
	CKMACHERR( kr, "Couldn't create iterator" );

	while( ( obj = IOIteratorNext( iter ) ) )
	{
		if( IOObjectConformsTo( obj, "PGPdiskDriver" ) )
		{
			cfDiskID = IORegistryEntryCreateCFProperty( obj, CFSTR( "diskid" ), kCFAllocatorDefault, 0 );
			CFNumberGetValue( cfDiskID, kCFNumberIntType, &thisDiskID );
			if( diskid == thisDiskID )
			{
				kr = IORegistryEntryCreateIterator( obj, kIOServicePlane, kIORegistryIterateRecursively, &childIter );
				CKMACHERR( kr, "Couldn't create iterator" );
				
				while( ( childObj = IOIteratorNext( childIter ) ) )
				{
					IORegistryEntryGetName( childObj, szName );
					if( IOObjectConformsTo( childObj, "IOMedia" ) && !strcmp( szName, "PGP Corporation PGPdisk Media" ) )
					{
						cfBSDName = IORegistryEntryCreateCFProperty( childObj, CFSTR( "BSD Name" ), kCFAllocatorDefault, 0 );
						if( IsntNull( cfBSDName ) )
						{
							pszBSDName = (char *) CFStringGetCStringPtr( cfBSDName, kCFStringEncodingMacRoman );
							err = PGPdiskEngineSetDiskDeviceName( ioDisk, pszBSDName ); CKERR;
							CFRelease( cfBSDName );
						}
					}
				}
			}
		}
	}

done:
	return err;
}


	static UInt32
sGetRequest(
	io_connect_t			dataPort,
	void *					data,
	PGPdiskIOParamBlock *	ioParamsFromKernel )
{
	PGPError				err			= kPGPError_NoErr;
	kern_return_t			kr			= 0;
	PGPdiskIOParamBlock		ioParamsToKernel;
	mach_msg_type_number_t	len			= 0;
	
	ioParamsToKernel.controlCode	= kPGPdiskDriverControlCode_GetRequest;
	ioParamsToKernel.startBlock		= 0;
	ioParamsToKernel.numBlocks		= 0;
	ioParamsToKernel.refCon			= 0;
	ioParamsToKernel.err			= 0;
	ioParamsToKernel.data			= data;
	
	ioParamsFromKernel->controlCode	= 0;
	ioParamsFromKernel->startBlock	= 0;
	ioParamsFromKernel->numBlocks	= 0;
	ioParamsFromKernel->refCon		= 0;
	ioParamsFromKernel->err			= 0;
	ioParamsFromKernel->data		= data;

	/* Send GetRequest message and wait for response */
	len = sizeof( PGPdiskIOParamBlock );
	kr = io_connect_method_structureI_structureO( dataPort,
		kPGPdiskDriverUserClient_DiskControl,
		(void *) &ioParamsToKernel, sizeof( PGPdiskIOParamBlock ),
		(void *) ioParamsFromKernel, &len );
	CKMACHERR( kr, "Couldn't get request from PGPdisk driver" );

done:
	if( IsPGPError( err ) )
	{
		/* Problem talking to driver - get rid of this disk */
		return kPGPdiskDriverControlCode_Inactive;
	}

	return ioParamsFromKernel->controlCode;
}

	static PGPError
sDoReadRequest(
	PGPdiskIODiskRef		disk,
	PGPdiskIOParamBlock *	ioParamsFromKernel,
	PGPdiskIOParamBlock *	ioParamsToKernel )
{
	PGPError				err				= kPGPError_NoErr;
	
	ioParamsToKernel->controlCode	= kPGPdiskDriverControlCode_ReadReply;
	ioParamsToKernel->startBlock	= ioParamsFromKernel->startBlock;
	ioParamsToKernel->numBlocks		= ioParamsFromKernel->numBlocks;
	ioParamsToKernel->refCon		= ioParamsFromKernel->refCon;
	ioParamsToKernel->err			= 0;
	ioParamsToKernel->data			= ioParamsFromKernel->data;
	
	err = PGPdiskEngineReadFromDisk( disk,
		ioParamsFromKernel->startBlock,
		ioParamsFromKernel->numBlocks,
		ioParamsToKernel->data ); CKERR;
	
done:
	return err;
}

	static PGPError
sDoWriteRequest(
	PGPdiskIODiskRef		disk,
	PGPdiskIOParamBlock *	ioParamsFromKernel,
	PGPdiskIOParamBlock *	ioParamsToKernel )
{
	PGPError				err				= kPGPError_NoErr;
	
	ioParamsToKernel->controlCode	= kPGPdiskDriverControlCode_WriteReply;
	ioParamsToKernel->startBlock	= ioParamsFromKernel->startBlock;
	ioParamsToKernel->numBlocks		= ioParamsFromKernel->numBlocks;
	ioParamsToKernel->refCon		= ioParamsFromKernel->refCon;
	ioParamsToKernel->err			= 0;
	ioParamsToKernel->data			= ioParamsFromKernel->data;
	
	err = PGPdiskEngineWriteToDisk( disk,
		ioParamsFromKernel->startBlock,
		ioParamsFromKernel->numBlocks,
		ioParamsToKernel->data ); CKERR;
	
done:	
	return err;
}

	static UInt32
sDoRequest(
	UInt32					controlCode,
	io_connect_t			dataPort,
	void *					data,
	PGPdiskIODiskRef		disk,
	PGPdiskIOParamBlock *	ioParamsFromKernel )
{
	PGPError				err			= kPGPError_NoErr;
	PGPdiskIOParamBlock		ioParamsToKernel;
	mach_msg_type_number_t	len			= 0;
	kern_return_t			kr			= 0;
	
	ioParamsToKernel.controlCode	= kPGPdiskDriverControlCode_None;
	ioParamsToKernel.startBlock		= 0;
	ioParamsToKernel.numBlocks		= 0;
	ioParamsToKernel.refCon			= 0;
	ioParamsToKernel.err			= 0;
	ioParamsToKernel.data			= data;
	
	ioParamsFromKernel->data		= data;

	switch( controlCode )
	{
		case kPGPdiskDriverControlCode_ReadRequest:
			err = sDoReadRequest( disk, ioParamsFromKernel, &ioParamsToKernel ); CKERR;
			break;
		case kPGPdiskDriverControlCode_WriteRequest:
			err = sDoWriteRequest( disk, ioParamsFromKernel, &ioParamsToKernel ); CKERR;
			break;
		case kPGPdiskDriverControlCode_Inactive:
		case kPGPdiskDriverControlCode_None:
			return controlCode;
		default:
			syslog( LOG_WARNING, "Error: Invalid control code (%ld) received from driver\n", controlCode );
			return kPGPdiskDriverControlCode_None;
	}
	
	/* Send Reply message */
	len = sizeof( PGPdiskIOParamBlock );
	kr = io_connect_method_structureI_structureO( dataPort,
		kPGPdiskDriverUserClient_DiskControl,
		(void *) &ioParamsToKernel, sizeof( PGPdiskIOParamBlock ),
		(void *) ioParamsFromKernel, &len );
	CKMACHERR( kr, "Couldn't send reply to PGPdisk driver" );

done:
	if( IsPGPError( err ) )
	{
		(void) memset( ioParamsFromKernel, 0, sizeof( PGPdiskIOParamBlock ) );
		return kPGPdiskDriverControlCode_None;
	}

	return ioParamsFromKernel->controlCode;
}

	static void
sGetKernelMessages(
	PGPdiskIODiskRef						disk,
	PGPdiskEngineKernelWorkerThreadData *	kwtd,
	io_connect_t							dataPort )
{
	PGPError				err			= kPGPError_NoErr;
	PGPdiskIOParamBlock		ioParamsFromKernel;
	void *					data		= NULL;
	UInt32					controlCode	= 0;
	
	data = PGPNewData( kwtd->mgr,
		kPGPdiskDriver_MaxTransferSize,
		kPGPMemoryMgrFlags_Clear );
	CKNULL( data );
	
	while( controlCode != kPGPdiskDriverControlCode_Inactive )
	{
		/* We could get the control code from sGetRequest _OR_ from sDoRequest */

		if( controlCode == kPGPdiskDriverControlCode_None )
			controlCode = sGetRequest( dataPort, data, &ioParamsFromKernel );
	
		controlCode = sDoRequest( controlCode, dataPort, data, disk, &ioParamsFromKernel );
		
		if( disk->szDeviceName[0] == '\0' )
		{
			/* We don't have the BSD Name of the disk yet.  Try to get it. */
			(void) sGetBSDName( kwtd->kernCon, kwtd->ioDisk, kwtd->diskid );
		}
	}
	
	syslog( LOG_NOTICE, "Ejecting device %s (%s)\n", disk->szDeviceName, disk->filename );
	
done:
	if( IsntNull( data ) )
		(void) PGPFreeData( data );
	return;
}

	static void *
sMountDiskThread(
	void *		arg )
{
	PGPdiskEngineKernelWorkerThreadData *	kwtd	= arg;
	PGPError						err				= kPGPError_NoErr;
	kern_return_t					kr				= 0;
	io_connect_t					controllerPort	= NULL;
	io_connect_t					driverPort		= NULL;
	CFStringRef						str;
	CFDataRef						data;
	PGPdiskCreateDiskParams			diskParams;
	mach_msg_type_number_t			len				= 0;
	io_service_t					driverDevice;
	char							currentPath[PATH_MAX];
	char							filename[PATH_MAX];
	PGPUInt32						diskid			= 0;
	PGPdiskKernelConnectionContext *	driverConn	= NULL;
	
	err = PGPdiskEngineSetDiskMounted( kwtd->ioDisk, kwtd->readonly ); CKERR;

	err = PGPdiskEngineGetIODiskFilename( kwtd->ioDisk, filename );
	syslog( LOG_NOTICE, "Mounting disk %s\n", filename );
	
	/* Create the drive */
	kr = IOServiceOpen( kwtd->kernCon->controllerDevice, mach_task_self(), 0, &controllerPort );
	CKMACHERR( kr, "Couldn't create user client");

	diskParams.diskSizeBytes = PGPdiskEngineGetDiskSize( kwtd->ioDisk );
	diskParams.mountReadOnly = kwtd->readonly;
	
	err = PGPdiskGetCurrentExecutablePath( currentPath ); CKERR;
	snprintf( filename, sizeof( filename ), "%s/mountedpgpdisk.icns", currentPath );
	str = CFStringCreateWithCString( kCFAllocatorDefault, filename, kCFStringEncodingUTF8 );
	data = CFStringCreateExternalRepresentation( kCFAllocatorDefault, str, kCFStringEncodingUTF8, 0x00 );
	pgpCopyMemory( CFDataGetBytePtr( data ), &diskParams.iconPath, CFDataGetLength( data ) );
	CFRelease( data );
	CFRelease( str );
	
	/* diskParams.ejectOnLogout = kwtd->ejectOnLogout; */
	/* Force ejectOnLogout to be TRUE for now */
	diskParams.ejectOnLogout = TRUE;

	diskParams.autoDiskMount = kwtd->autoDiskMount;
	
	kr = io_connect_method_structureI_structureO( controllerPort,
		kPGPdiskControllerUserClient_CreateDisk,
		(void *) &diskParams, sizeof( diskParams ),
		NULL, &len );
	CKMACHERR( kr, "Couldn't create disk" );
	
	IOServiceClose( controllerPort );
	
	syslog( LOG_DEBUG, "Disk created!\n" );

	driverDevice = sGetDevice( kwtd->kernCon->masterPort, kPGPdiskDriver_ClassName );
	if( !driverDevice )
		RETERR( kPGPError_UnknownError );
		
	kr = IOServiceOpen( driverDevice, mach_task_self(), 0, &driverPort );
	CKMACHERR( kr, "Couldn't create user client" );
	
	/* Activate the drive */
	diskid = 0;
	len = sizeof( diskid );
	kr = io_connect_method_scalarI_scalarO( driverPort,
		kPGPdiskDriverUserClient_ActivateDrive,
		NULL, 0,
		&diskid, &len );
	CKMACHERR( kr, "Couldn't activate disk" );

	kwtd->diskid = diskid;

	syslog( LOG_DEBUG, "Disk activated!\n" );
	
	/* Disk has been created! */
	
	driverConn = PGPNewData( kwtd->mgr,
		sizeof( PGPdiskKernelConnectionContext ),
		kPGPMemoryMgrFlags_Clear );
	CKNULL( driverConn );
	
	driverConn->driverPort = driverPort;
	driverConn->thread = pthread_self();
	err = PGPdiskEngineSetDiskKernelConnectionContext( kwtd->ioDisk, driverConn ); CKERR;
	
	sGetKernelMessages( kwtd->ioDisk, kwtd, driverPort );
	IOServiceClose( driverPort );

done:
	/* Disk has been ejected */
	(void) PGPdiskEngineSetDiskUnmounted( kwtd->ioDisk );
	(void) PGPdiskEngineCloseDisk( kwtd->ioDisk );
	
	if( IsntNull( driverConn ) )
		(void) PGPFreeData( driverConn );
	
	if( IsntNull( kwtd ) )
		(void) PGPFreeData( kwtd );
	
	return NULL;
}

	static PGPError
sFindMountPoint(
	char *				devName,
	char *				outMountedAt,
	PGPUInt32			outMountedAtLength )
{
	char				volName[PATH_MAX];
	struct statfs *		mntPoints	= NULL;
	PGPInt32			numMounts	= 0;
	PGPInt32			mountNum	= 0;
	PGPError			err			= kPGPError_NoErr;

	snprintf( volName, sizeof( volName ), "/dev/%s", devName );
	
	numMounts = getmntinfo( &mntPoints, 0 );

	for( mountNum = 0; mountNum < numMounts; mountNum++ )
	{
		if( strcmp( mntPoints[mountNum].f_mntfromname, volName ) == 0 )
			break;
	}
	if( mountNum == numMounts )
		RETERR( kPGPdiskError_DiskNotMounted );
		
	snprintf( outMountedAt, outMountedAtLength, "%s", mntPoints[mountNum].f_mntonname );
	
done:
	return err;
}

	static PGPError
sCopyDiskIcon(
	char *				mountPoint )
{
	PGPError			err				= kPGPError_NoErr;
	char				iconFilenameFrom[PATH_MAX];
	char				currentPath[PATH_MAX];
	char				iconFilenameTo[PATH_MAX];
	PGPInt32			fdFrom			= 0;
	PGPInt32			fdTo			= 0;
	PGPInt32			bytesRead		= 0;
	PGPInt32			bytesWritten	= 0;
	PGPInt32			retval			= 0;
	struct stat			sb;
	char				buf[10240];
	
	snprintf( iconFilenameTo, sizeof( iconFilenameTo ), "%s/.VolumeIcon.icns", mountPoint );
	
	err = PGPdiskGetCurrentExecutablePath( currentPath ); CKERR;
	snprintf( iconFilenameFrom, sizeof( iconFilenameFrom ), "%s/mountedpgpdisk.icns", currentPath );
	
	/* Check if custom volumeicon already exists */
	if( stat( iconFilenameTo, &sb ) == 0 )
		RETERR( kPGPError_NoErr );
		
	fdFrom = open( iconFilenameFrom, O_RDONLY, 0 );
	if( fdFrom < 0 )
		RETERR( kPGPError_UnknownError );
		
	fdTo = open( iconFilenameTo, O_WRONLY | O_CREAT | O_TRUNC, 0644 );
	if( fdTo < 0 )
		RETERR( kPGPError_UnknownError );
		
	/* Copy the data in the icon file, 1024 bytes at a time */
	do
	{
		bytesRead = read( fdFrom, buf, sizeof( buf ) );
		if( bytesRead > 0 )
		{
			bytesWritten = 0;
			while( bytesWritten < bytesRead )
			{
				retval = write( fdTo, buf + bytesWritten, bytesRead - bytesWritten );
				if( retval < 0 )
					RETERR( kPGPError_WriteFailed );
				bytesWritten += retval;
			}
		}
	} while( bytesRead > 0 );

done:
	if( fdFrom > 0 )
		(void) close( fdFrom );
	if( fdTo > 0 )
		(void) close( fdTo );
	return err;
}

	static PGPError
sSetDiskIcon(
	char *				mountPoint )
{
	CInfoPBRec			cpb;
	FSRef				fsRef;
	FSVolumeInfo		volInfo;
	FSCatalogInfo		info;
	FSVolumeRefNum		vRefNum		= -1;
	int					error		= 0;
	PGPError			err			= kPGPError_NoErr;
	

	error = FSPathMakeRef( mountPoint, &fsRef, NULL );
	if( error )
		RETERR( kPGPdiskError_DiskNotMounted );

	error = FSGetCatalogInfo( &fsRef, kFSCatInfoVolume,
		&info, NULL, NULL, NULL );
	if( error )
		RETERR( kPGPdiskError_DiskNotMounted );

	error = FSGetVolumeInfo( info.volume, 0, &vRefNum,
		kFSVolInfoFSInfo, &volInfo, NULL, &fsRef );
	if( error )
		RETERR( kPGPdiskError_DiskNotMounted );

	/* Set the custom icon bit of the root directoy catalog entry */
	memset( &cpb, 0, sizeof( cpb ) );
	cpb.dirInfo.ioVRefNum = vRefNum;
	
	error = PBGetCatInfoSync( &cpb );
	if( error )
		RETERR( kPGPdiskError_DiskNotMounted );

	cpb.dirInfo.ioDrUsrWds.frFlags |= kHasCustomIcon;

	error = PBSetCatInfoSync( &cpb );
	if( error )
		RETERR( kPGPdiskError_DiskNotMounted );
		
done:
	return err;
}

	static PGPError
sUpdateFinder( void )
{
	/*
	 * Supposedly sending a kAEFinderSuite,kAESync apple event will tell the Finder to
	 * update it's disk icons, but it doesn't seem to work right now.  :-(
	 */
	return kPGPError_NoErr;
}

	static PGPError
sMountDiskAtPath(
	const char *			deviceName,
	const char *			mountPoint )
{
	PGPError				err			= kPGPError_NoErr;
	struct hfs_mount_args	hfsargs;
	char					szFullDeviceName[PATH_MAX];
	PGPInt32				retval		= 0;
	
	memset( &hfsargs, 0, sizeof( hfsargs ) );
	snprintf( szFullDeviceName, sizeof( szFullDeviceName ), "/dev/%s", deviceName );
	hfsargs.fspec = szFullDeviceName;
	hfsargs.hfs_uid = getuid();
	hfsargs.hfs_gid = getgid();
	hfsargs.hfs_mask = ACCESSPERMS;	/* 0777 */
	
	retval = mount( "hfs", mountPoint, 0, &hfsargs );
	if( retval < 0 )
	{
		syslog( LOG_WARNING, "Could not mount %s: %s\n", deviceName, strerror( errno ) );
		RETERR( kPGPdiskError_DiskNotMounted );
	}
	
done:
	return err;
}

	static PGPError
sWaitUntilDiskMounted(
	PGPdiskKernelWorkerContextRef	kernCon,
	PGPdiskIODiskRef				disk,
	PGPBoolean						autoDiskMount,
	const char *					preferredMountPoint )
{
	PGPError						err						= kPGPError_NoErr;
	char							deviceName[PATH_MAX];
	char							mountPoint[PATH_MAX];
	time_t							startTime;
	
	deviceName[0] = '\0';
	
	/* Not too pretty - loop for no more than 15 seconds, waiting for the /dev/diskX entry to be created. */
	startTime = time( NULL );
	while( time( NULL ) < startTime + 15 )
	{
		if( disk->szDeviceName[0] != '\0' )
		{
			if( deviceName[0] == '\0' )
			{
				snprintf( deviceName, sizeof( deviceName ), "%s", disk->szDeviceName );
				break;
			}
			else
				usleep( 250000 );
		}
	}

	/* /dev/diskX entry has been created */		

	if( autoDiskMount )
	{
		startTime = time( NULL );
		while( time( NULL ) < startTime + 15 )
		{
			err = sFindMountPoint( deviceName, mountPoint, sizeof( mountPoint ) );
			if( IsPGPError( err ) )
				usleep( 250000 );
			else
				break;
		}
		/* If there was an error the last time we called sFindMountPoint, return */
		CKERR;
	}
	else if( preferredMountPoint[0] != '\0' )
	{
		err = sMountDiskAtPath( deviceName, preferredMountPoint ); CKERR;
		strncpy( mountPoint, preferredMountPoint, sizeof( mountPoint ) - 1 );
		mountPoint[sizeof(mountPoint) - 1] = '\0';
	}

	if( mountPoint[0] != '\0' )
	{
		/*
		* Once it does mount, copy mountedpgpdisk.icns to <NewVolume>/.VolumeIcon.icns and set the HasCustomIcon
		* HFS attribute
		*/
	
		(void) sCopyDiskIcon( mountPoint );
		(void) sSetDiskIcon( mountPoint );
	
		/* Supposedly the Finder is broken, so you need to send this twice */
		(void) sUpdateFinder();
		(void) sUpdateFinder();
	}

done:
	return err;
}

	PGPError
PGPdiskEngineMountDisk(
	PGPdiskBEContextRef				diskcon,
	PGPdiskIODiskRef				disk,
	PGPdiskIODiskUser *				user,
	PGPBoolean						readonly,
	PGPBoolean						ejectOnLogout,
	PGPBoolean						autoDiskMount,
	const char *					mountPoint,
	const char *					passphrase )
{
	PGPError								err					= kPGPError_NoErr;
	pthread_t								thread;
	PGPdiskEngineKernelWorkerThreadData *	kwtd				= NULL;
	PGPBoolean								bIsMounted			= FALSE;
	PGPBoolean								bStartedMountThread	= FALSE;

	/* Try to load driver if it's not already loaded */
	err = PGPdiskEngineLoadDriver(); CKERR;
	err = sInitializeController( diskcon->kernCon ); CKERR;
	
	kwtd = PGPNewData( diskcon->mgr,
		sizeof( PGPdiskEngineKernelWorkerThreadData ),
		kPGPMemoryMgrFlags_Clear );
	CKNULL( kwtd );
	
	kwtd->mgr = diskcon->mgr;
	kwtd->kernCon = diskcon->kernCon;
	kwtd->ioDisk = disk;
	kwtd->readonly = readonly;
	kwtd->ejectOnLogout = ejectOnLogout;
	kwtd->autoDiskMount = autoDiskMount;
	
	/* Make sure disk isn't already mounted */
	err = PGPdiskEngineGetDiskBooleanProperty( disk, kPGPdiskDiskProp_IsMounted, &bIsMounted ); CKERR;
	if( bIsMounted )
		RETERR( kPGPdiskError_DiskAlreadyMounted );

	/* Try to decrypt the disk */
	err = PGPdiskEngineDecryptDisk( diskcon, disk, user, passphrase ); CKERR;

	pthread_create( &thread, NULL, sMountDiskThread, kwtd );
	pthread_detach( thread );
	
	bStartedMountThread = TRUE;
	
	err = sWaitUntilDiskMounted( diskcon->kernCon, disk, autoDiskMount, mountPoint ); CKERR;
	/* For mounting disks at custom mount points - doesn't work right now
	if( !autoDiskMount && IsPGPError( err ) )
	{
		// Disk never mounted - close it and try again, letting autodiskmount pick a mount point for us
		err = PGPdiskEngineGetDiskKernelConnectionContext( disk, &driverConn ); CKERR;
		IOServiceClose( driverConn->driverPort );
		err = PGPdiskEngineSetDiskUnmounted( disk ); CKERR;

		err = PGPdiskEngineMountDisk( diskcon, disk, user, readonly, ejectOnLogout, TRUE, mountPoint, passphrase ); CKERR;
	}
	*/

done:
	if( IsPGPError( err ) && !bStartedMountThread )
	{
		if( IsntNull( kwtd ) )
			(void) PGPFreeData( kwtd );
	}
	return err;
}

	static PGPError
sTryToUnmountNicely(
	const char *	pszDeviceName )
{
	PGPError		err				= kPGPError_NoErr;
	char			szCommand[128];

	/*
	 * Assumption:
	 * if hdiutil returns 0, the disk was unmounted successfully
	 */

	snprintf( szCommand, sizeof( szCommand ), "/usr/bin/hdiutil eject %s >& /dev/null", pszDeviceName );

	if( system( szCommand ) != 0 )
		err = kPGPdiskError_CantUnmountDisk;

	return err;
}

	PGPError
PGPdiskEngineUnmountDisk(
	PGPdiskKernelWorkerContextRef	kernCon,
	PGPdiskIODiskRef				disk,
	PGPBoolean						bForced )
{
	PGPError						err				= kPGPError_NoErr;
	PGPBoolean						bIsMounted		= FALSE;
	PGPdiskKernelConnectionContext *driverConn		= NULL;

	err = PGPdiskEngineGetDiskBooleanProperty( disk, kPGPdiskDiskProp_IsMounted, &bIsMounted ); CKERR;
	if( !bIsMounted )
		RETERR( kPGPdiskError_DiskNotMounted );

	/* First try to unmount it nicely */
	err = sTryToUnmountNicely( disk->szDeviceName );
	if( IsPGPError( err ) && bForced )
	{
		/* Now close the driver port, removing the /dev/diskX entry */
		err = PGPdiskEngineGetDiskKernelConnectionContext( disk, &driverConn ); CKERR;
		IOServiceClose( driverConn->driverPort );
		pthread_cancel( driverConn->thread );
	}

	if( IsntPGPError( err ) )
	{
		/* Our connection to the driver doesn't close when we call IOServiceClose */
		(void) PGPdiskEngineSetDiskUnmounted( disk );
		if( bForced )
			(void) PGPdiskEngineCloseDisk( disk );
	}
	
done:
	return err;
}

	PGPError
PGPdiskEngineFormatDisk(
	PGPdiskBEContextRef				diskcon,
	PGPdiskIODiskRef				disk,
	PGPdiskIODiskUser *				adminUser,
	const char *					passphrase,
	const char *					volumeName,
	PGPdiskFilesystemType			fsType )
{
	PGPError						err				= kPGPError_NoErr;
	PGPBoolean						bIsMounted		= FALSE;
	time_t							startTime;
	char							szCommand[256];
	char							szDevice[PATH_MAX];
	PGPBoolean						badFSType		= FALSE;
	struct stat						sb;
	PGPUInt64						diskSize		= 0;
	PGPUInt64						journalSize		= 0;
	
	/* Make sure it's not already mounted */
	err = PGPdiskEngineGetDiskBooleanProperty( disk, kPGPdiskDiskProp_IsMounted, &bIsMounted ); CKERR;
	if( bIsMounted )
		RETERR( kPGPdiskError_DiskAlreadyMounted );
	
	/* Mount disk with autoDiskMount = FALSE */
	err = PGPdiskEngineMountDisk( diskcon,
		disk,
		adminUser,
		FALSE,	/* readonly */
		FALSE,	/* ejectOnLogout */
		FALSE,	/* autoDiskMount */
		"", /* mountPoint */
		passphrase ); CKERR;
		
	/* Wait for no longer than 10 seconds for the disk to get the BSD name */
	startTime = time( NULL );
	while( ( disk->szDeviceName[0] == '\0' ) && ( time( NULL ) < startTime + 10 ) )
		usleep( 250000 );
	
	if( disk->szDeviceName[0] == '\0' )
		RETERR( kPGPError_UnknownError );
	
	/* newfs the disk */
	switch( fsType )
	{
		case kPGPdiskFilesystemType_FAT12:
			snprintf( szCommand, sizeof( szCommand ), "/sbin/newfs_msdos -F 12 -v \"%s\" /dev/%s >& /dev/null", volumeName, disk->szDeviceName );
			break;
		case kPGPdiskFilesystemType_FAT16:
			snprintf( szCommand, sizeof( szCommand ), "/sbin/newfs_msdos -F 16 -v \"%s\" /dev/%s >& /dev/null", volumeName, disk->szDeviceName );
			break;
		case kPGPdiskFilesystemType_FAT12Or16:
			/* Unsupported */
			badFSType = TRUE;
			break;
		case kPGPdiskFilesystemType_FAT32:
			snprintf( szCommand, sizeof( szCommand ), "/sbin/newfs_msdos -F 32 -v \"%s\" /dev/%s >& /dev/null", volumeName, disk->szDeviceName );
			break;
		case kPGPdiskFilesystemType_NTFS:
			/* Unsupported */
			badFSType = TRUE;
			break;
		case kPGPdiskFilesystemType_HFS:
			snprintf( szCommand, sizeof( szCommand ), "/sbin/newfs_hfs -h -v \"%s\" /dev/%s >& /dev/null", volumeName, disk->szDeviceName );
			break;
		case kPGPdiskFilesystemType_HFSPlus:
			snprintf( szCommand, sizeof( szCommand ), "/sbin/newfs_hfs -v \"%s\" /dev/%s >& /dev/null", volumeName, disk->szDeviceName );
			break;
		case kPGPdiskFilesystemType_HFSPlusJournaled:
			diskSize = PGPdiskEngineGetDiskSize( disk );
			if( diskSize < 20 * 1024 * 1024 ) 		/* 20 MB */
				journalSize = 512 * 1024;		/* 512 KB */
			else if( diskSize < 50 * 1024 * 1024 )	/* 50 MB */
				journalSize = 2 * 1024 * 1024;	/* 2 MB */
			else if( diskSize < 100 * 1024 * 1024 )	/* 100 MB */
				journalSize = 4 * 1024 * 1024;	/* 4 MB */
			else
				journalSize = 8 * 1024 * 1024; 	/* 8 MB */
			snprintf( szCommand, sizeof( szCommand ), "/sbin/newfs_hfs -J %lld -v \"%s\" /dev/%s >& /dev/null",
				journalSize, volumeName, disk->szDeviceName );
			break;
		case kPGPdiskFilesystemType_UFS:
			snprintf( szCommand, sizeof( szCommand ), "/sbin/newfs -v \"%s\" /dev/%s >& /dev/null", volumeName, disk->szDeviceName );
			break;
		default:
			/* Unsupported */
			badFSType = TRUE;
			break;
	}
	
	if( !badFSType )
	{
		/* Loop for a maximum of 20 seconds, making sure the /dev/disk? is owned by us */
		startTime = time( NULL );
		snprintf( szDevice, sizeof( szDevice ), "/dev/%s", disk->szDeviceName );
		do
		{
			stat( szDevice, &sb );
			usleep( 250000 );
		} while( ( sb.st_uid != getuid() ) && ( time( NULL ) < startTime + 20 ) );
		
		if( sb.st_uid != getuid() )
			err = kPGPdiskError_CantFormatDisk;
		else 
		{
			if( system( szCommand ) != 0 )
				err = kPGPdiskError_CantFormatDisk;
		}
	}

	/* Unmount the disk */
	(void) PGPdiskEngineUnmountDisk(
		diskcon->kernCon,
		disk,
		TRUE );
	
done:
	if( badFSType )
		err = kPGPError_FeatureNotAvailable;
	return err;
}


/*__Editor_settings____

	Local Variables:
	tab-width: 4
	End:
	vi: ts=4 sw=4
	vim: si
_____________________*/
