/*____________________________________________________________________________
		Copyright (C) 2002 PGP Corporation
        All rights reserved.

        $Id: pgpDiskHeaders.h,v 1.12 2002/08/06 20:09:35 dallen Exp $
____________________________________________________________________________*/
#ifndef Included_pgpDiskHeaders_h	/* [ */
#define Included_pgpDiskHeaders_h

#include "pgpPubTypes.h"
#include "pgpBase.h"
#include "pgpDiskEngine.h"

/*
 * The basic form of the disk looks like
 *
 * | -
 * | |
 * | | PGPdiskOnDiskHeaderInfo
 * | |
 * | 
 * | PGPdiskOnDiskMainHeader
 * | -
 *
 * ... [DATA, encrypted in 512 byte blocks]
 *
 * Array of:
 *
 * | -
 * | |
 * | | PGPdiskOnDiskHeaderInfo
 * | |
 * |
 * | PGPdiskOnDiskUserHeader
 * | -
 * | |
 * | | PGPdiskOnDiskUserInfo
 * | |
 * |
 * | PGPdiskOnDiskUserWithPub or PGPdiskOnDiskUserWithSym
 * | -
 * | [DATA, encrypted disk symmetric key] (Only for public key users)
 * | -
 *
 * Note that for compatibility with Windows, all values written to disk
 * must be written as _little endian_.
 */


/* Arbitrary constants */
#define kMaxBytesDiskUser		(32 * 1024)
#define kPGPdiskMaxBlocksHeader						128

/* PGPdisk version constants */
#define kPGPdiskMainHeader_MajorVersion				6
#define kPGPdiskMainHeader_MinorVersion				1
#define kPGPdiskMainHeader_CompatMajorVersion		1
#define kPGPdiskMainHeader_PreSubKeyMajorVersion	5

#define kPGPdiskMainHeader_MaxSizeRoot				256
#define kPGPdiskReservedHeaderBlocks				4 

/* PGPdisk Magic numbers and types */
enum 
{
	kPGPdiskHeaderMagic		= 0x64504750,	// 'dPGP', backwards since little-endian

	/* Header types */
	kPGPdiskMainType		= 0x4E49414D,	// 'NIAM' 
	kPGPdiskUserType		= 0x52455355,	// 'RESU' 
	kPGPdiskOldPubKeyType	= 0x59454B50,	// 'YEKP' 

	/* User types */
	kPGPdiskUserMagic		= 0x52455355,	// 'RESU' 
	kPGPdiskUserWithSymType	= 0x4D4D5953,	// 'MMYS'
	kPGPdiskUserWithPubType	= 0x59454B50	// 'YEKP' 
};

/* A symmetric, unexpanded cryptographic key */
typedef union PGPdiskSymmetricKey
{
	PGPByte		bytes[128];
	PGPUInt32	words[128/sizeof(PGPUInt32)];
} PGPdiskSymmetricKey;

typedef PGPdiskSymmetricKey PGPdiskEncryptedKey;

/* Some bytes of data for passphrase verification */
typedef struct PGPdiskCheckBytes
{
	PGPByte	bytes[16];
} PGPdiskCheckBytes;

typedef union PGPdiskPassphraseSalt
{
	PGPByte		bytes[16];
	PGPUInt32	words[16/sizeof(PGPUInt32)];
} PGPdiskPassphraseSalt;

/* Only used for passphrase users */
typedef struct PGPdiskPassphraseKeyInfo
{
	PGPdiskEncryptedKey	encryptedKey;
	PGPdiskCheckBytes	checkBytes;

	PGPUInt16			hashReps;	// # of hashes on passphrase
	PGPUInt16			reserved1;
	PGPUInt32			reserved2;
} PGPdiskPassphraseKeyInfo;


/*
 * Generic Disk Header
 */
typedef struct PGPdiskOnDiskHeaderInfo
{
	PGPUInt32	headerMagic;		// Always kPGPdiskHeaderMagic
	PGPUInt32	headerType;			// One of the HeaderType enums
	PGPUInt32	headerSize;			// Total size of this header, in bytes

	PGPUInt32	headerCRC;			// CRC of this header
	PGPUInt64	nextHeaderOffset;	// Offset to next header from file start
									// 0 = no additional headers
	PGPUInt32	reserved[2];
} PGPdiskOnDiskHeaderInfo;

/*
 * Main Header
 */
typedef struct PGPdiskOnDiskMainHeader
{
	PGPdiskOnDiskHeaderInfo	hdrInfo;

	PGPUInt8	majorVersion;
	PGPUInt8	minorVersion;
	PGPUInt16	reserved;

	PGPUInt64	numFileBlocks;			// Size of the file, in blocks
	PGPUInt64	numHeaderBlocks;		// Size of leading header data
	PGPUInt64	numDataBlocks;			// Size of encrypted data

	PGPdiskEncryptionAlgorithm	algorithm;

	PGPdiskPassphraseSalt		salt;

	/* ...ECTWR ........ ........ ........ */
#if BYTE_ORDER == LITTLE_ENDIAN
	PGPUInt32	isBeingReEncrypted 			: 1;	// is disk being reencrypted?
	PGPUInt32	isWiped						: 1;	// users wiped?
	PGPUInt32	usesCustomTimeout			: 1;	// custom timeout?
	PGPUInt32	usesCustomEjectOnLogout		: 1;	// custom eject on logout setting?
	PGPUInt32	ejectOnLogout				: 1;	// eject on logout?
	PGPUInt32	unused						: 27;
#elif BYTE_ORDER == BIG_ENDIAN
	PGPUInt32	unused1						: 3;
	PGPUInt32	ejectOnLogout				: 1;	// eject on logout?
	PGPUInt32	usesCustomEjectOnLogout		: 1;	// custom eject on logout setting?
	PGPUInt32	usesCustomTimeout			: 1;	// custom timeout?
	PGPUInt32	isWiped						: 1;	// users wiped?
	PGPUInt32	isBeingReEncrypted 			: 1;	// is disk being reencrypted?
	PGPUInt32	unused2						: 24;
#else
#error BYTE_ORDER not defined!
#endif

	PGPUInt32	customTimeout;				// custom unmount timeout (in minutes, 0 = no auto unmount)
	PGPUInt64	numBlocksReEncrypted;		// number of blocks reencrypted

	char		defaultRoot[kPGPdiskMainHeader_MaxSizeRoot];	// Preferred mount point
} PGPdiskOnDiskMainHeader;

/*
 * User Header
 */
typedef struct PGPdiskOnDiskUserInfo
{
	PGPUInt32	userMagic;
	PGPUInt32	userType;
	PGPUInt32	userSize;

	/*
	 * This can be kinda funny.  On little endian machines, it actually looks like:
	 *     ....DALR ........ ........ ........
	 * so the big endian machines need to compensate for this strangeness
	 */
#if BYTE_ORDER == LITTLE_ENDIAN
	PGPUInt32	readOnly 	: 1;
	PGPUInt32	locked 		: 1;		/* User cannot be removed (ADK) */
	PGPUInt32	adminUser	: 1;
	PGPUInt32	disabled	: 1;
	PGPUInt32	unused		: 28;
#elif BYTE_ORDER == BIG_ENDIAN
	PGPUInt32	unused1		: 4;
	PGPUInt32	disabled	: 1;
	PGPUInt32	adminUser	: 1;
	PGPUInt32	locked 		: 1;		/* User cannot be removed (ADK) */
	PGPUInt32	readOnly 	: 1;
	PGPUInt32	unused2		: 24;
#else
#error BYTE_ORDER not defined!
#endif

	PGPUInt32	reserved[4];
} PGPdiskOnDiskUserInfo;

/* Public key user */
typedef struct PGPdiskOnDiskUserWithPub
{
	PGPdiskOnDiskUserInfo	header;
	
	PGPUInt32			encryptedKeyOffset;	// Relative to start of PGPdiskOnDiskUserWithPub
	PGPUInt32			encryptedKeySize;

	PGPKeyID			keyID;
	PGPdiskCheckBytes	checkBytes;
	PGPKeyID			subKeyID;

	PGPUInt32			reserved[8];

	// Encrypted key here...

	// Binary data here decrypted with the key referenced by keyID
	// is the key that the disk is encrypted to.
	// You can check this by checking that the key's fingerprint
	// encrypted with the key equals the checkBytes
	
	// Encrypted with PGPPublicKeyEncrypt and decrypted with
	// PGPPrivateKeyDecrypt.
	// messageFormat = kPGPPublicKeyMessageFormat_PGP
	// No other PGPOptionList options
} PGPdiskOnDiskUserWithPub;

/* Symmetric key (passphrase) user */
typedef struct PGPdiskOnDiskUserWithSym
{
	PGPdiskOnDiskUserInfo		header;

	char						userName[kPGPdiskMaxUserNameSize + 1];
	PGPdiskPassphraseKeyInfo	keyInfo;
} PGPdiskOnDiskUserWithSym;

/* Generic user header */
typedef struct PGPdiskOnDiskUserHeader
{
	PGPdiskOnDiskHeaderInfo	hdrInfo;

	/* .......N ........ ........ ........ */
#if BYTE_ORDER == LITTLE_ENDIAN
	PGPUInt32	notExported		: 1;
	PGPUInt32	unused			: 31;
#elif BYTE_ORDER == BIG_ENDIAN
	PGPUInt32	unused1			: 7;
	PGPUInt32	notExported		: 1;
	PGPUInt32	unused			: 24;
#else
#error BYTE_ORDER not defined!
#endif

	PGPUInt32	exportedUserOffset;		// Relative to start of PGPdiskOnDiskUserHeader
	PGPUInt32	exportedUserSize;

	PGPUInt32	reserved[4];

	// Exported PGPdisk user structure goes here...
	// (PGPdiskOnDiskUserWithPub or PGPdiskOnDiskUserWithSym)
} PGPdiskOnDiskUserHeader;

#endif /* ] Included_pgpDiskHeaders_h */

