/*____________________________________________________________________________
		Copyright (C) 2002 PGP Corporation
        All rights reserved.

        $Id: CPGPdiskUser.cpp,v 1.16 2002/11/02 19:22:38 wjb Exp $
____________________________________________________________________________*/

#include "pgpClassesConfig.h"

#include "CArray.h"
#include "CSecureObject.h"
#include "CString.h"
#include "UString.h"

#include "pgpClientErrors.h"
#include "CCipherContext.h"
#include "UCommonStrings.h"
#include "CPGPKey.h"

#include "CPGPdiskContext.h"
#include "CPGPdiskUser.h"

_USING_PGP

_PGP_BEGIN

// Types

// Align to 1.
#if PGP_WIN32
#pragma pack(push, 1)
#endif

struct OnDiskUserInfo
{
	enum
	{
		kMaxBytesDiskUser	= PFLConstants::kBytesPerKb * 32	// arbitrary
	};

#if defined(_M_IX86)
	enum 
	{
		kUserMagic			= 'RESU', 

		kUserWithSymType	= 'MMYS', 
		kUserWithPubType	= 'YEKP', 
	};
#endif	// _M_IX86

	PGPUInt32	userMagic;
	PGPUInt32	userType;
	PGPUInt32	userSize;

	PGPUInt32	readOnly 	: 1;
	PGPUInt32	locked 		: 1;
	PGPUInt32	adminUser	: 1;
	PGPUInt32	disabled	: 1;
	PGPUInt32	unused		: 28;

	PGPUInt32	reserved[4];
};

struct OnDiskUserWithPub
{
	OnDiskUserInfo	header;
	
	PGPUInt32	encryptedKeyOffset;
	PGPUInt32	encryptedKeySize;

	PGPKeyID			keyID;
	Crypto::CheckBytes	checkBytes;
	PGPKeyID			subKeyID;

	PGPUInt32	reserved[8];

	// Encrypted key here...
};

struct OnDiskUserWithSym
{
	OnDiskUserInfo	header;

	char						userName[kPGPdiskMaxUserNameSize + 1];
	Crypto::PassphraseKeyInfo	keyInfo;
};

// Restore alignment.
#if PGP_WIN32
#pragma pack(pop)
#endif

_PGP_END

// Class CPGPdiskUser member functions

CPGPdiskUser::CPGPdiskUser(const CPGPdiskContext *pContext)
{
	Init();

	pgpAssertAddrValid(pContext, CPGPdiskContext);
	mPContext = pContext;
}

CPGPdiskUser::CPGPdiskUser(const CPGPdiskUser& user)
{
	Init();
	Assign(user);
}

CPGPdiskUser::~CPGPdiskUser()
{
	Clear();
}

CPGPdiskUser& 
CPGPdiskUser::operator=(const CPGPdiskUser& user)
{
	Assign(user);
	return *this;
}

const CPGPdiskContext * 
CPGPdiskUser::Context() const
{
	pgpAssert(IsAssigned());
	return mPContext;
}

PGPBoolean 
CPGPdiskUser::IsEqual(const CPGPdiskUser& user) const
{
	if (this == &user)
		return TRUE;

	// Must both be assigned.
	if (!IsAssigned() || !user.IsAssigned())
		return FALSE;

	// Must be same type and same cipher.
	if (GetAsUserInfo()->userSize != user.GetAsUserInfo()->userSize)
		return FALSE;

	if (HasPublicKey())
	{
		// If public key, compare key IDs.
		return (mKeyID.IsEqual(user.mKeyID));
	}
	else
	{
		// If symmetric key, compare PassphraseKeyInfos and user names.
		CString	userName1, userName2;

		GetUserName(userName1);
		user.GetUserName(userName2);

		if (userName1 != userName2)
			return FALSE;

		return pgpMemoryEqual(&GetAsUserWithSym()->keyInfo, 
			&user.GetAsUserWithSym()->keyInfo, 
			sizeof(Crypto::PassphraseKeyInfo));
	}
}

PGPBoolean 
CPGPdiskUser::IsAssigned() const
{
	return mIsAssigned;
}

PGPBoolean 
CPGPdiskUser::HasPublicKey() const
{
	return (GetAsUserInfo()->userType == OnDiskUserInfo::kUserWithPubType);
}

PGPBoolean 
CPGPdiskUser::HasSymmetricKey() const
{
	return (GetAsUserInfo()->userType == OnDiskUserInfo::kUserWithSymType);
}

PGPBoolean 
CPGPdiskUser::IsReadOnly() const
{
	pgpAssert(IsAssigned());
	return (GetAsUserInfo()->readOnly);
}

PGPBoolean 
CPGPdiskUser::IsLocked() const
{
	pgpAssert(IsAssigned());
	return (GetAsUserInfo()->locked);
}

PGPBoolean 
CPGPdiskUser::IsAdminUser() const
{
	pgpAssert(IsAssigned());
	return (GetAsUserInfo()->adminUser);
}

PGPBoolean 
CPGPdiskUser::IsDisabled() const
{
	pgpAssert(IsAssigned());
	return (GetAsUserInfo()->disabled);
}

PGPBoolean 
CPGPdiskUser::IsCipheringEnabled() const
{
	pgpAssert(IsAssigned());
	return (mCipheringEnabled);
}

PGPBoolean 
CPGPdiskUser::CanSetReadOnly() const
{
	pgpAssert(IsAssigned());
	return TRUE;
}

PGPBoolean 
CPGPdiskUser::CanSetLocked() const
{
	pgpAssert(IsAssigned());
	return HasPublicKey();
}

PGPBoolean 
CPGPdiskUser::CanSetUserName() const
{
	pgpAssert(IsAssigned());
	return HasSymmetricKey();
}

void 
CPGPdiskUser::GetUserName(CString& userName) const
{
	pgpAssert(IsAssigned());

	if (HasPublicKey())
	{
		ImportPublicKey();

		if (mPGPKeyPresent)
		{
			mPGPKey.GetPrimaryUserIDName(userName);
		}
		else
		{
			CString	keyIDStr;
			mKeyID.GetKeyIDString(kPGPKeyIDString_Abbreviated, keyIDStr);

			userName.Format(UCommonStrings::Get(
				UCommonStrings::kUnknownUserWithIDString), keyIDStr.Get());
		}
	}
	else
	{
		userName = GetAsUserWithSym()->userName;
	}
}

PGPUInt32 
CPGPdiskUser::GetSizeForExport() const
{
	pgpAssert(IsAssigned());
	return GetAsUserInfo()->userSize;
}

void 
CPGPdiskUser::GetPGPKeyID(CPGPKeyID& keyID) const
{
	pgpAssert(IsAssigned());

	if (!HasPublicKey())
		THROW_PGPERROR(kPGPError_BadParams);

	keyID = mKeyID;
}

void 
CPGPdiskUser::GetPGPSubKeyID(CPGPKeyID& subKeyID) const
{
	pgpAssert(IsAssigned());

	if (!HasPublicKey())
		THROW_PGPERROR(kPGPError_BadParams);

	subKeyID = mSubKeyID;
}

void 
CPGPdiskUser::GetCipherContext(CCipherContext& context) const
{
	pgpAssert(IsAssigned());
	pgpAssert(IsCipheringEnabled());

	context = *mPCipher;
}

void 
CPGPdiskUser::SetIsReadOnly(PGPBoolean isReadOnly)
{
	pgpAssert(IsAssigned());
	GetAsUserInfo()->readOnly = isReadOnly;
}

void 
CPGPdiskUser::SetIsLocked(PGPBoolean isLocked)
{
	pgpAssert(IsAssigned());

	if (!CanSetLocked())
		THROW_PGPERROR(kPGPError_FeatureNotAvailable);

	GetAsUserInfo()->locked = isLocked;
}

void 
CPGPdiskUser::SetIsDisabled(PGPBoolean isDisabled)
{
	pgpAssert(IsAssigned());
	GetAsUserInfo()->disabled = isDisabled;
}

void 
CPGPdiskUser::SetUserName(const char *userName)
{
	pgpAssert(IsAssigned());

	if (HasPublicKey())
		THROW_PGPERROR(kPGPError_FeatureNotAvailable);

	UString::SmartStringCopy(GetAsUserWithSym()->userName, userName, 
		kPGPdiskMaxUserNameSize);
}

void 
CPGPdiskUser::EnableCiphering(
	const char						*passphrase,
	int								passkeylen,
	PGPdiskEncryptionAlgorithm		algorithm, 
	const Crypto::PassphraseSalt&	salt)
{
	pgpAssert(IsAssigned());

	if (IsCipheringEnabled())
		return;

	mPCipher = new CCipherContext(algorithm);

	if (HasPublicKey())
	{
		// Force reload of KeyDB... in case tokens have been inserted,etc
		// Note that because of our notification callback reloads only
		// occur when things have changed, we just tell disk when it
		// can safely update things -wjb
		Context()->ReloadPGPKeyDB();

		// Public key must be present on keyring.
		ImportPublicKey();

		if (!mPGPKeyPresent)
			THROW_PGPERROR(kPGPClientError_BadDiskPassKeyNotInRing);

		// Can we use it for decryption?
		if (!mPGPKey.GetBooleanProp(kPGPKeyProperty_CanDecrypt))
			// Changed error for private part missing from bad pass
			THROW_PGPERROR(kPGPClientError_BadDiskPassKeyNotInRing);

		// Does this passphrase unlock the private key?
		if (!mPGPKey.PassphraseIsValid(passphrase,passkeylen))
			THROW_PGPERROR(kPGPClientError_BadDiskPassphrase);

		// Decrypt the encrypted session key, assigning it to the cipher.
		PGPByte	*pEncryptedKey	= mUserBytes.Get() + 
			GetAsUserWithPub()->encryptedKeyOffset;

		if (mKeyID.IsEqual(mSubKeyID) && 
			(mPGPKey.GetNumericProp(kPGPKeyProperty_Version) > 
			kPGPKeyVersion_V3))
		{
			CPGPKeySet	pgpKeys(Context()->PGPKeyDB());
			CPGPKeyList	keyList(pgpKeys, kPGPKeyOrdering_Any);
			CPGPKeyIter	keyIter(keyList);
			CPGPKey		key;

			PGPBoolean	succeeded	= FALSE;

			while (keyIter.Next(key))
			{
				CPGPKeyID	keyID;
				key.GetKeyID(keyID);

				if (keyID.IsEqual(mKeyID))
				{
					CPGPKey	subKey;

					while (keyIter.NextSubKey(subKey))
					{
						try
						{
							DecryptUsingPublicKey(key, subKey, passphrase, passkeylen,
								pEncryptedKey, 
								GetAsUserWithPub()->encryptedKeySize, salt, 
								GetAsUserWithPub()->checkBytes, *mPCipher);

							mPGPSubKey.Attach(subKey);
							subKey.GetKeyID(mSubKeyID);

							succeeded = TRUE;
							break;
						}
						catch (CComboError&) {}
					}

					break;
				}
			}

			if (!succeeded)
				THROW_PGPERROR(kPGPClientError_BadDiskPassphrase);
		}
		else
		{
			DecryptUsingPublicKey(mPGPSubKey,mPGPSubKey, passphrase, passkeylen, pEncryptedKey, 
				GetAsUserWithPub()->encryptedKeySize, salt, 
				GetAsUserWithPub()->checkBytes, *mPCipher);
		}
	}
	else
	{
		pgpAssertStrValid(passphrase);

		// Decrypt encrypted session key, assigning it to the cipher.
		mPCipher->DecryptPassphraseKey(passphrase, salt, 
			GetAsUserWithSym()->keyInfo);
	}

	mCipheringEnabled = TRUE;
}

void 
CPGPdiskUser::DisableCiphering()
{
	pgpAssert(IsAssigned());

	if (!IsCipheringEnabled())
		return;

	delete mPCipher;
	mPCipher = NULL;

	mCipheringEnabled = FALSE;
}

void 
CPGPdiskUser::ChangeCipher(
	const CCipherContext&	newCipher, 
	const char				*passphrase)
{
	pgpAssert(newCipher.IsKeyAttached());

	CPGPdiskUser	tempUser(Context());

	if (HasPublicKey())
	{
		// Public key must be present for public PGPdisk keys.
		ImportPublicKey();

		if (!mPGPKeyPresent)
			THROW_PGPERROR(kPGPClientError_BadDiskPassKeyNotInRing);

		// Convert ADK users from READ ONLY 65DISKFIX -wjb
		PGPBoolean bReadOnly;

		bReadOnly=IsReadOnly();

		if (Context()->IsTherePGPdiskADK())
		{
			CPGPKey	adkKey;

			Context()->GetPGPdiskADKPubKey(adkKey);
			if(adkKey==mPGPKey)
				bReadOnly=FALSE;
		}
		// End ADK testing

		tempUser.Create(newCipher, mPGPKey, bReadOnly, IsLocked());
		tempUser.SetIsAdminUser(IsAdminUser());
	}
	else
	{
		CString	userName;
		GetUserName(userName);

		tempUser.Create(newCipher, userName, passphrase, IsReadOnly());
		tempUser.SetIsAdminUser(IsAdminUser());
	}

	Assign(tempUser);
}

void 
CPGPdiskUser::ChangePassphrase(const char *newPassphrase)
{
	pgpAssert(IsAssigned());
	pgpAssertStrValid(newPassphrase);

	if (!IsCipheringEnabled() || !HasSymmetricKey())
		THROW_PGPERROR(kPGPError_BadParams);

	mPCipher->EncryptPassphraseKey(newPassphrase, 
		GetAsUserWithSym()->keyInfo);
}

void 
CPGPdiskUser::Create(
	const CCipherContext&	cipher, 
	const CPGPKey&			pubKey, 
	PGPBoolean				readOnly, 
	PGPBoolean				isLocked)
{
	pgpAssert(cipher.IsInitialized());

	// Encrypt decypted key using the private key of the public key.
	CPGPKey	subKey;
	if (!pubKey.GetKeyForUsage(subKey))
		THROW_PGPERROR(kPGPError_PublicKeyUnimplemented);

	Crypto::CheckBytes		checkBytes;
	CSecureArray<PGPByte>	encryptedKey;
	PGPUInt32				encryptedKeySize;

	EncryptUsingPublicKey(subKey, cipher, encryptedKey, encryptedKeySize, 
		checkBytes);

	// Get key IDs.
	CPGPKeyID	keyID;
	pubKey.GetKeyID(keyID);

	CPGPKeyID	subKeyID;
	subKey.GetKeyID(subKeyID);

	// Import the key.
	ImportRawPublicKey(keyID, subKeyID, encryptedKey.Get(), 
		encryptedKeySize, checkBytes, readOnly, isLocked);
}

void 
CPGPdiskUser::Create(
	const CCipherContext&	cipher, 
	const char				*userName, 
	const char				*passphrase, 
	PGPBoolean				readOnly)
{
	pgpAssertStrValid(userName);
	pgpAssertStrValid(passphrase);

	// Initialize key info structure with session key.
	Crypto::PassphraseKeyInfo	keyInfo;
	cipher.EncryptPassphraseKey(passphrase, keyInfo);

	// Import symmetric key.
	ImportRawSymmetricKey(keyInfo, userName, readOnly);
}

void 
CPGPdiskUser::Import(const void *buf, PGPUInt32 size)
{
	pgpAssertAddrValid(buf, VoidAlign);

	if (size < sizeof(OnDiskUserInfo))
		THROW_PGPERROR(kPGPError_BadParams);

	Clear();

	// First get header.
	mUserBytes.Resize(sizeof(OnDiskUserInfo));
	pgpCopyMemory(buf, mUserBytes.Get(), sizeof(OnDiskUserInfo));

	// Validate header.
	if (GetAsUserInfo()->userMagic != OnDiskUserInfo::kUserMagic)
		THROW_PGPERROR(kPGPClientError_CorruptDiskHeader);

	if (!HasPublicKey() && !HasSymmetricKey())
		THROW_PGPERROR(kPGPClientError_CorruptDiskHeader);

	PGPUInt32	userSize	= GetAsUserInfo()->userSize;

	if (userSize > OnDiskUserInfo::kMaxBytesDiskUser)
		THROW_PGPERROR(kPGPClientError_CorruptDiskHeader);

	// Copy in entire user.
	mUserBytes.Resize(userSize);
	pgpCopyMemory(buf, mUserBytes.Get(), userSize);

	if (HasPublicKey())
	{
		if (GetAsUserWithPub()->encryptedKeyOffset + 
			GetAsUserWithPub()->encryptedKeySize > userSize)
		{
			THROW_PGPERROR(kPGPClientError_CorruptDiskHeader);
		}

		if (GetAsUserWithPub()->encryptedKeyOffset < 
			sizeof(OnDiskUserWithPub))
		{
			// beta header with no subkey ID
			mUserBytes.Resize(sizeof(OnDiskUserWithPub) + 
				GetAsUserWithPub()->encryptedKeySize);

			pgpCopyMemory(
				mUserBytes.Get() + 
					GetAsUserWithPub()->encryptedKeyOffset, 
				mUserBytes.Get() + 
					sizeof(OnDiskUserWithPub), 
				GetAsUserWithPub()->encryptedKeySize);

			GetAsUserWithPub()->encryptedKeyOffset = 
				sizeof(OnDiskUserWithPub);
			GetAsUserInfo()->userSize = 
				GetAsUserWithPub()->encryptedKeyOffset + 
				GetAsUserWithPub()->encryptedKeySize;
			GetAsUserWithPub()->subKeyID = 
				GetAsUserWithPub()->keyID;

			pgpClearMemory(GetAsUserWithPub()->checkBytes.bytes, 
				sizeof(GetAsUserWithPub()->checkBytes.bytes));
		}

		mKeyID = GetAsUserWithPub()->keyID;
		mSubKeyID = GetAsUserWithPub()->subKeyID;
	}

	mIsAssigned = TRUE;
}

void 
CPGPdiskUser::Export(void *buf, PGPUInt32 availSize) const
{
	pgpAssert(IsAssigned());
	pgpAssertAddrValid(buf, VoidAlign);

	if (availSize < GetSizeForExport())
		THROW_PGPERROR(kPGPError_BufferTooSmall);

	pgpCopyMemory(mUserBytes.Get(), buf, GetSizeForExport());
}

void 
CPGPdiskUser::ImportRawSymmetricKey(
	const Crypto::PassphraseKeyInfo&	keyInfo, 
	const char							*userName, 
	PGPBoolean							readOnly)
{
	pgpAssertStrValid(userName);

	Clear();

	// Allocate on disk user bytes.
	PGPUInt32	onDiskSize	= sizeof(OnDiskUserWithSym);

	mUserBytes.Resize(onDiskSize);
	mUserBytes.Wipe();

	// Initialize fields.
	GetAsUserInfo()->userMagic	= OnDiskUserInfo::kUserMagic;
	GetAsUserInfo()->userType	= OnDiskUserInfo::kUserWithSymType;
	GetAsUserInfo()->userSize	= onDiskSize;
	GetAsUserInfo()->readOnly	= readOnly;
	GetAsUserInfo()->locked		= FALSE;

	GetAsUserWithSym()->keyInfo = keyInfo;

	UString::SmartStringCopy(GetAsUserWithSym()->userName, userName, 
		kPGPdiskMaxUserNameSize);

	mIsAssigned = TRUE;
}

void 
CPGPdiskUser::ImportRawPublicKey(
	const PGPKeyID&				keyID, 
	const PGPKeyID&				subKeyID, 
	const PGPByte				*pEncryptedKey, 
	PGPUInt32					lengthEncKey, 
	const Crypto::CheckBytes&	checkBytes, 
	PGPBoolean					readOnly, 
	PGPBoolean					locked)
{
	pgpAssertAddrValid(pEncryptedKey, PGPByte);

	Clear();

	// Allocate on disk user bytes.
	PGPUInt32	onDiskSize	= sizeof(OnDiskUserWithPub) + lengthEncKey;

	mUserBytes.Resize(onDiskSize);
	mUserBytes.Wipe();

	// Initialize fields.
	GetAsUserInfo()->userMagic	= OnDiskUserInfo::kUserMagic;
	GetAsUserInfo()->userType	= OnDiskUserInfo::kUserWithPubType;
	GetAsUserInfo()->userSize	= onDiskSize;
	GetAsUserInfo()->readOnly	= readOnly;
	GetAsUserInfo()->locked		= locked;

	GetAsUserWithPub()->keyID = keyID;
	GetAsUserWithPub()->subKeyID = subKeyID;

	GetAsUserWithPub()->encryptedKeyOffset = sizeof(OnDiskUserWithPub);
	GetAsUserWithPub()->encryptedKeySize = lengthEncKey;
	GetAsUserWithPub()->checkBytes = checkBytes;

	PGPByte	*pData	= mUserBytes.Get() + 
		GetAsUserWithPub()->encryptedKeyOffset;
	pgpCopyMemory(pEncryptedKey, pData, lengthEncKey);

	mKeyID = keyID;
	mSubKeyID = subKeyID;
	mIsAssigned = TRUE;
}

void 
CPGPdiskUser::ImportRawPublicKey(
	const PGPByte				*pKeyID, 
	PGPSize						lengthKeyID, 
	const PGPByte				*pEncryptedKey, 
	PGPSize						lengthEncKey, 
	const Crypto::CheckBytes&	checkBytes, 
 	PGPPublicKeyAlgorithm		pubAlgorithm, 
	PGPBoolean					readOnly, 
	PGPBoolean					locked)
{
	pgpAssertAddrValid(pKeyID, PGPByte);
	pgpAssertAddrValid(pEncryptedKey, PGPByte);

	Clear();

	// This overload means we are being given old-style key ID. Convert it.
	PGPByte	*pRealKeyID	= const_cast<PGPByte *>(pKeyID);
	PGPSize	realLength	= lengthKeyID;

	if ((realLength == 5) || (realLength == 9))
	{
		realLength--;
		pRealKeyID++;
	}

	CPGPKeyID	keyID;
	keyID.SetKeyIDFromBytes(pRealKeyID, realLength, pubAlgorithm);

	ImportRawPublicKey(keyID, keyID, pEncryptedKey, lengthEncKey, 
		checkBytes, readOnly, locked);
}

void 
CPGPdiskUser::Assign(const CPGPdiskUser& user)
{
	if (&user == this)
		return;

	pgpAssert(user.IsAssigned());
	Clear();

	mPContext = user.mPContext;

	CArray<PGPUInt8>	exportedUser(user.GetSizeForExport());
	user.Export(exportedUser.Get(), exportedUser.Size());

	Import(exportedUser.Get(), exportedUser.Size());
}

void 
CPGPdiskUser::Clear()
{
	if (IsAssigned() && IsCipheringEnabled())
		DisableCiphering();

	mUserBytes.Wipe();

	mPGPKeyPresent = FALSE;
	mIsAssigned = FALSE;
}

void 
CPGPdiskUser::Init()
{
	mPContext			= NULL;
	mPCipher			= NULL;
	mCipheringEnabled	= FALSE;
	mIsAssigned			= FALSE;
	mPGPKeyPresent		= FALSE;
}

void 
CPGPdiskUser::SetIsAdminUser(PGPBoolean isAdminUser)
{
	pgpAssert(IsAssigned());
	GetAsUserInfo()->adminUser = isAdminUser;
}

void 
CPGPdiskUser::ImportPublicKey() const
{
	pgpAssert(mIsAssigned);
		
	// Get actual key.
	try
	{
		mPGPKey.FindKeyByKeyID(Context()->PGPKeyDB(), mKeyID);
		mPGPSubKey.FindKeyByKeyID(Context()->PGPKeyDB(), mSubKeyID);
		mPGPKeyPresent = TRUE;
	}
	catch (CComboError&)
	{
		mPGPKeyPresent = FALSE;
	}
}

void 
CPGPdiskUser::VerifyPubKeyCheckBytes(
	const void					*compareMainData, 
	const void					*compareData, 
	const Crypto::CheckBytes&	checkBytes, 
	const CCipherContext&		cipher) const
{
	pgpAssertAddrValid(compareData, VoidAlign);
	pgpAssert(cipher.IsInitialized());

	CArray<PGPUInt8>	encBuf(cipher.GetEncryptSize());
	encBuf.Wipe();

	// if checkbytes are zero, ignore
	if (pgpMemoryEqual(checkBytes.bytes, encBuf.Get(), 
		pgpMin(encBuf.Size(), sizeof(checkBytes.bytes))))
	{
		return;
	}

	cipher.Encrypt(compareData, encBuf.Get());

	if (!pgpMemoryEqual(checkBytes.bytes, encBuf.Get(), 
		pgpMin(encBuf.Size(), sizeof(checkBytes.bytes))))
	{
		// This second try is for 6.5x PGPdisks that compare
		// their checkbytes against the fingerprint of the
		// master key for some odd reason -wjb
		encBuf.Wipe();

		cipher.Encrypt(compareMainData, encBuf.Get());

		if (!pgpMemoryEqual(checkBytes.bytes, encBuf.Get(), 
			pgpMin(encBuf.Size(), sizeof(checkBytes.bytes))))
		{
			THROW_PGPERROR(kPGPClientError_CorruptDiskHeader);
		}
	}
}

void 
CPGPdiskUser::GeneratePubKeyCheckBytes(
	const void				*compareData, 
	const CCipherContext&	cipher, 
	Crypto::CheckBytes&		checkBytes) const
{
	pgpAssertAddrValid(compareData, VoidAlign);
	pgpAssert(cipher.IsInitialized());

	CArray<PGPByte>	encBuf(cipher.GetEncryptSize());
	encBuf.Wipe();

	cipher.Encrypt(compareData, encBuf.Get());
	pgpClearMemory(checkBytes.bytes, sizeof(checkBytes.bytes));

	pgpCopyMemory(encBuf.Get(), checkBytes.bytes, pgpMin(encBuf.Size(), 
		sizeof(checkBytes.bytes)));
}

void 
CPGPdiskUser::DecryptUsingPublicKey(
	const CPGPKey&					pubMainKey,
	const CPGPKey&					pubKey,
	const char						*passphrase,
	int								passkeylen,
	const PGPByte					*pEncryptedData,
	PGPSize							encryptedDataSize,
	const Crypto::PassphraseSalt&	salt, 
	const Crypto::CheckBytes&		checkBytes, 
	CCipherContext&					cipher) const
{
	pgpAssertAddrValid(pEncryptedData, VoidAlign);
	pgpAssert(cipher.IsInitialized());

	// Initialize private key context.
	CPGPPrivateKeyContext	privContext(pubKey, passphrase, passkeylen);

	// Get operation sizes.
	PGPSize	maxDecryptedBufferSize, maxEncryptedBufferSize, maxSignatureSize;

	privContext.GetPrivateKeyOperationSizes(maxDecryptedBufferSize, 
		maxEncryptedBufferSize, maxSignatureSize);

	// Get key properties.
	CArray<PGPByte>	fingerPrint(128);
	PGPSize			fingerPrintSize;

	fingerPrint.Wipe();
		
	pubKey.GetDataProp(kPGPKeyProperty_Fingerprint, fingerPrint.Get(), 
		fingerPrint.Size(), fingerPrintSize);

	// Get Main key properties.
	CArray<PGPByte>	mainFingerPrint(128);
	PGPSize			mainFingerPrintSize;

	mainFingerPrint.Wipe();
		
	pubMainKey.GetDataProp(kPGPKeyProperty_Fingerprint, mainFingerPrint.Get(), 
		mainFingerPrint.Size(), mainFingerPrintSize);

	// Decrypt the session key.
	CSecureArray<PGPByte>	decKeyData;
	PGPUInt32				decryptedDataSize;

	decKeyData.Resize(maxDecryptedBufferSize);

	privContext.PrivateKeyDecrypt(pEncryptedData, encryptedDataSize, 
		decKeyData.Get(), decryptedDataSize);

	// Check the result.
	if (decryptedDataSize > cipher.GetKeySize())
		THROW_PGPERROR(kPGPClientError_CorruptDiskHeader);

	CSecureObject<Crypto::SymmetricKey>	pDecryptedKey;
	pgpCopyMemory(decKeyData.Get(), pDecryptedKey->bytes, decryptedDataSize);

	// Initialize cipher context.
	cipher.AttachSymmetricKey(*pDecryptedKey, salt);

	// Check check bytes.
	VerifyPubKeyCheckBytes(mainFingerPrint.Get(),
		fingerPrint.Get(), checkBytes, cipher);
}

void 
CPGPdiskUser::EncryptUsingPublicKey(
	const CPGPKey&			pubKey, 
	const CCipherContext&	cipher, 
	CSecureArray<PGPByte>&	encryptedKey, 
	PGPUInt32&				encryptedKeySize, 
	Crypto::CheckBytes&		checkBytes) const
{
	// Initialize public key context.
	CPGPPublicKeyContext	pubContext(pubKey);

	// Get operation sizes.
	PGPSize	maxDecryptedBufferSize, maxEncryptedBufferSize, maxSignatureSize;

	pubContext.GetPublicKeyOperationSizes(maxDecryptedBufferSize, 
		maxEncryptedBufferSize, maxSignatureSize);

	// Get fingerprint.
	PGPByte	fingerPrint[128];
	PGPSize	fingerPrintSize;

	pgpClearMemory(fingerPrint, sizeof(fingerPrint));			
		
	pubKey.GetDataProp(kPGPKeyProperty_Fingerprint, fingerPrint, 
		sizeof(fingerPrint), fingerPrintSize);

	// Get session key.
	CSecureObject<Crypto::SymmetricKey>	pDecryptedKey;
	cipher.GetSymmetricKey(*pDecryptedKey);

	// Encrypt the session key.
	pgpAssert(cipher.GetKeySize() < maxDecryptedBufferSize - 10);

	encryptedKey.Resize(maxEncryptedBufferSize);

	pubContext.PublicKeyEncrypt(pDecryptedKey, cipher.GetKeySize(), 
		encryptedKey.Get(), encryptedKeySize);

	// Compute the check bytes.
	GeneratePubKeyCheckBytes(fingerPrint, cipher, checkBytes);
}
