/*____________________________________________________________________________
	Copyright (C) 2002 PGP Corporation
	All rights reserved.

	$Id: pgpNetLog.c,v 1.17 2002/08/06 20:10:25 dallen Exp $
____________________________________________________________________________*/
#include <stdio.h>
#include <string.h>

#include "pgpMemoryMgr.h"
#include "pgpMem.h"
#include "pgpPFLErrors.h"
#include "pgpFileUtilities.h"
#include "pgpEndianConversion.h"
#include "pgpUtilities.h"
#include "pgpPFLPriv.h"
#include "pgpStdFileIOPriv.h"

#if PGP_OSX
#	include "PGPnetDriverComm.h"
#endif

#include "pgpNetLog.h"

#if PGP_WIN32
#include "pgpNetLogUtils.h"
#endif

#if PGP_OSX
#include <sys/types.h>
#include <sys/stat.h>
#endif

typedef struct PGPnetLog	PGPnetLog;

struct PGPnetLog 
{
	PGPMemoryMgrRef	memoryMgr;
	PFLFileSpecRef	logFileSpec;
	PGPIORef		logIO;

	PGPUInt32		recLength;
	PGPUInt32		numRecords;
	PGPUInt32		dataLength;
	PGPUInt32		fileLength;
	PGPUInt32		currentFilePosition;

#if PGP_WIN32 || PGP_OSX
	/* pointer to buffer which we assume will contain the path of folder 
	   for event data files.  May not have been filled in at time of call
	   to PGPnetOpenLog  */
	char*			pszEventDirPath;
#endif
};


#define kPGPNetLogHeaderLength		sizeof(PGPnetLogFileHeader)
#define PGPValidateLog(logRef)		PGPValidateParam(logRef != NULL)

static PGPError sFreeLog(PGPnetLogRef logRef);
static PGPError sIOWriteEvent(PGPIORef fileIO, PGPnetLogEvent* pevent);
static PGPError sIOReadEvent(PGPIORef fileIO, PGPnetLogEvent* pevent);
static PGPError sDeleteOldEventFile(PGPnetLogRef logRef);
static PGPError sCheckForCurrentLogFile(PGPnetLogRef logRef, PGPBoolean	bWriteAccess);

/*____________________________________________________________________________

  Releases memory allocated for the PGPnetLog structure.
  ____________________________________________________________________________*/

static PGPError 
sFreeLog (
		PGPnetLogRef	logRef)
{
	PGPValidateLog (logRef);

	if (logRef)
	{
		if (pflFileSpecIsValid (logRef->logFileSpec))
			PFLFreeFileSpec (logRef->logFileSpec);

		if (PGPIOIsValid (logRef->logIO))
			PGPFreeIO (logRef->logIO);

		PGPFreeData (logRef);
		return kPGPError_NoErr;
	}
	else
		return kPGPError_BadParams;
}

/*____________________________________________________________________________

  Writes a PGPnetLogEvent to the filespec associated with fileIO.
  ____________________________________________________________________________*/

static PGPError 
sIOWriteEvent (
		PGPIORef		fileIO, 
		PGPnetLogEvent* pevent)
{
	PGPError		err = kPGPError_NoErr;
	PGPnetLogEvent	event;

	event.timeOfEvent	= pevent->timeOfEvent;
	event.typeOfEvent	= pevent->typeOfEvent;
	event.ipaddress		= pevent->ipaddress;

	pgpCopyMemory(pevent->eventDataFile, event.eventDataFile, sizeof(event.eventDataFile));
	pgpCopyMemory (pevent->info.buffer, event.info.buffer, sizeof(event.info));

	err = PGPIOWrite(fileIO, sizeof(PGPnetLogEvent), &event);

	return err;
}

/*____________________________________________________________________________

  Reads a PGPnetLogEvent from the filespec associated with fileIO.
  ____________________________________________________________________________*/

static PGPError 
sIOReadEvent (
		PGPIORef		fileIO, 
		PGPnetLogEvent* pevent)
{
	PGPnetLogEvent	event;
	PGPError		err;
	PGPSize			bytesRead;

	PGPValidateIO (fileIO);
	PGPValidatePtr (pevent);

	err = PGPIORead (fileIO, sizeof(PGPnetLogEvent), &event, &bytesRead);

	if (IsntPGPError (err))
	{
		pevent->timeOfEvent = event.timeOfEvent;
		pevent->typeOfEvent = event.typeOfEvent;
		pevent->ipaddress	= event.ipaddress;

		pgpCopyMemory(event.eventDataFile, pevent->eventDataFile, sizeof(event.eventDataFile));
		pgpCopyMemory (event.info.buffer, pevent->info.buffer, sizeof(event.info));
	}
	return err;
}

/*____________________________________________________________________________

  Reads an event from the log, and deletes the file associated with the event.

  If a log event makes reference to additional data (like a packet capture),
  the data is usually saved to disk for later retrieval. When log events are
  about to be overwritten, it's probably a good idea to delete their
  associated files, so that the disk doesn't become cluttered.
  ____________________________________________________________________________*/

static PGPError
sDeleteOldEventFile (
		PGPnetLogRef	logRef)
{
	PGPError		err = kPGPError_NoErr;
	PGPnetLogEvent	previousEvent;
	PGPMemoryMgrRef	memoryMgr;
#if PGP_MACINTOSH
	PFLFileSpecRef	oldEventFileSpec;
	FSSpec			oldEventFileFSSpec;
	OSStatus		macErr;
#elif PGP_WIN32 || PGP_OSX
	PFLFileSpecRef	oldEventFileSpec;
	PGPSize			requestSize = 0;
	char*			oldEventFile = NULL;
	char*			pszEventDir = NULL;
#if PGP_WIN32
	char			szEventDir[MAX_PATH];
#endif /* PGP_WIN32 */
	PGPUInt32		fileLen;
#endif /* PGP_WIN32 || PGP_OSX */

	/*
	Read an event from the log. If it contains valid (non-zero)
	eventDataFile fields, attempt to delete the file specified.
	*/
	err = sIOReadEvent(logRef->logIO, &previousEvent);
	if (IsPGPError(err))
		return err;
	
	memoryMgr = PGPIOGetMemoryMgr(logRef->logIO);
	err = PGPValidateMemoryMgr(memoryMgr);
		
	if (IsPGPError(err))
		return err;

	if (previousEvent.eventDataFile[0])
#if PGP_WIN32 || PGP_OSX
	{	

#if PGP_WIN32
		if (!logRef->pszEventDirPath)
		{
			err = pgpGetEventDirFullPath(szEventDir, sizeof(szEventDir));

			pszEventDir = szEventDir;

			if (IsPGPError(err))
				return err;
		}
#endif /* PGP_WIN32 */

		if (!pszEventDir)
			pszEventDir = logRef->pszEventDirPath;

		fileLen		= strlen(previousEvent.eventDataFile);
		requestSize	= strlen(pszEventDir) + fileLen + 1;

		oldEventFile = (char*) PGPNewData(memoryMgr, 
				requestSize, kPGPMemoryMgrFlags_Clear);

		if (oldEventFile)
		{
			strncpy(oldEventFile, pszEventDir, requestSize);
			strncat(oldEventFile, previousEvent.eventDataFile, fileLen);
			/* Determine if the file exists. If so, attempt to remove it */
			err = PFLNewFileSpecFromFullPath(memoryMgr, oldEventFile, &oldEventFileSpec);
			if (IsntPGPError(err))
			{
				PFLFileSpecDelete(oldEventFileSpec);
				PFLFreeFileSpec(oldEventFileSpec);
			}
			PGPFreeData(oldEventFile);
		}
		else
			return kPGPError_OutOfMemory;
	}
#elif PGP_MACINTOSH		/* Mac stuff */
	{
		if (previousEvent.eventDataFile[0])
		{
			macErr = FindPGPnetLogEventDataFolder(kOnSystemDisk,
												  &oldEventFileFSSpec.vRefNum,
												  &oldEventFileFSSpec.parID);
			/*
			FSMakeFSSpec returning anything other than noErr indicates that the
			file does NOT exist.
			*/
			macErr = FSMakeFSSpec(oldEventFileFSSpec.vRefNum,
								  oldEventFileFSSpec.parID,
								  previousEvent.eventDataFile,
								  &oldEventFileFSSpec);
			if (macErr == noErr)
			{
				err = PFLNewFileSpecFromFSSpec(memoryMgr,
											   &oldEventFileFSSpec,
											   &oldEventFileSpec);
				if (IsntPGPError(err))
					err = PFLFileSpecDelete(oldEventFileSpec);
					
				PFLFreeFileSpec(oldEventFileSpec);
			}
		}
	}
#else		/* Linux stuff */
	{
		/* Nothing to do */
	}
#endif

	return err;
}


PGPError 
PGPnetFlushLogBuffers (
		PGPnetLogRef 	logRef)
{
	PGPValidateLog (logRef);

	if (logRef)
		return PGPIOFlush (logRef->logIO);
	else
		return kPGPError_BadParams;
}



PGPError 
PGPnetClearLogFile (
		PGPnetLogRef 	logRef)
{
	PGPnetLogEvent	event;
	PGPUInt32		u;
	PGPIORef		newIORef;
	PGPError		err = kPGPError_NoErr;

	PGPValidateLog (logRef);

	if (logRef)
	{
		PGPFreeIO(logRef->logIO);
		err = PGPOpenFileSpec(logRef->logFileSpec,
					kPFLFileOpenFlags_ReadWrite,
					(PGPFileIORef*)&newIORef);
		if (IsntPGPError(err))
		{
			/* Free the old PGPIORef, and point logRef->logIO to the new one */
			logRef->logIO = newIORef;
		}

		PGPIOSetPos (logRef->logIO, kPGPNetLogHeaderLength);
		pgpClearMemory (&event, sizeof(PGPnetLogEvent));
		for (u=0; u<logRef->numRecords; u++)
		{
			PGPIOWrite (logRef->logIO, sizeof(PGPnetLogEvent), &event);
		}

		PGPIOSetPos (logRef->logIO, kPGPNetLogHeaderLength);
		logRef->currentFilePosition	= kPGPNetLogHeaderLength;

		return PGPIOFlush (logRef->logIO);
	}
	else
		return kPGPError_BadParams;
}


static PGPError
sCheckForCurrentLogFile (
		PGPnetLogRef	logRef,
		PGPBoolean		bWriteAccess)
{
	PGPBoolean			bFileExists;
	PGPError			err;
	PGPIORef			logIO;
	PGPnetLogFileHeader	header;
	PGPSize				bytesRead;

	/* first check if file exists */
	err = PFLFileSpecExists (logRef->logFileSpec, &bFileExists);
	if (IsntPGPError (err) && !bFileExists)
		err = kPGPError_FileNotFound;

	if (IsPGPError (err))
		return err;

	/* now check if file header corresponds */
	err = PGPOpenFileSpec(	logRef->logFileSpec, 
							kPFLFileOpenFlags_ReadOnly, 
							(PGPFileIORef*)&logIO);

	if (IsntPGPError (err))
	{
		err = PGPIORead (logIO, 
				sizeof(PGPnetLogFileHeader), &header, &bytesRead);

		if (IsntPGPError (err) && (bytesRead != sizeof(PGPnetLogFileHeader)))
			err = kPGPError_CantOpenFile;

		if (IsntPGPError (err))
		{
			if( header.logfileFormatVersion != kPGPnetLogFileFormatVersion)
				err = kPGPError_CantOpenFile;

			if( header.recordLengthBytes != logRef->recLength)
				err = kPGPError_CantOpenFile;

			/* if opening for write access, then the file must have the
			   the expected number of records */
			if (bWriteAccess)
			{
				if( header.fileLengthRecords != logRef->numRecords )
					err = kPGPError_CantOpenFile;
			}
			/* for read access, just use the number of records specified
			   in the header */
			else
			{
				logRef->numRecords = header.fileLengthRecords;
				logRef->dataLength = logRef->numRecords * 
										logRef->recLength;
				logRef->fileLength = logRef->dataLength +
										kPGPNetLogHeaderLength;
			}
		}

		PGPFreeIO (logIO);
	}

	return err;
}


PGPError 
PGPnetOpenLogFile (
		PFLFileSpecRef	logFileSpec, 
		PGPBoolean		bWriteAccess,
		PGPBoolean		bZeroOutFile,
#if PGP_WIN32 || PGP_OSX
		char*			eventfolder,
		PGPUInt32		eventDirLen,
#endif
		PGPnetLogRef*	plogRef)
{
	PGPMemoryMgrRef	memoryMgr	= NULL;
	PGPError		err;
	PGPnetLogEvent	event;

	if (IsntNull (plogRef))
		*plogRef = NULL;

	PGPValidatePtr (plogRef);
	PFLValidateFileSpec (logFileSpec);

	memoryMgr = PFLGetFileSpecMemoryMgr (logFileSpec);
	
	pgpAssert (PGPMemoryMgrRefIsValid (memoryMgr));
	if (!PGPMemoryMgrRefIsValid (memoryMgr))
		return kPGPError_BadParams;

	*plogRef = (PGPnetLogRef) PGPNewData (memoryMgr, 
								sizeof(PGPnetLog), 
								kPGPMemoryMgrFlags_Clear);

	pgpAssert (IsntNull (*plogRef));
	if (IsNull (*plogRef))
		return kPGPError_OutOfMemory;

	(*plogRef)->memoryMgr			= memoryMgr;
	(*plogRef)->logFileSpec			= NULL;
	(*plogRef)->logIO				= NULL;
	(*plogRef)->recLength			= sizeof(PGPnetLogEvent);

	/* to be changed? */
	(*plogRef)->numRecords			= kPGPnetLogMaxNumberOfEvents;

	(*plogRef)->dataLength			= (*plogRef)->numRecords * 
										(*plogRef)->recLength;

	(*plogRef)->fileLength			= (*plogRef)->dataLength +
										kPGPNetLogHeaderLength;

	(*plogRef)->currentFilePosition	= 0;

	err = PFLCopyFileSpec (logFileSpec, &((*plogRef)->logFileSpec));
	if (IsPGPError (err))
	{
		sFreeLog (*plogRef);
		*plogRef = NULL;
		return err;
	}

#if PGP_WIN32
	(*plogRef)->pszEventDirPath = eventfolder;

	err = pgpGetEventDirFullPath((*plogRef)->pszEventDirPath, eventDirLen);
#elif PGP_OSX
	(*plogRef)->pszEventDirPath = eventfolder;
#endif

	/* Create the file if it does not exist or 
	   if it is a different version and we have write access */
	err = sCheckForCurrentLogFile (*plogRef, bWriteAccess);
	if ((err == kPGPError_FileNotFound) ||
		(IsPGPError (err) && bWriteAccess))
	{
		PGPnetLogFileHeader header;

		if( err == kPGPError_FileNotFound )
		{
			PFLFileSpecCreate ((*plogRef)->logFileSpec);
#if PGP_OSX
			{
				/* Change the permissions to rw-r--r-- so non-root can at least *read* the file */
				char *		logFileName		= NULL;

				if( IsntPGPError( PFLGetFullPathFromFileSpec( (*plogRef)->logFileSpec, &logFileName ) ) )
					chmod( logFileName, 0644 );

				if( IsntNull( logFileName ) )
					PGPFreeData( logFileName );
			}
#endif
		}

		err = PGPSetFileSize ((*plogRef)->logFileSpec, 
								(*plogRef)->fileLength);

		if (IsntPGPError (err))
		{

			err = PGPOpenFileSpec((*plogRef)->logFileSpec, 
						kPFLFileOpenFlags_ReadWrite, 
						(PGPFileIORef*)(&((*plogRef)->logIO)));

			if (IsntPGPError (err))
			{
				/* initialize header and write it */
				pgpClearMemory (&header, sizeof(header));

				header.logfileFormatVersion = kPGPnetLogFileFormatVersion;
				header.recordLengthBytes = sizeof(PGPnetLogEvent);
				header.fileLengthRecords = (*plogRef)->numRecords;

				PGPIOSetPos ((*plogRef)->logIO, 0);

				err = PGPIOWrite ((*plogRef)->logIO, 
									kPGPNetLogHeaderLength, &header);

				/* we will need to wipe the entire file */
				bZeroOutFile = TRUE;
			}
		}

		PGPFreeIO ((*plogRef)->logIO);
	}
	
	pgpAssertNoErr (err);
	if (IsPGPError (err))
	{
		sFreeLog (*plogRef);
		*plogRef = NULL;
		return err;
	}


	/* open the file */
	if (bWriteAccess)
		err = PGPOpenFileSpec(	(*plogRef)->logFileSpec, 
							kPFLFileOpenFlags_ReadWrite, 
							(PGPFileIORef*)(&((*plogRef)->logIO)));
	else
		err = PGPOpenFileSpec(	(*plogRef)->logFileSpec, 
							kPFLFileOpenFlags_ReadOnly, 
							(PGPFileIORef*)(&((*plogRef)->logIO)));

	if (IsPGPError(err))
	{
		sFreeLog (*plogRef);
		*plogRef = NULL;
		return err;
	}
	
	pgpAssert (PGPIOIsValid ((*plogRef)->logIO));
	if (!PGPIOIsValid ((*plogRef)->logIO))
	{
		sFreeLog (*plogRef);
		*plogRef = NULL;
		return kPGPError_OutOfMemory;
	}

	/* zero-out file */
	if (bZeroOutFile)
		PGPnetClearLogFile (*plogRef);

	/* otherwise find starting position */
	else
	{
		PGPUInt32		currPos				= kPGPNetLogHeaderLength;
		PGPTime			timeOfPrevEvent;

		PGPIOSetPos ((*plogRef)->logIO, kPGPNetLogHeaderLength);

		event.timeOfEvent = 0;
		do {
			timeOfPrevEvent = event.timeOfEvent;

			err = sIOReadEvent ((*plogRef)->logIO, &event);
			if (IsPGPError (err))
				return err;
			currPos += sizeof(PGPnetLogEvent);

		} while ((event.timeOfEvent != 0) &&
				 (event.timeOfEvent >= timeOfPrevEvent) &&
				 (currPos < (*plogRef)->fileLength));

		if (currPos >= (*plogRef)->fileLength)
			currPos = kPGPNetLogHeaderLength;
		else
			currPos -= sizeof(PGPnetLogEvent);

		PGPIOSetPos ((*plogRef)->logIO, currPos);
		(*plogRef)->currentFilePosition	= currPos;
	}

	return err;
}

/*____________________________________________________________________________

  Writes a PGPnetLogEvent to the event log.

  Before writing the new event to disk, we read in the previous entry, attempt
  to perform any necessary garbage collection, then, if everything is cool, we
  write the new event to disk.
  ____________________________________________________________________________*/

PGPError 
PGPnetLogEventToFile (
		PGPnetLogRef 	logRef,
		PGPnetLogEvent*	pevent)
{
	PGPError		err;
	PGPFileOffset	currPos;
	PGPIORef		newIORef;

	PGPValidateLog (logRef);
	PGPValidatePtr (pevent);

	if (logRef)
	{
		if (pevent->timeOfEvent == 0)
			pevent->timeOfEvent = PGPGetTime ();

		/*
		sDeleteOldEventFile will read from the log, and delete any lingering
		files associated with the about-to-be-overwritten event. Before calling
		sDeleteOldEventFile, we make note of the current file position so that
		we can seek back to it after sDeleteOldEventFile is done.
		*/
		PGPIOGetPos(logRef->logIO, &currPos);
		err = sDeleteOldEventFile(logRef);
		if (IsntPGPError(err))
		{
			/*
			Note: This is a workaround for an irritating nuance in Microsoft's
			stdio fread/fwrite implementation. The symptoms are that after
			doing an fread from a file, the internal (FILE*)->_flag gets |= 
			with _IOREAD. In some cases, this doesn't pose a problem for fwrite,
			but in most of our circumstances, fwrite calls flsbuf. flsbuf makes
			the assumption that if _IOREAD is set (without looking at _IORW),
			that there is an error.

			This can easily be hacked with (FILE*)->_flag &= ~_IOREAD, but the
			more platform independent solution seems to be to reopen the file.
			*/
			PGPFreeIO(logRef->logIO);
			err = PGPOpenFileSpec(logRef->logFileSpec,
						kPFLFileOpenFlags_ReadWrite,
						(PGPFileIORef*)&newIORef);
			if (IsntPGPError(err))
			{
				/* Free the old PGPIORef, and point logRef->logIO to the new one */
				logRef->logIO = newIORef;
				PGPIOSetPos(logRef->logIO, currPos);
				err = sIOWriteEvent (logRef->logIO, pevent);
				PGPIOFlush(logRef->logIO);
				if (IsPGPError (err))
					return err;
				logRef->currentFilePosition += sizeof(PGPnetLogEvent);
				if (logRef->currentFilePosition >= logRef->fileLength)
				{
					logRef->currentFilePosition = kPGPNetLogHeaderLength;
					PGPIOSetPos (logRef->logIO, kPGPNetLogHeaderLength);
				}
			}
		}
		return err;
	}
	else 
		return kPGPError_BadParams;
}


PGPError 
PGPnetCloseLogFile (
		PGPnetLogRef	logRef)
{
	PGPValidateLog (logRef);

	if (logRef)
	{
		sFreeLog (logRef);
		return kPGPError_NoErr;
	}
	else
		return kPGPError_BadParams;
}



PGPError 
PGPnetReadLogData (
		PGPnetLogRef	logRef, 
		PGPSize*		pnumberEvents, 
		void**			ppEventBuffer)
{
	PGPUInt32		currPos			= kPGPNetLogHeaderLength;
	PGPUInt32		startPos		= kPGPNetLogHeaderLength;
	PGPUInt32		endPos			= kPGPNetLogHeaderLength;
	PGPUInt32		bufferPos		= 0;

	PGPSize			sizeBuffer;
	PGPByte*		pBuffer;
	PGPnetLogEvent	event;
	PGPTime			timeOfPrevEvent;
	PGPError		err;

	PGPValidatePtr (pnumberEvents);
	PGPValidatePtr (ppEventBuffer);

	*pnumberEvents = 0;
	*ppEventBuffer = NULL;

	PGPValidateLog (logRef);

	if (!logRef)
		return kPGPError_BadParams;

	PGPIOSetPos (logRef->logIO, kPGPNetLogHeaderLength);

	/* search for decreasing time, marking end of log data */
	event.timeOfEvent = 0;
	do {
		timeOfPrevEvent = event.timeOfEvent;

		err = sIOReadEvent (logRef->logIO, &event);
		if (IsPGPError (err))
			return err;
		currPos += sizeof(PGPnetLogEvent);

	} while ((event.timeOfEvent != 0) &&
			 (event.timeOfEvent >= timeOfPrevEvent) &&
			 (currPos < logRef->fileLength));

	if (event.timeOfEvent == 0)
	{
		startPos = kPGPNetLogHeaderLength;
		endPos = currPos;
	}
	else if (event.timeOfEvent < timeOfPrevEvent)
	{
		currPos -= sizeof(PGPnetLogEvent);
		startPos = currPos;
		endPos = currPos;
	}
	else 
	{
		startPos = kPGPNetLogHeaderLength;
		endPos = kPGPNetLogHeaderLength;
	}

	/* calculate size of buffer needed and allocate it */
	if (startPos < endPos)
		sizeBuffer = endPos - startPos - sizeof(PGPnetLogEvent);
	else
		sizeBuffer = logRef->numRecords * sizeof(PGPnetLogEvent);

	if (sizeBuffer == 0)
		return kPGPError_NoErr;

	pBuffer = (PGPByte *)PGPNewData (logRef->memoryMgr, 
								sizeBuffer, 
								kPGPMemoryMgrFlags_Clear);

	/* read the data from the file */
	PGPIOSetPos (logRef->logIO, startPos);
	currPos = startPos;

	while (bufferPos < sizeBuffer)
	{
		err = sIOReadEvent (logRef->logIO, 
						(PGPnetLogEvent*)&pBuffer[bufferPos]);
		if (IsPGPError (err))
			return err;
		(*pnumberEvents)++;
		bufferPos += sizeof(PGPnetLogEvent);
		currPos += sizeof(PGPnetLogEvent);
		if (currPos >= logRef->fileLength) 
		{
			PGPIOSetPos (logRef->logIO, kPGPNetLogHeaderLength);
			currPos = kPGPNetLogHeaderLength;
		}
	}

	*ppEventBuffer = pBuffer;

	return kPGPError_NoErr;
}


PGPError 
PGPnetFreeLogData (
		void*	pData)
{
	PGPValidatePtr (pData);

	PGPFreeData (pData);
	return kPGPError_NoErr;
}


#if PGP_MACINTOSH

#include "MacStrings.h"

// Called just like FindFolder()
// except the folder is always created if not present.

	OSStatus
FindPGPnetLogEventDataFolder(
	short 	vRefNum,
	short 	*foundVRefNum,
	long 	*foundDirID)
{
	OSStatus	err;
	FSSpec		folderSpec;
	
	pgpAssertAddrValid( foundVRefNum, short );
	pgpAssertAddrValid( foundDirID, long );
	
	err = FindPGPPreferencesFolder( vRefNum, &folderSpec.vRefNum, &folderSpec.parID );
	if( IsntErr( err ) )
	{
		CInfoPBRec	cpb;
		
		*foundVRefNum = folderSpec.vRefNum;
		
		// This string should _not_ be internationalized.
		CopyPString( "\pPGPnet Log Event Data", folderSpec.name );
		
		err = FSpGetCatInfo( &folderSpec, &cpb );
		if( IsntErr( err ) )
		{
			if( cpbIsFolder( &cpb ) )
			{
				*foundDirID = cpb.dirInfo.ioDrDirID;
			}
			else
			{
				pgpDebugMsg( "PGPnet Log Event Data is a file!" );
				err = dupFNErr;
			}
		}
		else
		{
			err = FSpDirCreate( &folderSpec, smSystemScript, foundDirID );
		}
	}

	return( err );
}
#endif


/*__Editor_settings____

	Local Variables:
	tab-width: 4
	End:
	vi: ts=4 sw=4
	vim: si
_____________________*/

