/*____________________________________________________________________________
    Copyright (C) 2002 PGP Corporation
    All rights reserved.

    $Id: UAppleEvents.cp,v 1.2 2002/08/06 20:10:27 dallen Exp $
____________________________________________________________________________*/
#include "UAppleEvents.h"

// ===========================================================================
//	StAEDescriptor Class
// ===========================================================================
//
//	StAEDescriptor class encapsulates an Apple Event Descriptor (AEDesc).
//
//	The Constructor creates an empty AEDesc. GetParamDesc and
//	GetOptionalParamDesc load information from an Apple Event into
//	the AEDesc. The Destructor disposes of the AEDesc.
//
//	The public member mDesc is an AEDesc record which you can pass
//	to AppleEvent routines or other functions.
//
//	Use StAEDescriptors as stack-based objects to ensure that Descriptors
//	are properly disposed, even if an exception occurs. For example,
//
//		RADescriptor	theDesc;
//		theDesc.GetParamDesc(myAppleEvent, myKeyword, typeQDRectangle);
//		UExtractFromAEDesc::TheRect(theDesc.mDesc);
//
//	This class has constructors for creating Descriptors for the following
//	common data types:
//		Boolean, Short Integer, Long Interger, OSType, Pascal String

// ---------------------------------------------------------------------------
//	 StAEDescriptor						Default Constructor
// ---------------------------------------------------------------------------

StAEDescriptor::StAEDescriptor()
{
	InitToNull();
}


//	Warning:
//
//	This constructor will NOT fail if the indicated parameter does
//	not exist.  Instead, the StAEDescriptor will be initialized to typeNull.
//	This way, the caller can explicitly check to see if the indicated
//	parameter was non-existent (typeNull), or just fail when trying to
//	get data out of the typeNull descriptor.

StAEDescriptor::StAEDescriptor(
	const AEDesc&	inDesc,
	AEKeyword		inKeyword,
	DescType		inDesiredType)
{
	mDesc.descriptorType = typeNull;
	mDesc.dataHandle     = nil;
	
	OSErr	err;
	
	switch (inDesc.descriptorType) {

		case typeAERecord:
		case typeAppleEvent: {
			err = ::AEGetParamDesc(&inDesc, inKeyword, inDesiredType, &mDesc);
			//	Don't throw.
			//	The typeNull value assigned above is sufficient.
			break;
		}
		
		case typeNull:
			break;
		
		default: {
			StAEDescriptor	temp;
			
			::AECoerceDesc(&inDesc, typeAERecord, &temp.mDesc);
			
			err = ::AEGetParamDesc(&temp.mDesc, inKeyword, inDesiredType, &mDesc);
			
			//	Don't throw if the descriptor doesn't exist.
			//	The typeNull value assigned above is sufficient.
			if (err != errAEDescNotFound)
			{}
			break;
		}
	}
}


StAEDescriptor::StAEDescriptor(
	DescType	inType,
	const void	*inData,
	SInt32		inSize)
{
	InitToNull();
	
	::AECreateDesc(inType, inData, inSize, &mDesc);
}


StAEDescriptor::StAEDescriptor(
	Boolean		inBoolean)
{
	InitToNull();
	Assign(inBoolean);
}


StAEDescriptor::StAEDescriptor(
	SInt16		inInt16)
{
	InitToNull();
	Assign(inInt16);
}


StAEDescriptor::StAEDescriptor(
	SInt32		inInt32)
{
	InitToNull();
	Assign(inInt32);
}


StAEDescriptor::StAEDescriptor(
	OSType		inOSType)
{
	InitToNull();
	Assign(inOSType);
}


StAEDescriptor::StAEDescriptor(
	ConstStringPtr	inString)
{
	InitToNull();
	Assign(inString);
}


StAEDescriptor::StAEDescriptor(
	const FSSpec&	inFileSpec)
{
	InitToNull();
	Assign(inFileSpec);
}


// ---------------------------------------------------------------------------
//	 ~StAEDescriptor						Destructor				  [public]
// ---------------------------------------------------------------------------

StAEDescriptor::~StAEDescriptor()
{
	if (mDesc.dataHandle != nil) {
		::AEDisposeDesc(&mDesc);
	}
}


StAEDescriptor&
StAEDescriptor::Assign(
	Boolean		inBoolean)
{
	Dispose();
	
	::AECreateDesc(typeBoolean, &inBoolean, sizeof(inBoolean), &mDesc);

	return *this;
}


StAEDescriptor&
StAEDescriptor::Assign(
	SInt16		inInt16)
{
	Dispose();
	
	::AECreateDesc(typeShortInteger, &inInt16, sizeof(inInt16), &mDesc);

	return *this;
}


StAEDescriptor&
StAEDescriptor::Assign(
	SInt32		inInt32)
{
	Dispose();
	
	::AECreateDesc(typeLongInteger, &inInt32, sizeof(inInt32), &mDesc);

	return *this;
}


StAEDescriptor&
StAEDescriptor::Assign(
	OSType		inOSType)
{
	Dispose();
	
	::AECreateDesc(typeType, &inOSType, sizeof(inOSType), &mDesc);

	return *this;
}


StAEDescriptor&
StAEDescriptor::Assign(
	ConstStringPtr	inString)
{
	Dispose();
	
	::AECreateDesc(typeChar, inString + 1, inString[0], &mDesc);

	return *this;
}


StAEDescriptor&
StAEDescriptor::Assign(
	const FSSpec&	inFileSpec)
{
	Dispose();
	
	::AECreateDesc(typeFSS, &inFileSpec, sizeof(FSSpec), &mDesc);

	return *this;
}


// ---------------------------------------------------------------------------
//	 Dispose
// ---------------------------------------------------------------------------
//	Dispose AEDesc and reset it to a null descriptor

void
StAEDescriptor::Dispose()
{
	if (mDesc.dataHandle != nil) {
		::AEDisposeDesc(&mDesc);
		InitToNull();
	}
}


// ---------------------------------------------------------------------------
//	 GetParamDesc
// ---------------------------------------------------------------------------
//	Load Descriptor data from an AppleEvent
//
//	Throws an OSErr exception if it can't get the data

void
StAEDescriptor::GetParamDesc(
	const AppleEvent&	inAppleEvent,
	AEKeyword			inKeyword,
	DescType			inDesiredType)
{
	Dispose();
	
	::AEGetParamDesc(&inAppleEvent, inKeyword, inDesiredType, &mDesc);
}

// ---------------------------------------------------------------------------
//	 GetOptionalParamDesc
// ---------------------------------------------------------------------------
//	Load optional Descriptor data from an AppleEvent
//
//	Differs from GetParamDesc in that it does not throw an exception
//	if the request fails because the specified keyword data does not
//	exist. Use this function to extract optional parameters from
//	an AppleEvent.

void
StAEDescriptor::GetOptionalParamDesc(
	const AppleEvent&	inAppleEvent,
	AEKeyword			inKeyword,
	DescType			inDesiredType)
{
	Dispose();
	
	OSErr	err = ::AEGetParamDesc(&inAppleEvent, inKeyword, inDesiredType,
									&mDesc);
	if ((err != errAEDescNotFound) && (err != noErr)) {
	}
}

