/*____________________________________________________________________________
	Copyright (C) 2002 PGP Corporation
	All rights reserved.

	$Id: pgpECKey.c,v 1.17 2002/10/24 00:22:00 ajivsov Exp $
____________________________________________________________________________*/
#include "pgpSDKBuildFlags.h"

/* This entire module is dependent on EC being enabled */
#if PGP_EC

#include "pgpConfig.h"
#include <string.h>
#include <stddef.h>

#include "pgpDebug.h"
#include "pgpKeyMisc.h"
#include "pgpECKey.h"
#include "bn.h"
#include "ec.h"
#include "pgpCFBPriv.h"
#include "pgpSymmetricCipherPriv.h"
#include "pgpHashPriv.h"
#include "pgpMem.h"
#include "pgpErrors.h"
#include "pgpPubKey.h"
#include "pgpRandomX9_17.h"
#include "pgpStr2Key.h"
#include "pgpContext.h"
#include "pgpEnv.h"
#include "pgpEndianConversion.h"
#include "pgpFIPSPriv.h"



#define ASSERTEC(alg) pgpAssert( \
			      (ALGMASK(alg))==kPGPPublicKeyAlgorithm_ECEncrypt || \
			      (ALGMASK(alg))==kPGPPublicKeyAlgorithm_ECSign)
#define ASSERTECSIG(alg) \
		pgpAssert((ALGMASK(alg))==kPGPPublicKeyAlgorithm_ECSign)
#define ASSERTECENC(alg) \
		pgpAssert((ALGMASK(alg))==kPGPPublicKeyAlgorithm_ECEncrypt)


/* Size of encryption mask buffer - big enough to hold 256 bit keys */
#define MASKBUF_SIZE	40

/* Defines for ecScalars and BigNums */
#define SECURE		1
#define INSECURE	0


/* A structure to hold a public key */
typedef struct ECpub
{
	ecContextRef	ec;			/* Context */
	ecPointRef		y;			/* Public point */
	ecPointRef		g;			/* Generator */
	ecScalarRef		order;		/* Order of generator */
	PGPUInt32		cofactor;	/* # curve points / order */
	DEBUG_STRUCT_CONSTRUCTOR( ECpub )
} ECpub;

/* A structure to hold a secret key */
typedef struct ECsec
{
	ecContextRef	ec;			/* Context */
	ecPointRef		y;			/* Public point */
	ecPointRef		g;			/* Generator */
	ecScalarRef		order;		/* Order of generator */
	PGPUInt32		cofactor;	/* # curve points / order */
	ecScalarRef		x;			/* Private exponent */
	DEBUG_STRUCT_CONSTRUCTOR( ECsec )
} ECsec ;




/* A PGPSecKey's priv points to this, an ECsec plus the encrypted form... */
/* This struct is always allocated using PGPNewSecureData */
typedef struct ECsecPlus
{
	PGPContextRef	context;
	
	ECsec			s;
	PGPByte *		cryptkey;
	PGPSize			ckalloc;
	PGPSize			cklen;
	int				locked;
	DEBUG_STRUCT_CONSTRUCTOR( ECsecPlus )
} ECsecPlus ;



/** Predefined curves **/

#define BITSIZE1	163
#define NAME1		"sect163r2"
#define COFACTOR1	2
static PGPUInt32 nistb1[] = {
 0x4a3205fd, 0x512f7874, 0x1481eb10, 0xb8c953ca, 0x0a601907, 0x02
};
static PGPByte nistg1[] = {
    0x03, 0x03, 0xF0, 0xEB, 0xA1, 0x62,
	0x86, 0xA2, 0xD5, 0x7E, 0xA0, 0x99, 0x11, 0x68,
	0xD4, 0x99, 0x46, 0x37, 0xE8, 0x34, 0x3E, 0x36
};
static PGPByte nisto1[] = {
    0x04, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0x92, 0xFE,
	0x77, 0xE7, 0x0C, 0x12, 0xA4, 0x23, 0x4C, 0x33
};

#define BITSIZE2	283
#define NAME2		"sect283r1"
#define COFACTOR2	2
static PGPUInt32 nistb2[] = {
	0x3B79A2F5,
	0xF6263E31, 0xA581485A, 0x45309FA2, 0xCA97FD76,
	0x19A0303F, 0xA5A4AF8A, 0xC8B8596D, 0x027B680A
};
static PGPByte nistg2[] = {
    0x03, 0x05, 0xF9, 0x39, 0x25,
	0x8D, 0xB7, 0xDD, 0x90, 0xE1, 0x93, 0x4F, 0x8C,
	0x70, 0xB0, 0xDF, 0xEC, 0x2E, 0xED, 0x25, 0xB8,
	0x55, 0x7E, 0xAC, 0x9C, 0x80, 0xE2, 0xE1, 0x98,
	0xF8, 0xCD, 0xBE, 0xCD, 0x86, 0xB1, 0x20, 0x53
};
static PGPByte nisto2[] = {
    0x03, 0xFF, 0xFF, 0xFF,
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xEF, 0x90,
	0x39, 0x96, 0x60, 0xFC, 0x93, 0x8A, 0x90, 0x16,
	0x5B, 0x04, 0x2A, 0x7C, 0xEF, 0xAD, 0xB3, 0x07
};


typedef struct ECcurve
{
	PGPUInt32		bitsize;
	const char		*name;
	PGPUInt32		cofactor;
	PGPUInt32		a;
	PGPUInt32		*b;
	PGPByte			*g;
	PGPByte			*order;
	DEBUG_STRUCT_CONSTRUCTOR( ECcurve )
} ECcurve ;

static ECcurve sCurves[] = {
	{ BITSIZE1, NAME1, COFACTOR1, 1, nistb1, nistg1, nisto1 },
	{ BITSIZE2, NAME2, COFACTOR2, 1, nistb2, nistg2, nisto2 }
};



/** Elliptic Curve / BigNum conversion functions **/


/* Convert an EC scalar into a BigNum */
	static PGPError
sScalartoBN( ecScalarRef ec, BigNum *bn, PGPMemoryMgrRef mgr,
			   PGPUInt32 scalarSize )
{
	PGPByte *buf;

	buf = PGPNewSecureData( mgr, scalarSize, 0 );
	if( buf == NULL )
		return kPGPError_OutOfMemory;
	if( ecScalarExtractBytes( ec, buf ) < 0
		|| bnInsertBigBytes( bn, buf, 0, scalarSize ) < 0 )
	{
		PGPFreeData( buf );
		return kPGPError_OutOfMemory;
	}
	PGPFreeData( buf );
	return kPGPError_NoErr;
}

/* Convert an EC point into a BigNum */
	static PGPError
sPointtoBN( ecPointRef ec, BigNum *bn, PGPMemoryMgrRef mgr,
			   PGPUInt32 pointSize )
{
	PGPByte *buf;

	buf = PGPNewSecureData( mgr, pointSize, 0 );
	if( buf == NULL )
		return kPGPError_OutOfMemory;
	if( ecPointCompress( ec ) < 0
		|| ecPointExtractBytes( ec, buf, 0 ) < 0
		|| bnInsertBigBytes( bn, buf, 0, pointSize ) < 0 )
	{
		PGPFreeData( buf );
		return kPGPError_OutOfMemory;
	}
	PGPFreeData( buf );
	return kPGPError_NoErr;
}

/* Convert an EC point's X coordinate into a BigNum */
	static PGPError
sPointXtoBN( ecPointRef ec, BigNum *bn, PGPMemoryMgrRef mgr,
			   PGPUInt32 coordSize )
{
	PGPByte *buf;

	buf = PGPNewSecureData( mgr, coordSize, 0 );
	if( buf == NULL )
		return kPGPError_OutOfMemory;
	if( ecPointExtractXYBytes( ec, buf, NULL, 0 ) < 0 ||
		bnInsertBigBytes( bn, buf, 0, coordSize ) < 0 )
	{
		PGPFreeData( buf );
		return kPGPError_OutOfMemory;
	}
	PGPFreeData( buf );
	return kPGPError_NoErr;
}

/* Convert a BigNum into an EC scalar */
	static PGPError
sBNtoScalar( BigNum *bn, ecScalarRef ec, PGPMemoryMgrRef mgr,
			   PGPUInt32 scalarSize )
{
	PGPByte *buf;

	buf = PGPNewSecureData( mgr, scalarSize, 0 );
	if( buf == NULL )
		return kPGPError_OutOfMemory;

	bnExtractBigBytes( bn, buf, 0, scalarSize );
	if( ecScalarInsertBytes( ec, buf, NULL ) < 0 )
	{
		PGPFreeData( buf );
		return kPGPError_OutOfMemory;
	}
	PGPFreeData( buf );
	return kPGPError_NoErr;
}

/* Convert a BigNum into an EC point */
	static PGPError
sBNtoPoint( BigNum *bn, ecPointRef ec, PGPMemoryMgrRef mgr,
			   PGPUInt32 pointSize )
{
	PGPByte *buf;

	buf = PGPNewSecureData( mgr, pointSize, 0 );
	if( buf == NULL )
		return kPGPError_OutOfMemory;

	bnExtractBigBytes( bn, buf, 0, pointSize );
	if( ecPointInsertBytes( ec, buf, 0 ) < 0 )
	{
		PGPFreeData( buf );
		return kPGPError_OutOfMemory;
	}
	PGPFreeData( buf );
	return kPGPError_NoErr;
}


/** Search for known curves **/

	static PGPInt32
sCurveIndexFromBitsize( PGPUInt32 bitsize )
{
	PGPUInt32 i;

	for( i=0; i<sizeof(sCurves)/sizeof(sCurves[0]); ++i )
	{
		if( sCurves[i].bitsize == bitsize )
			return (PGPInt32)i;
	}
	return -1;
}

	static PGPInt32
sCurveIndexFromScalarsize( PGPUInt32 scalarsize )
{
	PGPUInt32 i;

	for( i=0; i<sizeof(sCurves)/sizeof(sCurves[0]); ++i )
	{
		if( (sCurves[i].bitsize+7)/8 == scalarsize )
			return (PGPInt32)i;
	}
	return -1;
}

	static PGPInt32
sCurveIndexFromName( const char * name, PGPSize namelen )
{
	PGPUInt32 i;

	for( i=0; i<sizeof(sCurves)/sizeof(sCurves[0]); ++i )
	{
		if( strlen( sCurves[i].name ) == namelen
			&& strncmp( sCurves[i].name, name, namelen ) == 0 )
			return (PGPInt32)i;
	}
	return -1;
}

/* Special entry point used by keyring parsing code to get bitsize */
	PGPInt32
pgpECurveBitsFromName( const char * name, PGPSize namelen )
{
	PGPUInt32 i;

	for( i=0; i<sizeof(sCurves)/sizeof(sCurves[0]); ++i )
	{
		if( strlen( sCurves[i].name ) == namelen
			&& strncmp( sCurves[i].name, name, namelen ) == 0 )
			return sCurves[i].bitsize;
	}
	return -1;
}


/* Create mask buffer for EC encrypt/decrypt from shared secret curve
 * point and hash algorithm.  Use algorithm MGF1 from IEEE P1363.
 */
static PGPError
sFillEncryptMaskBuf( PGPContextRef context, ecPointRef ec,
					 PGPSize coordSize, PGPHashAlgorithm hashalg,
					 PGPByte *hashbuf, PGPSize hashbuflen )
{
	PGPHashContextRef	hc;
	PGPSize				hashsize;
	PGPUInt32			counter;
	PGPByte				counterbuf[4];
	void				*hbuf = NULL;
	void				*xbuf = NULL;
	PGPUInt32			off;
	PGPMemoryMgrRef		mgr = PGPPeekContextMemoryMgr( context );
	PGPError			err = kPGPError_NoErr;

	err = PGPNewHashContext( context, hashalg, &hc );
	pgpAssert( IsntPGPError( err ) );

	PGPGetHashSize( hc, &hashsize );

	/* Get X coordinate as shared secret */
	xbuf = PGPNewSecureData( mgr, coordSize, 0 );
	if( IsNull( xbuf ) ) {
		err = kPGPError_OutOfMemory;
		goto error;
	}
	if( ecPointExtractXYBytes( ec, xbuf, NULL, 0 ) < 0 ) {
		err = kPGPError_BadParams;
		goto error;
	}

	hbuf = PGPNewSecureData( mgr, hashsize, 0 );
	if( IsNull( hbuf ) ) {
		err = kPGPError_OutOfMemory;
		goto error;
	}

	counter = 0;
	for( off = 0; off < hashbuflen; off += hashsize )
	{
		PGPUInt32ToEndian( counter++, kPGPBigEndian, counterbuf );
		PGPContinueHash( hc, counterbuf, sizeof(counterbuf) );
		PGPContinueHash( hc, xbuf, coordSize );
		PGPFinalizeHash( hc, hbuf );
		pgpCopyMemory( hbuf, hashbuf+off, pgpMin(hashsize,hashbuflen-off) );
	}

 error:
	if( IsntNull( hbuf ) )
		PGPFreeData( hbuf );
	if( IsntNull( xbuf ) )
		PGPFreeData( xbuf );
	PGPFreeHashContext( hc );

	return err;
}



/** Public key functions **/

static void
ecPubDestroy(PGPPubKey *pubkey)
{
	ECpub *pub = (ECpub *)pubkey->priv;
	PGPContextRef	context;

	pgpAssertAddrValid( pubkey, PGPPubKey );
	context	= pubkey->context;

	ASSERTEC(pubkey->pkAlg);
	
	ecPointFree( pub->y );
	ecPointFree( pub->g );
	ecScalarFree( pub->order );
	ecFreeContext( pub->ec );
	pgpClearMemory( pub,  sizeof(pub));
	pgpContextMemFree( context, pub);
	pgpClearMemory( pubkey,  sizeof(pubkey));
	pgpContextMemFree(context, pubkey);
}

/* Return the largest possible PGPESK size for a given key */
static PGPSize
ecPubMaxesk(PGPPubKey const *pubkey, PGPPublicKeyMessageFormat format)
{
	ECpub const *pub = (ECpub *)pubkey->priv;
	PGPUInt32 pointSize;

	ASSERTECENC(pubkey->pkAlg);

	ecGetBufferSize( pub->ec, NULL, NULL, NULL, &pointSize );

	if (format == kPGPPublicKeyMessageFormat_PGP)
		return 1 + 2 + pointSize + 2 + MASKBUF_SIZE;
	else if (format == kPGPPublicKeyMessageFormat_PKCS1 ||
			 format == kPGPPublicKeyMessageFormat_IKE)
		return 1 + pointSize + MASKBUF_SIZE;
	else if (format == kPGPPublicKeyMessageFormat_X509) {
		/* SEQUENCE, length, INT, INT */
		PGPUInt32 len;
		len = pgpBnX509LenLen(pointSize+1) + 1 + pointSize+1 +
			pgpBnX509LenLen(MASKBUF_SIZE+1) + 1 + MASKBUF_SIZE+1;
		return 1 + 1 + pgpBnX509LenLen(len) + len;
	}
	pgpAssert(0);
	return 0;
}

/* Return the largest possible input size for ecEncrypt */
static PGPSize
ecPubMaxdecrypted(PGPPubKey const *pubkey, PGPPublicKeyMessageFormat format)
{
	(void) format;
	ASSERTECENC(pubkey->pkAlg);

	return MASKBUF_SIZE - 3;
}

/* Return the largest possible signature input to ecVerify */
static PGPSize
ecPubMaxsig(PGPPubKey const *pubkey, PGPPublicKeyMessageFormat format)
{
	ECpub const *pub = (ECpub *)pubkey->priv;
	PGPUInt32 scalarSize;

	ASSERTECSIG(pubkey->pkAlg);

	ecGetBufferSize( pub->ec, NULL, &scalarSize, NULL, NULL );
	if (format == kPGPPublicKeyMessageFormat_PGP)
		return 2*( 2 + scalarSize );
	else if (format == kPGPPublicKeyMessageFormat_PKCS1 ||
			 format == kPGPPublicKeyMessageFormat_IKE)
		return 2 * scalarSize;
	else if (format == kPGPPublicKeyMessageFormat_X509) {
		/* SEQUENCE, length, INT, INT */
		PGPUInt32 len;
		len = 2*(pgpBnX509LenLen(scalarSize+1) + 1 + scalarSize+1);
		return 1 + pgpBnX509LenLen(len) + len;
	}
	pgpAssert(0);
	return 0;
}

/*
 * Set keyid
 */
static void
ecPubSetKeyID(PGPPubKey *pubkey, PGPByte *keyid)
{
	pgpCopyMemory(keyid, pubkey->keyID, sizeof(pubkey->keyID));
}


static int
ecPubParams(PGPPubKey const *pubkey,
	BigNum *x1, BigNum *x2, BigNum *x3, BigNum *x4, BigNum *x5, BigNum *x6,
	BigNum *x7, BigNum *x8)
{
	ASSERTEC( pubkey->pkAlg );

	(void) x1;
	(void) x2;
	(void) x3;
	(void) x4;
	(void) x5;
	(void) x6;
	(void) x7;
	(void) x8;
	return kPGPError_PublicKeyUnimplemented;
}

/*
 * Given a buffer of at least "maxesk" bytes, make an PGPESK
 * into it and return the size of the PGPESK, or <0.
 */
static int
ecEncrypt(PGPPubKey const *pubkey, PGPByte const *key,
           PGPSize keylen, PGPByte *esk, PGPSize *esklen,
           PGPRandomContext const *rc, PGPPublicKeyMessageFormat format)
{
#if PGP_ENCRYPT_DISABLE /* [ */

	(void)pubkey;
	(void)key;
	(void)keylen;
	(void)esk;
	(void)esklen;
	(void)rc;
	(void)format;
	return kPGPError_FeatureNotAvailable;

#else /* PGP_ENCRYPT_DISABLE */  /* ]  [ */

	ECpub const *pub = (ECpub *)pubkey->priv;
	ecContextRef ec = pub->ec;
	unsigned t;
	int i;
	PGPUInt32 coordSize, scalarSize, pointSize;
	BigNum bn, bntq, bnorder;
	ecScalarRef td=NULL, td2=NULL;
	ecPointRef tq=NULL, dh=NULL;
	PGPByte *tdbuf=NULL, *maskbuf=NULL, *kbuf=NULL;
	PGPMemoryMgrRef		mgr	= NULL;
	
	mgr	= PGPPeekContextMemoryMgr( pubkey->context );

	ASSERTECENC(pubkey->pkAlg);

	ecGetBufferSize( ec, &coordSize, &scalarSize, NULL, &pointSize );

	if (keylen > MASKBUF_SIZE)
		return kPGPError_PublicKeyTooSmall; /* data too big for pubkey */

	if( format == kPGPPublicKeyMessageFormat_PGP ) {
		if (keylen+2 > MASKBUF_SIZE)
			return kPGPError_PublicKeyTooSmall; /* data too big for pubkey */
		/* Add checksum to key, place temporarily in esk buffer */
		t = 0;
		esk[0] = key[0];
		for (i = 1; i < (int)keylen; i++)
			t += esk[i] = key[i];
		esk[keylen] = (PGPByte)(t >> 8 & 255);
		esk[keylen+1] = (PGPByte)(t & 255);
		keylen += 2;
		key = esk;
	}

	bnBegin(&bn, mgr, SECURE );
	bnBegin(&bntq, mgr, SECURE );
	bnBegin(&bnorder, mgr, INSECURE );

	if( ecScalarCreate( ec, &td, SECURE ) < 0
		|| ecScalarCreate( ec, &td2, SECURE ) < 0
		|| ecPointCreate( ec, &tq ) < 0
		|| ecPointCreate( ec, &dh ) < 0
		|| sScalartoBN( pub->order, &bnorder, mgr, scalarSize ) < 0 )
		goto nomem;

	/* Choose the random td scalar to be used for this encryption */
	tdbuf = PGPNewSecureData( mgr, scalarSize+8, 0 );
	if( IsNull( tdbuf ) )
		goto nomem;
	pgpRandomGetBytes( rc, tdbuf, scalarSize+8 );
	if( bnInsertBigBytes( &bn, tdbuf, 0, scalarSize+8 ) < 0
		|| bnMod( &bn, &bn, &bnorder ) < 0 
		|| sBNtoScalar( &bn, td, mgr, scalarSize ) < 0 )
		goto nomem;

	/* Also calculate td*cofactor % order */
	if( bnMulQ( &bn, &bn, pub->cofactor ) < 0
		|| bnMod( &bn, &bn, &bnorder ) < 0
		|| sBNtoScalar( &bn, td2, mgr, scalarSize ) < 0 )
		goto nomem;

	/* Do EC multiplication to get tq value */
	if( ecPointMul( pub->g, td, FALSE, tq ) < 0 )
		goto nomem;

	/* Calculate shared secret curve point dh */
	if( ecPointMul( pub->y, td2, FALSE, dh ) < 0 )
		goto nomem;

	/* Create maskbuf from shared secret */
	maskbuf = PGPNewSecureData( mgr, MASKBUF_SIZE, 0 );
	if( IsNull( maskbuf ) )
		goto nomem;
	i = sFillEncryptMaskBuf( pubkey->context, dh, coordSize,
							 kPGPHashAlgorithm_SHA, maskbuf, MASKBUF_SIZE );
	if( i < 0 )
		goto done;

	/* PKCS-1 pad the key to encode its length */
	if (pgpPKCSPack(&bn, key, keylen, PKCS_PAD_ENCRYPTED, MASKBUF_SIZE, rc)<0)
		goto nomem;
	kbuf = PGPNewSecureData( mgr, MASKBUF_SIZE, 0 );
	if( IsNull( kbuf ) )
		goto nomem;
	bnExtractBigBytes( &bn, kbuf, 0, MASKBUF_SIZE );

	/* XOR the key with the shared secret */
	for( i=0; i<MASKBUF_SIZE; ++i )
		kbuf[i] ^= maskbuf[i];
	
	/* Prepare to output results */
	t = 0;
	esk[t++] = kPGPHashAlgorithm_SHA;
	if( sPointtoBN( tq, &bntq, mgr, pointSize ) < 0
		|| bnInsertBigBytes( &bn, kbuf, 0, MASKBUF_SIZE ) < 0 )
		goto nomem;

	if (format == kPGPPublicKeyMessageFormat_X509) {
		/* Put in SEQUENCE header for 509 encrypted data */
		PGPUInt32 len_seq, lenlen_seq;
		/* Count size of sequence, counting a 0 byte if hi bit is set */
		if (8*pointSize == bnBits(&bntq))
			len_seq = pgpBnX509LenLen(pointSize+1) + 1 + pointSize+1;
		else
			len_seq = pgpBnX509LenLen(pointSize) + 1 + pointSize;
		if (8*MASKBUF_SIZE == bnBits(&bn))
			len_seq += pgpBnX509LenLen(MASKBUF_SIZE+1) + 1 + MASKBUF_SIZE+1;
		else
			len_seq += pgpBnX509LenLen(MASKBUF_SIZE) + 1 + MASKBUF_SIZE;
		lenlen_seq = pgpBnX509LenLen(len_seq);
		esk[t++] = X509_TAG_SEQUENCE | X509_TAG_CONSTRUCTED;
		if (--lenlen_seq == 0) {
			esk[t++] = len_seq;
		} else {
			esk[t++] = 0x80 | lenlen_seq;
			len_seq <<= 8 * (4-lenlen_seq);
			while (lenlen_seq--) {
				esk[t++] = (PGPByte)(len_seq >> 24);
				len_seq <<= 8;
			}
		}
	}

	/* Pack the two values into the esk, first tq and then xbuf^key */
	t += pgpBnPutFormatted(&bntq, esk+t, pointSize, format);
	t += pgpBnPutFormatted(&bn, esk+t, MASKBUF_SIZE, format);

	if (esklen)
		*esklen = (PGPSize)t;
	
	i = 0;
	goto done;
nomem:
	i = kPGPError_OutOfMemory;
done:
	bnEnd(&bn);
	bnEnd(&bntq);
	bnEnd(&bnorder);
	if( IsntNull( tq ) )
		ecPointFree(tq);
	if( IsntNull( dh ) )
		ecPointFree(dh);
	if( IsntNull( td ) )
		ecScalarFree(td);
	if( IsntNull( td2 ) )
		ecScalarFree(td2);
	if( IsntNull( maskbuf ) )
		PGPFreeData( maskbuf );
	if( IsntNull( tdbuf ) )
		PGPFreeData( tdbuf );
	if( IsntNull( kbuf ) )
		PGPFreeData( kbuf );
	return i;


#endif /* PGP_ENCRYPT_DISABLE */ /* ] */
}


/*
 * Return 1 if (sig,siglen) is a valid MPI which signs
 * hash, of type h.  Check the DER-encoded prefix and the
 * hash itself.
 */
static int
ecVerify(PGPPubKey const *pubkey, PGPByte const *sig,
	PGPSize siglen, PGPHashVTBL const *h, PGPByte const *hash,
	PGPPublicKeyMessageFormat format)
{
#if PGP_VERIFY_DISABLE /* [ */

	(void)pubkey;
	(void)sig;
	(void)siglen;
	(void)h;
	(void)hash;
	(void)format;
	return kPGPError_FeatureNotAvailable;

#else /* PGP_VERIFY_DISABLE */  /* ]  [ */

	ECpub const *pub = (ECpub *)pubkey->priv;
	ecContextRef ec = pub->ec;
	PGPUInt32 coordSize, scalarSize, pointSize;
	BigNum r, s, w, u2, bnorder;
	ecPointRef gu1=NULL, yu2=NULL, ecv=NULL;
	ecScalarRef ecu1=NULL, ecu2=NULL;
	int i;
	PGPSize off;
	PGPMemoryMgrRef	mgr	= PGPPeekContextMemoryMgr( pubkey->context );

	ASSERTECSIG(pubkey->pkAlg);

	ecGetBufferSize( ec, &coordSize, &scalarSize, NULL, &pointSize );

#if 0
	/* Hashsize must be at least as big as size of field for legal sig */
	if (h->hashsize < scalarSize) {
		return 0;
	}
#endif

#if 0
	/* Allow generalizations of SHA, as long as they are big enough */
	if (h->algorithm != kPGPHashAlgorithm_SHA)
		return 0;	/* No match for sure! */
#endif

	bnBegin(&r, mgr, INSECURE );
	bnBegin(&s, mgr, INSECURE );
	bnBegin(&w, mgr, INSECURE );
	bnBegin(&u2, mgr, INSECURE );
	bnBegin(&bnorder, mgr, INSECURE );

	if( ecPointCreate( ec, &gu1 ) < 0
		|| ecPointCreate( ec, &yu2 ) < 0
		|| ecPointCreate( ec, &ecv ) < 0
		|| ecScalarCreate( ec, &ecu1, INSECURE ) < 0
		|| ecScalarCreate( ec, &ecu2, INSECURE ) < 0
		|| sScalartoBN( pub->order, &bnorder, mgr, scalarSize ) < 0 )
		goto nomem;

	/* sig holds two values.  Get first, r, from sig. */
	off = 0;
	if (format == kPGPPublicKeyMessageFormat_X509) {
		/* Parse SEQUENCE header for 509 sig data */
		PGPByte const *sigp = sig + off;
		PGPUInt32 len;
		if (pgpBnX509TagLen(&sigp, &len) != X509_TAG_SEQUENCE) {
			i = kPGPError_MalformedKeyComponent;
			goto done;
		}
		off += sigp - sig;
		if (len != siglen - off) {
			i = kPGPError_MalformedKeyComponent;
			goto done;
		}
	}
	i = pgpBnGetFormatted(&r, sig+off, siglen-off, scalarSize, format);
	if (i <= 0)
		goto fail;
	/* Get 2nd value, s, from SIG */
	off += i;
	i = pgpBnGetFormatted(&s, sig+off, siglen-off, scalarSize, format);
	if (i <= 0)
		goto fail;
	off += i;
	if (off != siglen) {
		i = kPGPError_BadSignatureSize;
		goto done;
	}

	/*
	 * Sanity-check r and s against the group order.  Both should
	 * be less than that.  If not, the signature is clearly bad.
	 */
	if (bnCmp(&r, &bnorder) >= 0 || bnCmp(&s, &bnorder) >= 0) {
		i = 0;	/* FAIL */
		goto done;
	}
	
	/* Reconstruct hash as u2 */
	if (bnInsertBigBytes(&u2, hash, 0, h->hashsize) < 0)
		goto nomem;

	/*
	 * Calculate DSS check function....
	 * Given signature (r,s) and hash H (in bn), compute:
	 * w = s^-1 mod q
	 * u1 = H * w mod q
	 * u2 = r * w mod q
	 * v = X coord of (u1*G + u2*Y)
	 * if v == r the signature checks.
	 *
	 * To save space, we put u1 into s, H into u2, and v into w.
	 */
	if (bnInv(&w, &s, &bnorder) < 0)
		goto nomem;
	if (bnMul(&s, &u2, &w) < 0 || bnMod(&s, &s, &bnorder) < 0)
		goto nomem;
	if (bnMul(&u2, &r, &w) < 0 || bnMod(&u2, &u2, &bnorder) < 0)
		goto nomem;
	if( sBNtoScalar( &s, ecu1, mgr, scalarSize ) < 0 )
		goto nomem;
	if( sBNtoScalar( &u2, ecu2, mgr, scalarSize ) < 0 )
		goto nomem;

	/* Now for the expensive part... */

	if( ecPointMul( pub->g, ecu1, FALSE, gu1 ) < 0
		|| ecPointMul( pub->y, ecu2, FALSE, yu2 ) < 0
		|| ecPointAdd( gu1, yu2, ecv ) < 0 )
		goto nomem;

	if( sPointXtoBN( ecv, &w, mgr, coordSize ) < 0
		|| bnMod( &w, &w, &bnorder ) < 0 )
		goto nomem;

	/* Compare result with r, should be equal */
	i = bnCmp(&w, &r) == 0;

	goto done;

fail:
	if (!i)
		i = kPGPError_BadSignatureSize;
	goto done;
nomem:
	i = kPGPError_OutOfMemory;
	goto done;
done:
	bnEnd(&bnorder);
	bnEnd(&u2);
	bnEnd(&w);
	bnEnd(&s);
	bnEnd(&r);
	if( IsntNull( gu1 ) )
		ecPointFree( gu1 );
	if( IsntNull( yu2 ) )
		ecPointFree( yu2 );
	if( IsntNull( ecv ) )
		ecPointFree( ecv );
	if( IsntNull( ecu1 ) )
		ecScalarFree( ecu1 );
	if( IsntNull( ecu2 ) )
		ecScalarFree( ecu2 );

	return i;

#endif /* PGP_VERIFY_DISABLE */ /* ] */
}


/*
 * Turn a PGPPubKey into the algorithm-specific parts of a public key.
 * A public key's EC-specific part is:
 *
 *  0      2    Curve type (=0)
 *  2      2+s  Curve name (s=strlen(name))
 * 4+s     2+p  Public curve point (p=pointSize)
 * 6+s+p
 */
static PGPSize
ecPubBufferLength(PGPPubKey const *pubkey)
{
	ECpub const *pub = (ECpub *)pubkey->priv;
	PGPUInt32 pointSize;
	PGPUInt32 scalarSize;
	PGPInt32 curve;

	ASSERTEC(pubkey->pkAlg);

	ecGetBufferSize( pub->ec, NULL, &scalarSize, NULL, &pointSize );

	curve = sCurveIndexFromScalarsize( scalarSize );
	pgpAssert( curve >= 0 );
	return 6 + pointSize + strlen( sCurves[curve].name );
}

static void
ecPubToBuffer(PGPPubKey const *pubkey, PGPByte *buf)
{
	ECpub const *pub = (ECpub *)pubkey->priv;
	PGPUInt32 pointSize;
	PGPUInt32 scalarSize;
	PGPInt32 curve;
	PGPSize namelen;
	PGPUInt32 i;
	BigNum bny;
	PGPMemoryMgrRef mgr = PGPPeekContextMemoryMgr( pubkey->context );

	ASSERTEC(pubkey->pkAlg);

	ecGetBufferSize( pub->ec, NULL, &scalarSize, NULL, &pointSize );

	i = 0;
	buf[i++] = 0;
	buf[i++] = 0;

	curve = sCurveIndexFromScalarsize( scalarSize );
	pgpAssert( curve >= 0 );

	namelen = strlen( sCurves[curve].name );
	buf[i++] = ((8*namelen-1) >> 8) & 0xff;
	buf[i++] = ((8*namelen-1) >> 0) & 0xff;
	pgpCopyMemory( sCurves[curve].name, buf+1, namelen );
	i += namelen;

	bnBegin( &bny, mgr, INSECURE );
	(void)sPointtoBN( pub->y, &bny, mgr, pointSize );
	i += pgpBnPutPlain( &bny, buf+i );
	bnEnd( &bny );
}


/* A little helper function that's used twice */
static void
ecFillPubkey(PGPPubKey *pubkey, ECpub *pub)
{
	pubkey->next	 = NULL;
	pubkey->pkAlg	 = kPGPPublicKeyAlgorithm_ECSign;
	pubkey->priv	 = pub;
	pubkey->destroy  = ecPubDestroy;
	pubkey->maxesk   = ecPubMaxesk;
	pubkey->maxdecrypted   = ecPubMaxdecrypted;
	pubkey->maxsig   = ecPubMaxsig;
	pubkey->encrypt  = ecEncrypt;
	pubkey->verify   = ecVerify;
	pubkey->bufferLength  = ecPubBufferLength;
	pubkey->toBuffer = ecPubToBuffer;
	pubkey->pubparams= ecPubParams;
	pubkey->setkeyid= ecPubSetKeyID;
}


/*
 * Turn the algorithm-specific parts of a public key into a PGPPubKey
 * structure.  A public key's EC-specific part is:
 *
 *  0      2    Curve type (=0)
 *  2      2+s  Curve name (s=strlen(name))
 * 4+s     2+p  Public curve point (p=pointSize)
 * 6+s+p
 */
PGPPubKey *
ecPubFromBuf(
	PGPContextRef	context,
	PGPByte const *	buf,
	PGPSize			size,
	PGPError *		error)
{
	PGPPubKey *pubkey;
	ECpub *pub;
	unsigned i, t;
	PGPInt32 curve;
	PGPError	err	= kPGPError_OutOfMemory;
	PGPMemoryMgrRef	mgr	= PGPPeekContextMemoryMgr( context );

	bnInit();

	if (size < 6)
		return NULL;

	if( buf[0] != 0 || buf[1] != 0 )
	{
		*error = kPGPError_MalformedKeyComponent;
		return NULL;
	}

	i = ((unsigned)buf[2] << 8) + buf[3];
	if (!i || buf[3] >> ((i-1) & 7) != 1) {
		*error = kPGPError_MalformedKeyComponent;
		return NULL;	/* Bad bit length */
	}
	i = (i+7)/8;
	if (size < 4+i) {
		*error = kPGPError_KeyPacketTruncated;
		return NULL;
	}
	curve = sCurveIndexFromName( buf+4, i );
	if( curve < 0 )
	{
		/* Unrecognized name */
		*error = kPGPError_MalformedKeyComponent;
		return NULL;
	}

	t = ((unsigned)buf[4+i] << 8) + buf[5+i];
	if (!t || buf[6+i] >> ((t-1) & 7) != 1) {
		*error = kPGPError_MalformedKeyComponent;
		return NULL;	/* Bad bit length */
	}
	t = (t+7)/8;
	if (size < 6+i+t) {
		*error = kPGPError_KeyPacketTruncated;
		return NULL;
	}

	pub = (ECpub *)pgpContextMemAlloc( context,
		sizeof(*pub), kPGPMemoryMgrFlags_Clear );
	if (pub) {
		pubkey = (PGPPubKey *)pgpContextMemAlloc( context,
			sizeof(*pubkey), kPGPMemoryMgrFlags_Clear);
		if (pubkey) {
			ecContextRef ec;
			BigNum bnorder;
			PGPUInt32 scalarSize;

			pubkey->context	= context;

			if( ecCreate2mContext( mgr, sCurves[curve].bitsize,
								   EC_MEM_USAGE_HIGH, &pub->ec )<0 )
			{
				pgpContextMemFree( context, pubkey );
				*error = kPGPError_OutOfMemory;
				return NULL;
			}
			ec = pub->ec;

			ecSetEC2mParamAInt( ec, sCurves[curve].a );
			ecSetEC2mParamB( ec, sCurves[curve].b );

			ecGetBufferSize( pub->ec, NULL, &scalarSize, NULL, NULL );

			bnBegin( &bnorder, mgr, INSECURE );
			pub->cofactor = sCurves[curve].cofactor;

			if( ecPointCreate( ec, &pub->y ) >= 0
				&& ecPointCreate( ec, &pub->g ) >= 0
				&& ecScalarCreate( ec, &pub->order, INSECURE ) >= 0
				&& ecPointInsertBytes( pub->g, sCurves[curve].g, 0 ) >= 0
				&& bnInsertBigBytes( &bnorder, sCurves[curve].order,
									 0, scalarSize ) >= 0
				&& sBNtoScalar( &bnorder, pub->order, mgr, scalarSize )>=0
				&& ecPointInsertBytes( pub->y, buf+6+i, 0 ) >= 0 )
			{
				ecFillPubkey(pubkey, pub);
				bnEnd( &bnorder );
				*error = 0;
				return pubkey;
			}
			/* Failed = clean up and return NULL */
			err = kPGPError_OutOfMemory;
			bnEnd( &bnorder );
			if( IsntNull( pub->y ) )
				ecPointFree( pub->y );
			if( IsntNull( pub->g ) )
				ecPointFree( pub->g );
			if( IsntNull( pub->order ) )
				ecScalarFree( pub->order );
		}
		pgpContextMemFree( context, pub);
	}

	*error = err;
	return NULL;
}

/*
 * Return the size of the public portion of a key buffer.
 */
int
ecPubKeyPrefixSize(PGPByte const *buf, PGPSize size)
{
	return 2 + pgpBnParse(buf+2, size-2, 2, NULL, NULL);
}



/** Secret key functions **/

static void
ecSecDestroy(PGPSecKey *seckey)
{
	ECsecPlus *sec = (ECsecPlus *)seckey->priv;
	PGPContextRef	context;

	pgpAssertAddrValid( seckey, PGPPubKey );
	context	= seckey->context;

	ASSERTEC(seckey->pkAlg);

	if( !sec->locked )
	{
		/* Locked keys have no data in these fields */
		ecPointFree( sec->s.y );
		ecPointFree( sec->s.g );
		ecScalarFree( sec->s.order );
		ecScalarFree( sec->s.x );
		ecFreeContext( sec->s.ec );
	}

	pgpClearMemory(sec->cryptkey, sec->ckalloc);
	pgpContextMemFree( context, sec->cryptkey);
	PGPFreeData( sec );			/* Wipes as it frees */
	pgpClearMemory( seckey,  sizeof(seckey));
	pgpContextMemFree( context, seckey);
}

/*
 * Generate a PGPPubKey from a PGPSecKey
 */
static PGPPubKey *
ecPubkey(PGPSecKey const *seckey)
{
	/* Not implemented - only needed for X.509 and smart card keys */
	(void) seckey;
	pgpAssert( 0 );
	return NULL;
}

/*
 * Set keyid
 */
static void
ecSecSetKeyID(PGPSecKey *seckey, PGPByte *keyid)
{
	pgpCopyMemory(keyid, seckey->keyID, sizeof(seckey->keyID));
}

/*
 * Yes, there *is* a reason that this is a function and no a variable.
 * On a hardware device with an automatic timeout,
 * it actually might need to do some work to find out.
 */
static int
ecIslocked(PGPSecKey const *seckey)
{
	ECsecPlus const *sec = (ECsecPlus *)seckey->priv;

	ASSERTEC(seckey->pkAlg);
	return sec->locked;
}


/*
 * Return the algorithm and (symmetric) key size used for locking/unlocking
 * the secret key.
 */
static PGPError
ecLockingAlgorithm(
	PGPSecKey const *seckey,
	PGPCipherAlgorithm *pAlg,
	PGPSize *pAlgKeySize
	)
{
	ECsecPlus *sec = (ECsecPlus *)seckey->priv;
	PGPCipherVTBL const *cipher;
	PGPByte alg;
	int i;

	ASSERTEC(seckey->pkAlg);

	if( IsntNull( pAlg ) )
		*pAlg = (PGPCipherAlgorithm) 0;
	if( IsntNull( pAlgKeySize ) )
		*pAlgKeySize = (PGPSize) 0;

	/* Check packet for basic consistency */
	i = pgpBnParse(sec->cryptkey+2, sec->cklen-2, 2, NULL, NULL);
	if (i < 0)
		return (PGPError)i;
	i += 2;

	/* Get the encryption algorithm (cipher number).  0 == no encryption */
	alg = sec->cryptkey[i] & 255;

	/* New style has 0xff or 0xfe then algorithm value */
	if (alg == 0xff || alg == 0xfe)
		alg = sec->cryptkey[i+1] & 255;

	cipher = pgpCipherGetVTBL( (PGPCipherAlgorithm)alg);
	if (!cipher)
		return kPGPError_BadCipherNumber;

	/* Success */
	if( IsntNull( pAlg ) )
		*pAlg = (PGPCipherAlgorithm) alg;
	if( IsntNull( pAlgKeySize ) )
		*pAlgKeySize = cipher->keysize;

	return kPGPError_NoErr;
}


/*
 * Return the StringToKey type for unlocking the given key.  We use
 * kPGPStringToKey_Literal to flag a secret split unlocking buffer.
 * Returns kPGPStringToKey_Simple if key has no passphrase.
 */
static PGPError
ecS2KType(
	PGPSecKey const *seckey,
	PGPStringToKeyType *s2kType
	)
{
	ECsecPlus *sec = (ECsecPlus *)seckey->priv;
	PGPByte alg;
	int i;

	ASSERTEC(seckey->pkAlg);

	/* note that 0 is a valid type, but use it as default anyway */
	if( IsntNull( s2kType ) )
		*s2kType = (PGPStringToKeyType) 0;

	/* Check packet for basic consistency */
	i = pgpBnParse(sec->cryptkey+2, sec->cklen-2, 2, NULL, NULL);
	if (i < 0)
		return (PGPError)i;
	i += 2;

	/* Get the encryption algorithm (cipher number).  0 == no encryption */
	alg = sec->cryptkey[i] & 255;

	if (alg == 0xff || alg == 0xfe) {
		/* New style has 0xff or 0xfe then algorithm value then S2K */
		*s2kType = (PGPStringToKeyType) sec->cryptkey[i+2];
	} else {
		/* Unencrypted or old-style simple encryption */
		*s2kType = kPGPStringToKey_Simple;
	}

	return kPGPError_NoErr;
}


/*
 * Convert a passphrase into a s2k literal buffer for the key.
 * Returns error code.  Output buffer will be size of the *pAlgKeySize
 * parameter from pgpSecKeyLockingalgorithm.
 */
static PGPError
ecConvertPassphrase(PGPSecKey *seckey, PGPEnv const *env,
	  char const *phrase, PGPSize plen, PGPByte *outbuf)
{
	ECsecPlus *sec = (ECsecPlus *)seckey->priv;
	PGPStringToKey *s2k;
	PGPByte alg;
	PGPBoolean hasS2K;
	PGPCipherVTBL const *cipher;
	int i;
	PGPContextRef	context	= pgpenvGetContext( env );

	ASSERTEC(seckey->pkAlg);
	pgpAssert (IsntNull( outbuf ) );

	/* Check packet for basic consistency */
	i = pgpBnParse(sec->cryptkey+2, sec->cklen-2, 2, NULL, NULL, NULL, NULL);
	if (i < 0)
		return (PGPError)i;
	i += 2;

	/* Get the encryption algorithm (cipher number).  0 == no encryption */
	alg = sec->cryptkey[i++] & 255;

	hasS2K = (alg == 0xff || alg == 0xfe);

	/* New style has 255 then algorithm value */
	if (hasS2K)
		alg = sec->cryptkey[i++] & 255;

	/* Now we are looking at the s2k object if there is one. */
	if (alg == 0) {
		/* Key is not locked */
		return kPGPError_BadParams;
	}
	cipher = pgpCipherGetVTBL( (PGPCipherAlgorithm)alg);
	if( IsNull( cipher ) )
		return kPGPError_BadCipherNumber;

	if (hasS2K) {
		pgpS2Kdecode(&s2k, context, sec->cryptkey+i, sec->cklen-i);
	} else {
		s2k = pgpS2Ksimple(context, pgpHashByNumber(kPGPHashAlgorithm_MD5));
	}
	if (IsNull( s2k ) )
		return kPGPError_OutOfMemory;
	pgpStringToKey(s2k, phrase, plen, outbuf, cipher->keysize);
	pgpS2Kdestroy (s2k);

	return kPGPError_NoErr;
}


/*
 * Try to decrypt the secret key wih the given passphrase.  Returns >0
 * if it was the correct passphrase. =0 if it was not, and <0 on error.
 * Does not alter the key even if it's the wrong passphrase and already
 * unlocked.  A NULL passphrae will work if the key is unencrypted.
 * 
 * A (secret) key's EC-specific part is:
 *
 *  0      2    Curve type (=0)
 *  2      2+s  Curve name (s=strlen(name))
 * 4+s     2+p  Public curve point (p=pointSize)
 * 6+s+p   e    Encryption algorithm data, S2K info
 * 6+s+p+e 2+w  MPI for x, secret part (w=scalarSize)
 * 8+s+p+e+w
 */

static int
ecUnlock(PGPSecKey *seckey,
	  char const *phrase, PGPSize plen, PGPBoolean hashedPhrase)
{
	ECsecPlus *sec = (ECsecPlus *)seckey->priv;
	ecContextRef ec;
	BigNum bnx, bny, bnorder;
	PGPCFBContext *cfb = NULL;	/* Necessary */
	PGPUInt32 v, t, yoff;
	
	PGPByte alg /* actual algorithm */, alg0 /* first algorithm descriptor */;
        PGPHashContextRef checksumHash = NULL;
        PGPHashVTBL const *hashEntry = NULL;
        PGPByte checksumStored[20], checksum[20];       /* size no less then the max return value from pgpChecksumGetEx */
        unsigned checksumSize;
	
	PGPUInt32 scalarSize, pointSize;
	PGPInt32 curve;
	int i;
	PGPMemoryMgrRef		mgr	= NULL;

	if( pgpFIPSModeEnabled() )
	{
		/* FIPS does not allow using keys with NULL passphrases */
		if( IsNull( phrase ) || plen == 0 )
			return( 0 );
	}
	
	mgr	= PGPPeekContextMemoryMgr( seckey->context );

	ASSERTEC(seckey->pkAlg);
	bnInit();

	if (sec->cklen < 8)
		return kPGPError_KeyPacketTruncated;
	if( sec->cryptkey[0] != 0 || sec->cryptkey[1] != 0 )
		return kPGPError_MalformedKeyComponent;

	v = ((unsigned)sec->cryptkey[2] << 8) + sec->cryptkey[3];
	if (!v || sec->cryptkey[4] >> ((v-1) & 7) != 1) {
		return kPGPError_MalformedKeyComponent;
	}
	v = (v+7)/8;
	if (sec->cklen < 8+v)
		return kPGPError_KeyPacketTruncated;
	curve = sCurveIndexFromName( sec->cryptkey+4, v );
	if( curve < 0 )
	{
		/* Unrecognized name */
		return kPGPError_MalformedKeyComponent;
	}

	t = ((unsigned)sec->cryptkey[4+v] << 8) + sec->cryptkey[5+v];
	if (!t || sec->cryptkey[6+v] >> ((t-1) & 7) != 1) {
		return kPGPError_MalformedKeyComponent;
	}
	t = (t+7)/8;
	if (sec->cklen < 6+v+t)
		return kPGPError_KeyPacketTruncated;
	yoff = 6+v;
	v += t + 6;
	if (sec->cklen < v+1)
		return kPGPError_KeyPacketTruncated;

	/* Get the encryption algorithm (cipher number).  0 == no encryption */
	alg0 = alg = sec->cryptkey[v];
	if( alg0 == 0xff || alg0 == 0xfe  )
		alg = sec->cryptkey[v+1];

	/* If the phrase is empty, set it to NULL */
	if (plen == 0)
		phrase = NULL;
	/*
	 * We need a pass if it is encrypted, and we cannot have a
	 * password if it is NOT encrypted.  I.e., this is a logical
	 * xor (^^)
	 */
	if (!phrase != !alg0)
		return 0;

	hashEntry = pgpHashByNumberWithMask( alg0!=0xfe ? kPGPHashAlgorithm_Checksum16 : kPGPHashAlgorithm_SHA, 0xffffffff );
	checksumHash = pgpHashCreate( mgr, hashEntry );
	if( checksumHash == NULL )  { 
		i = kPGPError_BadHashNumber;
		goto done;
	}

	i = pgpCipherSetup(sec->cryptkey + v, sec->cklen - v, phrase, plen,
					   hashedPhrase, FALSE, seckey->context, &cfb);
	if (i < 0)
		return i;
	v += i;

	PGPResethash( checksumHash );
	bnBegin( &bnx, mgr, SECURE );
	bnBegin( &bny, mgr, INSECURE );
	bnBegin( &bnorder, mgr, INSECURE );

	if (bnInsertBigBytes(&bny, sec->cryptkey+yoff, 0, t) < 0)
		goto nomem;

	i = pgpBnGet(&bnx, sec->cryptkey + v, sec->cklen - v, cfb, alg0, FALSE/*old*/, checksumHash);
	if (i <= 0)
		goto badpass;
	v += i;

        /* Check that we ended in the right place */
	checksumSize = pgpChecksumGet( sec->cryptkey+v, NULL, alg, NULL );
	pgpAssert( checksumSize <= sizeof(checksumStored) );    /* programmer's error */
	if (sec->cklen - v != checksumSize)  {
		i = kPGPError_KEY_LONG;
		goto fail;
	}

	pgpChecksumGet(sec->cryptkey+v, cfb, alg, checksumStored);
	PGPFinalizeHash( checksumHash, checksumHash );
	if( !pgpMemoryEqual( checksumStored, checksumHash, checksumSize ) )
		goto badpass;
	
	if( ecCreate2mContext( mgr, sCurves[curve].bitsize,
						   EC_MEM_USAGE_HIGH, &sec->s.ec ) < 0 )
		goto nomem;
	ec = sec->s.ec;

	ecSetEC2mParamAInt( ec, sCurves[curve].a );
	ecSetEC2mParamB( ec, sCurves[curve].b );
	ecGetBufferSize( ec, NULL, &scalarSize, NULL, &pointSize );

	sec->s.cofactor = sCurves[curve].cofactor;

	if( ecPointCreate( ec, &sec->s.y ) >= 0
		&& ecPointCreate( ec, &sec->s.g ) >= 0
		&& ecScalarCreate( ec, &sec->s.order, INSECURE ) >= 0
		&& ecScalarCreate( ec, &sec->s.x, SECURE ) >= 0
		&& ecPointInsertBytes( sec->s.g, sCurves[curve].g, 0 )>=0
		&& bnInsertBigBytes( &bnorder, sCurves[curve].order,
							 0, scalarSize ) >= 0
		&& sBNtoScalar( &bnorder, sec->s.order, mgr, scalarSize ) >= 0
		&& sBNtoPoint( &bny, sec->s.y, mgr, pointSize ) >= 0
		&& sBNtoScalar( &bnx, sec->s.x, mgr, scalarSize ) >= 0 )
	{

#if 0
// Debugging sanity check that g^x == y
{
ecPointRef zero, neggx;
ecScalarRef negx;
BigNum bnnegx;
bnBegin( &bnnegx, mgr, SECURE );
bnCopy( &bnnegx, &bnorder );
bnSub( &bnnegx, &bnx );
ecScalarCreate( ec, &negx, SECURE );
sBNtoScalar( &bnnegx, negx, mgr, scalarSize );
ecPointCreate( ec, &neggx );
ecPointMul( sec->s.g, negx, FALSE, neggx );
ecPointCreate( ec, &zero );
ecPointAdd( neggx, sec->s.y, zero );
pgpAssert( ecPointIsZero( zero ) );
ecPointFree( zero );
ecPointFree( neggx );
ecScalarFree( negx );
bnEnd (&bnnegx);
} 
#endif

		sec->locked = 0;
		if (cfb)
			PGPFreeCFBContext (cfb);
		bnEnd(&bny);
		bnEnd(&bnx);
		bnEnd(&bnorder);
		return 1;	/* Decrypted */
	}

	/* Fail, undo above */
	if( IsntNull( sec->s.y ) )
		ecPointFree( sec->s.y );
	if( IsntNull( sec->s.g ) )
		ecPointFree( sec->s.g );
	if( IsntNull( sec->s.order ) )
		ecScalarFree( sec->s.order );
	if( IsntNull( sec->s.x ) )
		ecScalarFree( sec->s.x );
	pgpClearMemory( &sec->s, sizeof( ECsec ) );
	/* Fall through */

nomem:
	i = kPGPError_OutOfMemory;
	goto done;
badpass:
	i = 0;	/* Incorrect passphrase */
	goto done;
fail:
	if (!i)
		i = kPGPError_KeyPacketTruncated;
	goto done;
done:
	if (cfb)
		PGPFreeCFBContext (cfb);
       if( checksumHash != NULL )
                PGPFreeHashContext( checksumHash );
	bnEnd( &bny );
	bnEnd( &bnorder );
	bnEnd( &bnx );
	return i;
}

/*
 * Relock the key.
 */
static void
ecLock(PGPSecKey *seckey)
{
	ECsecPlus *sec = (ECsecPlus *)seckey->priv;

	ASSERTEC(seckey->pkAlg);
	sec->locked = 1;

	ecPointFree( sec->s.y );
	ecPointFree( sec->s.g );
	ecScalarFree( sec->s.order );
	ecScalarFree( sec->s.x );
	ecFreeContext( sec->s.ec );
	pgpClearMemory( &sec->s, sizeof( ECsec ) );
}

static PGPSize
ecSecMaxdecrypted(PGPSecKey const *seckey, PGPPublicKeyMessageFormat format);


/*
 * Try to decrypt the given esk.  If the key is locked, try the given
 * passphrase.  It may or may not leave the key unlocked in such a case.
 * (Some hardware implementations may insist on a password per usage.)
 */
static int
ecDecrypt(PGPSecKey *seckey, PGPByte const *esk, PGPSize esklen,
		   PGPByte *key, PGPSize *keylen,
		   char const *phrase, PGPSize plen,
		   PGPPublicKeyMessageFormat format)
{
#if PGP_DECRYPT_DISABLE /* [ */

	(void)seckey;
	(void)esk;
	(void)esklen;
	(void)key;
	(void)keylen;
	(void)phrase;
	(void)plen;
	(void)format;
	return kPGPError_FeatureNotAvailable;

#else /* PGP_DECRYPT_DISABLE */  /* ]  [ */

	ECsecPlus *sec = (ECsecPlus *)seckey->priv;
	ecContextRef ec = sec->s.ec;
	PGPUInt32 coordSize, scalarSize, pointSize;
	BigNum k, bn, bnorder, bnx;
	ecPointRef tq=NULL, dh=NULL;
	ecScalarRef td2=NULL;
	int i, j;
	unsigned t;
	PGPUInt32 hashalg;
	PGPUInt32 masksize;
	PGPSize len;
	PGPSize off;
	PGPByte *secbuf = NULL;
	PGPByte *maskbuf = NULL;
	PGPMemoryMgrRef		mgr	= NULL;

	mgr	= PGPPeekContextMemoryMgr( seckey->context );

	ASSERTECENC(seckey->pkAlg);
	if (sec->locked) {
		i = ecUnlock(seckey, phrase, plen, FALSE);
		if (i <= 0)
			return i ? i : kPGPError_KeyIsLocked;
		pgpAssert(!sec->locked);
	}

	if (esklen < 2)
		return kPGPError_BadSessionKeySize;
	
	ecGetBufferSize( ec, &coordSize, &scalarSize, NULL, &pointSize );

	off = 0;
	hashalg = esk[off++];
	if (hashalg != kPGPHashAlgorithm_SHA)
		return kPGPError_MalformedKeyComponent;

	bnBegin(&bn, mgr, SECURE);
	bnBegin(&k, mgr, SECURE);
	bnBegin(&bnx, mgr, SECURE);
	bnBegin(&bnorder, mgr, INSECURE);

	if( ecPointCreate( ec, &tq ) < 0
		|| ecPointCreate( ec, &dh ) < 0
		|| ecScalarCreate( ec, &td2, SECURE ) < 0
		|| sScalartoBN( sec->s.order, &bnorder, mgr, scalarSize ) < 0
		|| sScalartoBN( sec->s.x, &bnx, mgr, scalarSize ) < 0 )
		goto nomem;

	/* ESK holds two values.  Get first, tq, from ESK. */
	if (format == kPGPPublicKeyMessageFormat_X509) {
		/* Parse SEQUENCE header for 509 encrypted data */
		PGPByte const *eskp = esk + off;
		PGPUInt32 len;
		if (pgpBnX509TagLen(&eskp, &len) != X509_TAG_SEQUENCE) {
			i = kPGPError_MalformedKeyComponent;
			goto done;
		}
		off += eskp - esk;
		if (len != esklen - off) {
			i = kPGPError_BadSessionKeySize;
			goto done;
		}
	}
	i = pgpBnGetFormatted(&bn, esk+off, esklen-off, pointSize, format);
	if (i <= 0)
		goto fail;
	/* Get 2nd value, k, from ESK */
	off += i;
	i = pgpBnGetFormatted(&k, esk+off, esklen-off, MASKBUF_SIZE, format);
	if (i <= 0)
		goto fail;
	off += i;
	if (off != esklen) {
		i = kPGPError_BadSessionKeySize;
		goto done;
	}
	
	/* Convert tq to EC point; k is key xor shared secret */
	if( sBNtoPoint( &bn, tq, mgr, pointSize ) < 0 )
		goto nomem;
	masksize = bnBytes(&k);
	secbuf = PGPNewSecureData( mgr, masksize, 0 );
	if( IsNull( secbuf ) )
		goto nomem;
	bnExtractBigBytes( &k, secbuf, 0, masksize );

	/* Calculate td2 = x*cofactor % order */
	if( bnMulQ( &bn, &bnx, sec->s.cofactor ) < 0
		|| bnMod( &bn, &bn, &bnorder ) < 0
		|| sBNtoScalar( &bn, td2, mgr, scalarSize ) < 0 )
		goto nomem;

	/* Calculate shared secret curve point dh */
	if( ecPointMul( tq, td2, FALSE, dh ) < 0 )
		goto nomem;

	/* Create maskbuf from shared secret */
	maskbuf = PGPNewSecureData( mgr, masksize, 0 );
	if( IsNull( maskbuf ) )
		goto nomem;
	i = sFillEncryptMaskBuf( seckey->context, dh, coordSize,
							 kPGPHashAlgorithm_SHA, maskbuf, masksize );
	if( i < 0 )
		goto fail;

	/* XOR into secbuf to recover PKCS1-padded value */
	for( t=0; t<masksize; t++ )
		secbuf[t] ^= maskbuf[t];

	if( bnInsertBigBytes( &bn, secbuf, 0, masksize ) < 0 )
		goto nomem;
	len = ecSecMaxdecrypted( seckey, format );
	i = pgpPKCSUnpack( key, len, &bn, PKCS_PAD_ENCRYPTED, masksize );
	if( i < 0 )
		goto fail;
	
	if (format == kPGPPublicKeyMessageFormat_PGP) {
		/* Check checksum */
		t = 0;
		for (j = 1; j < i-2; j++)
			t += key[j];
		if (t != ((unsigned)key[i-2]<<8) + key[i-1])
		{
			i = kPGPError_CorruptData;
			goto fail;
		}
		pgpClearMemory(key+i-2, 2);

		/* The actual key */
		if (keylen)
			*keylen = (PGPSize)i-2;
	} else {
		if( keylen)
			*keylen = (PGPSize)i;
	}
	
	i = 0;
	goto done;

fail:
	if (!i)
		i = kPGPError_BadSessionKeySize;
	goto done;
nomem:
	i = kPGPError_OutOfMemory;
	goto done;
done:
	if( IsntNull( secbuf ) )
		PGPFreeData( secbuf );
	if( IsntNull( maskbuf ) )
		PGPFreeData( maskbuf );
	bnEnd(&k);
	bnEnd(&bn);
	bnEnd(&bnx);
	bnEnd(&bnorder);
	if( IsntNull( tq ) )
		ecPointFree( tq );
	if( IsntNull( dh ) )
		ecPointFree( dh );
	if( IsntNull( td2 ) )
		ecScalarFree( td2 );
	return i;

#endif /* PGP_DECRYPT_DISABLE */ /* ] */
}


/*
 * Return the size of the buffer needed, worst-case, for the decrypted
 * output.
 */
static PGPSize
ecSecMaxdecrypted(PGPSecKey const *seckey, PGPPublicKeyMessageFormat format)
{
	(void) format;
	ASSERTECENC(seckey->pkAlg);

	return MASKBUF_SIZE - 3;
}

/* Return the largest possible PGPESK size for a given key */
static PGPSize
ecSecMaxesk(PGPSecKey const *seckey, PGPPublicKeyMessageFormat format)
{
	ECsecPlus const *sec = (ECsecPlus *)seckey->priv;
	PGPUInt32 pointSize;

	ASSERTECENC(seckey->pkAlg);

	ecGetBufferSize( sec->s.ec, NULL, NULL, NULL, &pointSize );

	if (format == kPGPPublicKeyMessageFormat_PGP)
		return 1 + 2 + pointSize + 2 + MASKBUF_SIZE;
	else if (format == kPGPPublicKeyMessageFormat_PKCS1 ||
			 format == kPGPPublicKeyMessageFormat_IKE)
		return 1 + pointSize + MASKBUF_SIZE;
	else if (format == kPGPPublicKeyMessageFormat_X509) {
		/* SEQUENCE, length, INT, INT */
		PGPUInt32 len;
		len = pgpBnX509LenLen(pointSize+1) + 1 + pointSize+1 +
			pgpBnX509LenLen(MASKBUF_SIZE+1) + 1 + MASKBUF_SIZE+1;
		return 1 + 1 + pgpBnX509LenLen(len) + len;
	}
	pgpAssert(0);
	return 0;
}

static PGPSize
ecSecMaxsig(PGPSecKey const *seckey, PGPPublicKeyMessageFormat format)
{
	ECsecPlus const *sec = (ECsecPlus *)seckey->priv;
	PGPUInt32 scalarSize;

	ASSERTECSIG(seckey->pkAlg);

	ecGetBufferSize( sec->s.ec, NULL, &scalarSize, NULL, NULL );
	if (format == kPGPPublicKeyMessageFormat_PGP)
		return 2*( 2 + scalarSize );
	else if (format == kPGPPublicKeyMessageFormat_PKCS1 ||
			 format == kPGPPublicKeyMessageFormat_IKE)
		return 2 * scalarSize;
	else if (format == kPGPPublicKeyMessageFormat_X509) {
		/* SEQUENCE, length, INT, INT */
		PGPUInt32 len;
		len = 2*(pgpBnX509LenLen(scalarSize+1) + 1 + scalarSize+1);
		return 1 + pgpBnX509LenLen(len) + len;
	}
	pgpAssert(0);
	return 0;
}

static int
ecSign(PGPSecKey *seckey, PGPHashVTBL const *h, PGPByte const *hash,
	PGPByte *sig, PGPSize *siglen, PGPRandomContext const *rc,
	PGPPublicKeyMessageFormat format)
{
#if PGP_SIGN_DISABLE /* [ */

	(void)seckey;
	(void)h;
	(void)hash;
	(void)sig;
	(void)siglen;
	(void)rc;
	(void)format;
	return kPGPError_FeatureNotAvailable;

#else

	ECsecPlus *sec = (ECsecPlus *)seckey->priv;
	ecContextRef ec = sec->s.ec;
	PGPUInt32 coordSize, scalarSize, pointSize;
	BigNum r, s, bn, k, bnx, bnorder;
	ecScalarRef eck=NULL;
	ecPointRef kg=NULL;
	unsigned t;
	int i;
	PGPRandomContext *rc2 = NULL;
	PGPMemoryMgrRef		mgr	= NULL;
	PGPCipherAlgorithm	cipherAlg;
	
	mgr	= PGPPeekContextMemoryMgr( seckey->context );

	ASSERTECSIG(seckey->pkAlg);

	ecGetBufferSize( ec, &coordSize, &scalarSize, NULL, &pointSize );

	/* Allow generalizations of SHA as long as they are big enough */
#if 0
	pgpAssert(h->algorithm == kPGPHashAlgorithm_SHA);
#else
//	pgpAssert(h->hashsize >= scalarSize);
#endif

	if (sec->locked)
		return kPGPError_KeyIsLocked;

	/*
	 * Of these values, only k is inherently sensitive, but others may
	 * hold some intermediate results we would prefer not to have leaked.
	 * So mark all as sensitive.
	 */
	bnBegin(&r, mgr, SECURE );
	bnBegin(&s, mgr, SECURE );
	bnBegin(&bn, mgr, SECURE );
	bnBegin(&k, mgr, SECURE );
	bnBegin(&bnx, mgr, SECURE );
	bnBegin(&bnorder, mgr, INSECURE );

	if( ecScalarCreate( ec, &eck, SECURE ) < 0
		|| ecPointCreate( ec, &kg ) < 0
		|| sScalartoBN( sec->s.order, &bnorder, mgr, scalarSize ) < 0
		|| sScalartoBN( sec->s.x, &bnx, mgr, scalarSize ) < 0 )
		goto nomem;

	/*
	 * EC requires a secret k.  This k is *very* important
	 * to keep secret.  Consider, the EC signing equations are:
	 * r = x-coordinate of k*G mod n
	 * s = k^-1 * (H(m) + x*r) mod q,
	 * so if you know k (and, the signature r, s and H), then
	 * x = r^-1 * (k*s - H(m))
	 * If we ever pick two k values the same, then
	 * r = x-coordinate of k*G mod n is the same for both signatures, and
	 * s1 = k^-1 * (H1 + x * r) 
	 * s2 = k^-1 * (H2 + x * r) 
	 * k = (H1-H2) / (s1-s2)
	 * and proceed from there.
	 *
	 * So we need to make *very* sure there's no problem.  To make
	 * sure, we add a layer on top of the passed-in RNG.  We assume
	 * the passed-in RNG is good enough to never repeat (not a
	 * difficult task), and apply an additional X9.17 generator on
	 * top of that, seeded with the secret x, which is destroyed
	 * before leaving this function.
	 *
	 * In addition, we add entropy from the hash to the original RNG.
	 * This will prevent us from using the same k value twice if the
	 * messages are different.
	 */
	pgpRandomAddBytes(rc, hash, h->hashsize);
	
	if( pgpFIPSModeEnabled() )
	{
		cipherAlg = kPGPCipherAlgorithm_3DES;
	}
	else
	{
		cipherAlg = kPGPCipherAlgorithm_CAST5;
	}
	
	rc2 = pgpRandomCreateX9_17( rc->context, cipherAlg, rc);
	if (!rc2)
		goto nomem;
	pgpRandomBnSeed(rc2, &bnx);

	/*
	 * Choose the random k value to be used for this signature.
	 * Make it a bit bigger than order so it is fairly uniform mod order.
	 */
	if (pgpBnGenRand(&k, rc2, 8*(scalarSize+8), 0, 1, 8*scalarSize) < 0
	    || bnMod(&k, &k, &bnorder) < 0
		|| sBNtoScalar( &k, eck, mgr, scalarSize ) < 0 )
		goto nomem;
	
	/* Multiply k*G and get r as x-coord */
	if( ecPointMul( sec->s.g, eck, FALSE, kg ) < 0
		|| sPointXtoBN( kg, &r, mgr, coordSize ) < 0
		|| bnMod( &r, &r, &bnorder ) < 0 )
		goto nomem;
	      
	/* r*x mod q into s */
	if (bnMul(&s, &r, &bnx) < 0 ||
	    bnMod(&s, &s, &bnorder) < 0)
		goto nomem;

	/* Pack message hash M into buffer bn */
	if (bnInsertBigBytes(&bn, hash, 0, h->hashsize) < 0
		|| bnMod(&bn, &bn, &bnorder) < 0)
		goto nomem;

	/* Add into s */
	if (bnAdd(&s, &bn) < 0 ||
	    bnMod(&s, &s, &bnorder) < 0)
		goto nomem;

	/* Divide by k, mod q (k inverse held in bn) */
	if (bnInv(&bn, &k, &bnorder) < 0 ||
	    bnMul(&s, &s, &bn) < 0 ||
	    bnMod(&s, &s, &bnorder) < 0)
		goto nomem;

	/* That's it, now to pack r and then s into the buffer */
	t = 0;
	if (format == kPGPPublicKeyMessageFormat_X509) {
		/* Put in SEQUENCE header for 509 sig data */
		PGPUInt32 len_seq, lenlen_seq;
		/* Count size of sequence, counting a 0 byte if hi bit is set */
		if (8*scalarSize == bnBits(&r))
			len_seq = pgpBnX509LenLen(scalarSize+1) + 1 + scalarSize+1;
		else
			len_seq = pgpBnX509LenLen(scalarSize) + 1 + scalarSize;
		if (8*scalarSize == bnBits(&s))
			len_seq += pgpBnX509LenLen(scalarSize+1) + 1 + scalarSize+1;
		else
			len_seq += pgpBnX509LenLen(scalarSize) + 1 + scalarSize;
		lenlen_seq = pgpBnX509LenLen(len_seq);
		sig[t++] = X509_TAG_SEQUENCE | X509_TAG_CONSTRUCTED;
		if (--lenlen_seq == 0) {
			sig[t++] = len_seq;
		} else {
			sig[t++] = 0x80 | lenlen_seq;
			len_seq <<= 8 * (4-lenlen_seq);
			while (lenlen_seq--) {
				sig[t++] = (PGPByte)(len_seq >> 24);
				len_seq <<= 8;
			}
		}
	}
	t += pgpBnPutFormatted(&r, sig+t, scalarSize, format);
	t += pgpBnPutFormatted(&s, sig+t, scalarSize, format);
	if (siglen)
		*siglen = (PGPSize)t;

	i = 0;
	goto done;

nomem:
	i = kPGPError_OutOfMemory;
	/* fall through */
done:
	bnEnd(&k);
	bnEnd(&bn);
	bnEnd(&s);
	bnEnd(&r);
	bnEnd(&bnx);
	bnEnd(&bnorder);
	if( IsntNull( eck ) )
		ecScalarFree( eck );
	if( IsntNull( kg ) )
		ecPointFree( kg );
	if( IsntNull( rc2 ) ) {
		pgpRandomDestroy(rc2);
	}
	return i;

#endif /* PGP_SIGN_DISABLE */ /* ] */
}


/*
 * Re-encrypt a PGPSecKey to turn a PGPSecKey into a secret key.
 * A secret key is, after a non-specific prefix:
 *  0       1    Version (= 2 or 3)
 *  1       4    Timestamp
 *  5       2    Validity (=0 at present)
 *  7       1    Algorithm (=1 for EC)
 * The following:
 *  0      2    Curve type (=0)
 *  2      2+s  Curve name (s=strlen(name))
 * 4+s     2+p  Public curve point (p=pointSize)
 * 6+s+p   e    Encryption algorithm data, S2K info
 * 6+s+p+e 2+w  MPI for x, secret part (w=scalarSize)
 * 8+s+p+e+w
 *
 */

#if PGP_MACINTOSH
#pragma global_optimizer on
#endif
static int
ecChangeLock(PGPSecKey *seckey, PGPEnv const *env, 
	PGPRandomContext const *rc, char const *ophrase, PGPSize oplen,
	PGPBoolean oHashedPhrase, char const *phrase, PGPSize plen,
	PGPStringToKeyType s2ktype)
{
	ECsecPlus *sec = (ECsecPlus *)seckey->priv;
	ecContextRef ec = sec->s.ec;
	PGPUInt32 coordSize, scalarSize, pointSize;
	PGPStringToKey *s2k = NULL;	/* Shut up warnings */
	PGPCipherVTBL const *cipher = NULL;	/* Shut up warnings */
	PGPCFBContext *cfb = NULL;	/* This is realy needed */
	BigNum bnx, bny;
	PGPByte *p;
	PGPByte key[PGP_CIPHER_MAXKEYSIZE];
	PGPBoolean oldf = 0;				/* Shut up warnings */
	PGPInt32 curve;
	PGPSize len;
	PGPSize namelen;
	PGPUInt32 checksum;
	PGPContextRef context = pgpenvGetContext( env );
	PGPMemoryMgrRef	mgr	= PGPPeekContextMemoryMgr( seckey->context );
	
	ASSERTEC(seckey->pkAlg);
	if (sec->locked)
		if( IsPGPError(ecUnlock( seckey, ophrase, oplen, oHashedPhrase )) )
			return kPGPError_KeyIsLocked;

	ecGetBufferSize( ec, &coordSize, &scalarSize, NULL, &pointSize );

	curve = sCurveIndexFromScalarsize( scalarSize );
	pgpAssert( curve >= 0 );
	namelen = strlen(sCurves[curve].name);

	bnBegin( &bny, mgr, INSECURE );
	bnBegin( &bnx, mgr, INSECURE );

	sPointtoBN( sec->s.y, &bny, mgr, pointSize );
	sScalartoBN( sec->s.x, &bnx, mgr, scalarSize );

	len = 11 + namelen + bnBytes(&bny) + bnBytes(&bnx);

	if (phrase) {
		s2k = pgpS2Kcreate(context, rc, s2ktype);
		if (!s2k) {
			bnEnd( &bnx );
			bnEnd( &bny );
			return kPGPError_OutOfMemory;
		}
		cipher = pgpCipherDefaultKey(env);
		pgpAssert(cipher);
		if (!cipher) {
			bnEnd( &bnx );
			bnEnd( &bny );
			pgpS2Kdestroy(s2k);
			return kPGPError_OutOfMemory;
		}
		len += cipher->blocksize;
		cfb = pgpCFBCreate( mgr, cipher);
		if (!cfb) {
			bnEnd( &bnx );
			bnEnd( &bny );
			pgpS2Kdestroy(s2k);
			return kPGPError_OutOfMemory;
		}
		oldf = pgpS2KisOldVers(s2k);
		if (!oldf)
			len += 1 + s2k->encodelen;
	}
	if (len > sec->ckalloc) {
		PGPError err = kPGPError_NoErr;
		if( IsNull( sec->cryptkey ) ) {
			sec->cryptkey = (PGPByte *)
				pgpContextMemAlloc( sec->context, len, 0 );
			if( IsNull( sec->cryptkey ) ) {
				err = kPGPError_OutOfMemory;
			}
		} else {
			err = pgpContextMemRealloc( sec->context,
				(void **)&sec->cryptkey, len, 0 );
		}
		if( IsPGPError( err ) ) {
			bnEnd( &bnx );
			bnEnd( &bny );
			PGPFreeCFBContext(cfb);
			pgpS2Kdestroy(s2k);
			return err;
		}
		sec->ckalloc = (PGPSize)len;
	}
	sec->cklen = len;
	p = sec->cryptkey;

	/* Okay, no more errors possible!   Start installing data */
	
	*p++ = 0;
	*p++ = 0;

	/* Assume namelen has 2nd to highest bit set, typical of ascii */
	*p++ = ((8*namelen-1) >> 8) & 0xff;
	*p++ = ((8*namelen-1) >> 0) & 0xff;
	pgpCopyMemory( sCurves[curve].name, p, namelen );
	p += namelen;

	p += pgpBnPutPlain(&bny, p);

	/* Encryption parameters */
	if (!phrase) {
		*p++ = 0;	/* Unencrypted */
	} else {
		if (oldf) {
			*p++ = cipher->algorithm;
		} else {
			*p++ = 0xff;
			*p++ = cipher->algorithm;
			memcpy(p, s2k->encoding, s2k->encodelen);
			p += s2k->encodelen;
		}
		/* Create IV */
		pgpRandomGetBytes(rc, p, cipher->blocksize);
		pgpStringToKey(s2k, phrase, plen, key, cipher->keysize);
		PGPInitCFB(cfb, key, p);
		pgpS2Kdestroy(s2k);
		p += cipher->blocksize;
		/* Wipe key *immediately* */
		pgpClearMemory( key,  cipher->keysize);
	}

	/* Now install x, encrypted */
	checksum = 0;
	p += pgpBnPut(&bnx, p, cfb, &checksum, FALSE);
	pgpChecksumPut(checksum, p, cfb, FALSE);
	p += 2;
	pgpAssert((ptrdiff_t)len == p - sec->cryptkey);

	if (cfb)
		PGPFreeCFBContext(cfb);
	bnEnd(&bnx);
	bnEnd(&bny);

	return 0;	/* Success */
}
#if PGP_MACINTOSH
#pragma global_optimizer reset
#endif

static PGPSize
ecSecBufferLength(PGPSecKey const *seckey)
{
	ECsecPlus const *sec = (ECsecPlus *)seckey->priv;

	return sec->cklen;
}

static void
ecSecToBuffer(PGPSecKey const *seckey, PGPByte *buf)
{
	ECsecPlus const *sec = (ECsecPlus *)seckey->priv;

	memcpy(buf, sec->cryptkey, sec->cklen);
}


static int
ecSecParams(PGPSecKey const *seckey,
	BigNum *x1, BigNum *x2, BigNum *x3, BigNum *x4, BigNum *x5, BigNum *x6,
	BigNum *x7, BigNum *x8)
{
	ECsecPlus const *sec = (ECsecPlus *)seckey->priv;

	ASSERTEC( seckey->pkAlg );
	pgpAssert( !sec->locked );
	(void) sec;
	(void) x1;
	(void) x2;
	(void) x3;
	(void) x4;
	(void) x5;
	(void) x6;
	(void) x7;
	(void) x8;
	return kPGPError_PublicKeyUnimplemented;
}


/* Fill in secret key structure */
static void
ecFillSecKey(PGPSecKey *seckey, ECsecPlus *sec)
{
	seckey->pkAlg	            = kPGPPublicKeyAlgorithm_ECSign;
	seckey->priv	            = sec;
	seckey->destroy             = ecSecDestroy;
	seckey->pubkey              = ecPubkey;
	seckey->islocked            = ecIslocked;
	seckey->lockingalgorithm    = ecLockingAlgorithm;
	seckey->s2ktype             = ecS2KType;
	seckey->convertpassphrase   = ecConvertPassphrase;
	seckey->unlock              = ecUnlock;
	seckey->lock                = ecLock;
	seckey->decrypt             = ecDecrypt;
	seckey->maxdecrypted        = ecSecMaxdecrypted;
	seckey->maxsig              = ecSecMaxsig;
	seckey->maxesk              = ecSecMaxesk;
	seckey->sign                = ecSign;
	seckey->changeLock          = ecChangeLock;
	seckey->bufferLength        = ecSecBufferLength;
	seckey->toBuffer            = ecSecToBuffer;
	seckey->secparams			= ecSecParams;
	seckey->setkeyid			= ecSecSetKeyID;
}


PGPSecKey *
ecSecFromBuf(
	PGPContextRef	context,
	PGPByte const *	buf,
	PGPSize			size,
	PGPError *		error)
{
	PGPSecKey *seckey;
	ECsecPlus *sec;
	PGPByte *cryptk;
	PGPError	err	= kPGPError_OutOfMemory;
	PGPMemoryMgrRef		mgr	= PGPPeekContextMemoryMgr( context );

	bnInit();
	cryptk = (PGPByte *)pgpContextMemAlloc(context,
		size, kPGPMemoryMgrFlags_Clear);
	if (cryptk) {
		sec = (ECsecPlus *)PGPNewSecureData( mgr, sizeof(*sec), 0 );
		if (sec) {
			pgpClearMemory( sec, sizeof(*sec) );
			sec->context	= context;
			seckey = (PGPSecKey *)
				pgpContextMemAlloc(context,
					sizeof(*seckey), kPGPMemoryMgrFlags_Clear);
			if (seckey) {
				seckey->context	= context;
				
				memcpy(cryptk, buf, size);
				sec->cryptkey = cryptk;
				sec->cklen = sec->ckalloc = size;
				sec->locked = 1;
				/* We only need this to try unlocking... */
				seckey->pkAlg = kPGPPublicKeyAlgorithm_ECSign;
				seckey->priv = sec;
				
				if (ecUnlock(seckey, NULL, 0, FALSE) >= 0) {
					ecFillSecKey(seckey, sec);
					*error = 0;
					return seckey;	/* Success! */
				}

				/* Ka-boom.  Delete and free everything. */
				pgpClearMemory( cryptk,  size);
				pgpContextMemFree( context, seckey);
			}
			PGPFreeData( sec );			/* Wipes as it frees */
		}
		pgpContextMemFree(context, cryptk);
	}
	*error = err;
	return NULL;
}




/*
 * Generate an EC secret key with prime of the specified number of bits.
 * We only support the pre-computed sizes, with fixed curves.
 * Make callbacks to progress function periodically.
 * Secret key is returned in the unlocked form, with no passphrase set.
 * fastgen has no effect.
 */
PGPSecKey *
ecSecGenerate(
	PGPContextRef	context,
	unsigned bits, PGPBoolean fastgen,
	PGPRandomContext const *rc,
	int progress(void *arg, int c), void *arg, PGPError *error)
{
	PGPSecKey *seckey	= NULL;
	ECsecPlus *sec;
	BigNum bnorder, bnx;
	ecContextRef ec = NULL;
	PGPInt32 curve;
	PGPUInt32 coordSize, scalarSize, pointSize;
	PGPByte *xbuf = NULL;
	PGPMemoryMgrRef		mgr	= PGPPeekContextMemoryMgr( context );
	PGPEnv *			pgpEnv = pgpContextGetEnvironment( context );

	(void) fastgen;

	*error = kPGPError_NoErr;

	curve = sCurveIndexFromBitsize( bits );
	if( curve < 0 )
	{
		*error = kPGPError_BadParams;
		return NULL;
	}

	/* Initialize local pointers (simplify cleanup below) */
	seckey = NULL;
	sec = NULL;
	
	/* Allocate data structures */
	seckey = (PGPSecKey *)pgpContextMemAlloc( context,
		sizeof(*seckey), kPGPMemoryMgrFlags_Clear);
	if (!seckey)
		goto memerror;
	seckey->context	= context;
	sec = (ECsecPlus *)PGPNewSecureData( mgr, sizeof(*sec), 0 );
	if (!sec)
		goto memerror;
	pgpClearMemory( sec, sizeof(*sec) );
	sec->context	= context;
	
	bnBegin( &bnorder, mgr, INSECURE );
	bnBegin( &bnx, mgr, SECURE );

	if( ecCreate2mContext( mgr, sCurves[curve].bitsize,
						   EC_MEM_USAGE_HIGH, &sec->s.ec ) < 0 )
		goto memerror;
	ec = sec->s.ec;

	ecSetEC2mParamAInt( ec, sCurves[curve].a );
	ecSetEC2mParamB( ec, sCurves[curve].b );

	ecGetBufferSize( ec, &coordSize, &scalarSize, NULL, &pointSize );

	if( ecPointCreate( ec, &sec->s.y ) < 0
		|| ecPointCreate( ec, &sec->s.g ) < 0
		|| ecScalarCreate( ec, &sec->s.order, INSECURE ) < 0
		|| ecScalarCreate( ec, &sec->s.x, SECURE ) < 0 )
		goto memerror;
	sec->s.cofactor = sCurves[curve].cofactor;

	/* Use fixed field and generator */
	if ( ecPointInsertBytes( sec->s.g, sCurves[curve].g, 0 ) < 0
		|| bnInsertBigBytes( &bnorder, sCurves[curve].order,
							 0, scalarSize ) < 0
		|| sBNtoScalar( &bnorder, sec->s.order, mgr, scalarSize ) < 0 )
		goto memerror;

	/* Choose the random x scalar for the secret key */
	xbuf = PGPNewSecureData( mgr, scalarSize+8, 0 );
	if( IsNull( xbuf ) )
		goto memerror;
	pgpRandomGetBytes( rc, xbuf, scalarSize+8 );
	if( bnInsertBigBytes( &bnx, xbuf, 0, scalarSize+8 ) < 0
		|| bnMod( &bnx, &bnx, &bnorder ) < 0
		|| sBNtoScalar( &bnx, sec->s.x, mgr, scalarSize ) < 0 )
		goto memerror;

	if (progress != NULL)
		progress(arg, '.');

	/* And calculate x*g as public key */
	if( ecPointMul( sec->s.g, sec->s.x, FALSE, sec->s.y ) < 0 )
		goto memerror;

	if (progress != NULL)
		progress(arg, ' ');

	/* Fill in structs */
	sec->cryptkey = NULL;
	sec->ckalloc = sec->cklen = 0;
	sec->locked = 0;
	ecFillSecKey(seckey, sec);

	/* Fill in cryptkey structure, unencrypted */
	if( ecChangeLock (seckey, pgpEnv, NULL, NULL, 0, FALSE, NULL, 0,
					  kPGPStringToKey_Simple) < 0 )
		goto memerror;

	goto done;

memerror:
	if ( IsntNull( seckey ) )
		pgpContextMemFree( context, seckey);
	if ( IsntNull( sec ) ) {
		if( IsntNull( sec->s.y ) )
			ecPointFree( sec->s.y );
		if( IsntNull( sec->s.g ) )
			ecPointFree( sec->s.g );
		if( IsntNull( sec->s.x ) )
			ecScalarFree( sec->s.x );
		if( IsntNull( sec->s.order ) )
			ecScalarFree( sec->s.order );
		PGPFreeData( sec );			/* Wipes as it frees */
	}
	if( IsntNull( ec ) )
		ecFreeContext( ec );
	seckey = NULL;
	*error = kPGPError_OutOfMemory;
	/* Fall through */
done:
	bnEnd( &bnorder );
	bnEnd( &bnx );
	if( IsntNull( xbuf ) )
		PGPFreeData( xbuf );

	return seckey;
}


#endif /* PGP_EC */
