/*____________________________________________________________________________
	Copyright (C) 2002 PGP Corporation
	All rights reserved.

	$Id: pgp-agent.c,v 1.11 2002/08/06 20:11:19 dallen Exp $
____________________________________________________________________________*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/time.h>
#include <pthread.h>
#include <signal.h>
#include <pwd.h>
#include <errno.h>

#include "pgpParseCommandLine.h"
#include "pgpPFLErrors.h"
#include <pgpUtilities.h>
#include <pgpThreads.h>
#include <pgpRPCMsg.h>

int					debug				= FALSE;
int					notimeout			= FALSE;
extern PGPMutex_t	sRPCMutex;
extern PGPBoolean	haveSomethingCached;

int					numConnections		= 0;
PGPMutex_t			numConnectionsMutex;

PGPBoolean			running				= TRUE;

pthread_t			mainThread;

	static PGPError
sRead(
	PGPInt32		fd,
	void *			data,
	PGPUInt32		size )
{
	PGPInt32		retval		= 0;
	PGPUInt32		bytesRead	= 0;
	do
	{
		retval = read( fd, data + bytesRead, size - bytesRead );
		if( retval < 0 )
			return kPGPError_ReadFailed;
		else if( retval == 0 )
			return kPGPError_EOF;
		bytesRead += retval;
	} while( bytesRead < size );

	return kPGPError_NoErr;
}

	static PGPError
sWrite(
	PGPInt32		fd,
	void *			data,
	PGPUInt32		size )
{
	PGPError		err				= kPGPError_NoErr;
	PGPInt32		retval			= 0;
	PGPUInt32		bytesWritten	= 0;
	do
	{
		retval = write( fd, data + bytesWritten, size - bytesWritten );
		if( retval < 0 )
			return kPGPError_WriteFailed;
		bytesWritten += retval;
	} while( bytesWritten < size );

	return kPGPError_NoErr;
}

void *
sClientConnection( void *arg )
{
	PGPError		err			= kPGPError_NoErr;
	int				asock		= (int) arg;
	int				n			= 0;
	char			*buf		= NULL;
	char			*next		= NULL;
	PGPPackMsg		pkt;
	PGPInt32		length		= 0;
	PGPInt32		currentSize	= 0;
	const PGPInt32	headerSize	= sizeof( PGPInt32 ) * 3;
	PGPContextRef	context		= kInvalidPGPContextRef;
	char			*outbuf		= NULL;
	int				outbuflen	= 0;
	char			*bufcopy	= NULL;

	printf( "Client connection (asock=%d)\n", asock );

	err = PGPNewContext( kPGPsdkAPIVersion, &context );
	if( IsPGPError( err ) )
		goto out;

	buf = (char *) PGPNewSecureData( PGPPeekContextMemoryMgr( context ), headerSize, 0 );
	if( buf == NULL )
		goto out;

	currentSize = headerSize;

	while( TRUE )
	{
		/* Read header */
		err = sRead( asock, buf, headerSize );
		if( IsPGPError( err ) )
			goto out;

		pkt.ptr = 0;
		pkt.length = 0;
		pkt.base = buf;

		/* skip first two ints in header */
		(void) unpack_int32( &pkt );
		(void) unpack_int32( &pkt );
		length = unpack_int32( &pkt );

		if( length != 0 )
		{
			if( length > currentSize )
			{
				currentSize = length;
				err = PGPReallocData( PGPPeekContextMemoryMgr( context ),
					(void **)&buf, currentSize, kPGPMemoryMgrFlags_Clear );
				if( IsPGPError( err ) )
					goto out;
			}
		}
		else
			goto out;

		if( length > headerSize )
		{
			next = buf;
			next += headerSize;
			err = sRead( asock, next, length - headerSize );
			if( IsPGPError( err ) )
				goto out;
		}
		PGPMutexLock( &sRPCMutex );

		rcv_generic( buf, n, (PGPByte **)&outbuf, &outbuflen );

		if( outbuflen < currentSize )
		{
			memcpy( buf, outbuf, outbuflen );
		}
		else
		{
			currentSize = outbuflen;
			if( PGPReallocData( PGPPeekContextMemoryMgr( context ), (void **)&buf, currentSize, 0 ) != kPGPError_NoErr )
			{
				PGPMutexUnlock( &sRPCMutex );	
				goto out;
			}
			memcpy( buf, outbuf, outbuflen );
		}
		if( !PGPIsInitRPCBuf( outbuf ) )
			PGPFreeData( outbuf );

		PGPMutexUnlock( &sRPCMutex );

		err = sWrite( asock, buf, outbuflen );
		if( IsPGPError( err ) )
			goto out;
	}

out:
	printf( "Client disconnected (asock=%d)\n", asock );

	PGPMutexLock( &numConnectionsMutex );
	numConnections--;
	PGPMutexUnlock( &numConnectionsMutex );

	if( buf != NULL )
		PGPFreeData( buf );

	PGPMutexLock( &sRPCMutex );

	if( context != kInvalidPGPContextRef )
		PGPFreeContext( context );

	PGPMutexUnlock( &sRPCMutex );

	close( asock );

	return;
}

void sSigIntHandler( int sig )
{
	pthread_t	me;

	printf( "Caught SIGINT.  Exiting.\n" );
	/* Workaround for these threading signal mask problems I'm seeing */
	me = pthread_self();

#if !PGP_UNIX_HPUX
	/* HP-UX's pthreads implementation is retarded */
	if( pthread_equal( mainThread, me ) )
		return;
#endif

	running = FALSE;
}

void sSigTermHandler( int sig )
{
	printf( "Caught SIGTERM.  Exiting.\n" );

	running = FALSE;
}

static PGPError
sMakeSockPath( char *sockpath )
{
	struct passwd *pw;

	pw = getpwuid( getuid() );
	sprintf( sockpath, "/tmp/.pgp-agent-%s-%d", pw->pw_name, getuid() );

	return kPGPError_NoErr;
}

int
main( int argc, char *argv[] )
{
	int							sock;
	int							n;
	struct sockaddr_un			sunaddr;
	fd_set						readfds;
	pthread_t					thread;
	char						sockpath[PATH_MAX];
	PGPUInt32					i			= 0;
	PGPCommandLineContextRef	clCon		= kInvalidPGPCommandLineContextRef;
	PGPError					err			= kPGPError_NoErr;
	PGPUInt32					flag		= 0;

	mainThread = pthread_self();

	/* Don't barf on SIGPIPE */
	signal( SIGPIPE, SIG_IGN );
	signal( SIGTERM, sSigTermHandler );
	signal( SIGINT, sSigIntHandler );

	PGPMutexCreate( &sRPCMutex, NULL );
	PGPMutexCreate( &numConnectionsMutex, NULL );

	pgpRPCCreateBEContext(); /* does PGPsdkInit() too..  */

	err = PGPNewCommandLineContext( PGPGetDefaultMemoryMgr(),
		argc, argv, NULL, 0, &clCon );
	if( IsntPGPError( err ) )
	{
		while( IsntPGPError( PGPCommandLineNextFlag( clCon, &flag ) ) )
		{
			switch( flag )
			{
				case 'd':		debug = TRUE;		break;
				case 'n':		notimeout = TRUE;	break;
				default:							break;
			}
		}
		(void) PGPFreeCommandLineContext( clCon );
	}

	(void) sMakeSockPath( sockpath );
	if( debug )
		printf( "Listening on %s\n", sockpath );
 
	sock = socket( PF_UNIX, SOCK_STREAM, 0 );
	if( sock == -1 )
	{
		perror( "socket(AF_UNIX)" );
		return EXIT_FAILURE;
	}

	bzero( &sunaddr, sizeof( sunaddr ) );
	sunaddr.sun_family = AF_UNIX;
	strcpy( sunaddr.sun_path, sockpath );

	n = connect( sock, (struct sockaddr *) &sunaddr, sizeof( sunaddr ) );
	if( n >= 0 )
	{
		/* pgp-agent already running */
		printf( "pgp-agent is already running.\n" );
		return EXIT_FAILURE;
	}

	unlink( sockpath );

	n = bind( sock, (struct sockaddr *) &sunaddr, sizeof( sunaddr ) );
	if( n == -1 )
	{
		perror( "bind" );
		return EXIT_FAILURE;
	}

	/* Only the owner can read/write from this socket */
	chmod( sockpath, 0600 );

	listen( sock, 5 );

	/* Become a daemon */
	if( !debug )
	{
		if( fork() != 0 )
			exit( EXIT_SUCCESS );
		(void) setsid();

		close( 0 );
		close( 1 );
		close( 2 );
	}

	printf( "Waiting for connections (sock=%d)\n", sock );

	while( running )
	{
		PGPInt32		nfds		= 0;
		struct timeval	tv			= { 20, 0 };

		FD_ZERO( &readfds );
		FD_SET( sock, &readfds );

		nfds = select( sock + 1, &readfds, NULL, NULL, &tv );
		if( ( nfds == 0 ) && !notimeout && !haveSomethingCached )
		{
			/* Timeout */
			PGPMutexLock( &numConnectionsMutex );
			if( numConnections == 0 )
			{
				PGPMutexUnlock( &numConnectionsMutex );
				printf( "No connections.  Timing out.\n" );
				break;
			}
			PGPMutexUnlock( &numConnectionsMutex );
		}
		if( nfds == -1 )
		{
			if( running == FALSE )
				break;

			if( errno == EINTR )
				continue;
			else
			{
				perror( "select" );
				break;
			}
		}

		/* Check for new connection */
		if( FD_ISSET( sock, &readfds ) )
		{
			PGPInt32			fromlen	= 0;
			PGPInt32			rv		= 0;
			PGPInt32			asock	= 0;

			fromlen = sizeof( sunaddr );
			asock = accept( sock, (struct sockaddr *) &sunaddr, &fromlen );
			if( asock == -1 )
			{
				perror( "accept()" );
				return EXIT_FAILURE;
			}
			PGPMutexLock( &numConnectionsMutex );
			numConnections++;
			PGPMutexUnlock( &numConnectionsMutex );

#if PGP_UNIX_HPUX
			pthread_create( &thread, pthread_attr_default, sClientConnection, (void *) asock );
			pthread_detach( &thread );
#else
			pthread_create( &thread, NULL, sClientConnection, (void *) asock );
			pthread_detach( thread );
#endif
		}
	}

	PGPMutexDestroy( &sRPCMutex );
	PGPMutexDestroy( &numConnectionsMutex );
	pgpRPCCleanupBEContext();

	unlink( sockpath );

	printf( "Exiting\n" );

	return EXIT_SUCCESS;
}

