/*____________________________________________________________________________
	Copyright (C) 2002 PGP Corporation
	All rights reserved.

	$Id: pgpLicenseNumberNet.c,v 1.16 2002/11/12 21:40:16 ajivsov Exp $
____________________________________________________________________________*/

#include "pgpLicenseNumber.h"
#include "pgpLicenseNumberNet.h"
#include "pgpUtilities.h"
#include "pgpMem.h"
#include "pgpHash.h"
#include "pgpSymmetricCipher.h"
#include "pgpCBC.h"
#include "pgpKeys.h"
#include "pgpEncode.h"
#include "string.h"
#include "pgpErrors.h"
#include "pgpSockets.h"
#include "pgpBigNum.h"
#include "pgpPublicKey.h"
#include "pgpRandomPool.h"
#include "pgpUnicode.h"

#define kPGPLANet_UrlMax  256
#define kPGPLANet_SendSize 2048
#define kPGPLANet_ReceiveSize 2048

static const char tok_ln[]="ln";
static const char tok_companyName[]="companyName";
static const char tok_customerName[]="customerName";
static const char tok_ticket[]="ticket";
static const char tok_scookie[]="scookie";
static const char tok_ticket_stub[]="ticket_stub";
static const char tok_la[]="la";
static const char tok_nameHash[]="nameHash";

/* controls all debug dump on the stderr */
#define PGP_DEBUG_PRINT PGP_DEBUG

/* implementation of DH DL in the generric API that will work for
   other DH types. The idea is to have alternative to dhInitDL, say dhInitEC, 
   while keeping other dh* functions inchanged */
typedef struct DHExchange {
	PGPContextRef ctx;
	PGPMemoryMgrRef mem;
	PGPBoolean is_initiator;

	PGPSize sizeInTransit;

	PGPBigNumRef x;
	PGPBigNumRef t;
	PGPBigNumRef gy;
	PGPBigNumRef g;
	PGPBigNumRef p;

	PGPByte *buf;
	PGPSize bufSize;

} DHExchange;
typedef DHExchange * pgpDHExchangeRef;

typedef PGPByte byte;
/* IKE group 2 */
/* The prime is 2^1024 - 2^960 - 1 + 2^64 * { [2^894 pi] + 129093 } */
static const PGPByte mod1024bytesG0[] = {
	(byte)0xFF,(byte)0xFF,(byte)0xFF,(byte)0xFF,(byte)0xFF,(byte)0xFF,(byte)0xFF,(byte)0xFF,
	(byte)0xC9,(byte)0x0F,(byte)0xDA,(byte)0xA2,(byte)0x21,(byte)0x68,(byte)0xC2,(byte)0x34,
	(byte)0xC4,(byte)0xC6,(byte)0x62,(byte)0x8B,(byte)0x80,(byte)0xDC,(byte)0x1C,(byte)0xD1,
	(byte)0x29,(byte)0x02,(byte)0x4E,(byte)0x08,(byte)0x8A,(byte)0x67,(byte)0xCC,(byte)0x74,
	(byte)0x02,(byte)0x0B,(byte)0xBE,(byte)0xA6,(byte)0x3B,(byte)0x13,(byte)0x9B,(byte)0x22,
	(byte)0x51,(byte)0x4A,(byte)0x08,(byte)0x79,(byte)0x8E,(byte)0x34,(byte)0x04,(byte)0xDD,
	(byte)0xEF,(byte)0x95,(byte)0x19,(byte)0xB3,(byte)0xCD,(byte)0x3A,(byte)0x43,(byte)0x1B,
	(byte)0x30,(byte)0x2B,(byte)0x0A,(byte)0x6D,(byte)0xF2,(byte)0x5F,(byte)0x14,(byte)0x37,
	(byte)0x4F,(byte)0xE1,(byte)0x35,(byte)0x6D,(byte)0x6D,(byte)0x51,(byte)0xC2,(byte)0x45,
	(byte)0xE4,(byte)0x85,(byte)0xB5,(byte)0x76,(byte)0x62,(byte)0x5E,(byte)0x7E,(byte)0xC6,
	(byte)0xF4,(byte)0x4C,(byte)0x42,(byte)0xE9,(byte)0xA6,(byte)0x37,(byte)0xED,(byte)0x6B,
	(byte)0x0B,(byte)0xFF,(byte)0x5C,(byte)0xB6,(byte)0xF4,(byte)0x06,(byte)0xB7,(byte)0xED,
	(byte)0xEE,(byte)0x38,(byte)0x6B,(byte)0xFB,(byte)0x5A,(byte)0x89,(byte)0x9F,(byte)0xA5,
	(byte)0xAE,(byte)0x9F,(byte)0x24,(byte)0x11,(byte)0x7C,(byte)0x4B,(byte)0x1F,(byte)0xE6,
	(byte)0x49,(byte)0x28,(byte)0x66,(byte)0x51,(byte)0xEC,(byte)0xE6,(byte)0x53,(byte)0x81,
	(byte)0xFF,(byte)0xFF,(byte)0xFF,(byte)0xFF,(byte)0xFF,(byte)0xFF,(byte)0xFF,(byte)0xFF
};

static void dhFree( pgpDHExchangeRef exchange )  {
	if( !exchange )
		return;

	if( exchange->g )
		PGPFreeBigNum( exchange->g );
	if( exchange->x )
		PGPFreeBigNum( exchange->x );
	if( exchange->t )
		PGPFreeBigNum( exchange->t );
	if( exchange->gy )
		PGPFreeBigNum( exchange->gy );
	if( exchange->p )
		PGPFreeBigNum( exchange->p );

	if( exchange->buf )
		PGPFreeData( exchange->buf );
}

#if PGP_DEBUG_PRINT
static void dhPrintBuffer(pgpDHExchangeRef exchange )  {
	PGPUInt32 i;

	for( i=0; i<exchange->bufSize; i++ )  {
		fprintf( stderr, "%02x", exchange->buf[i] );
	}
}
#endif

/* initialize DL DH exchange */
static PGPError dhInitDL( PGPContextRef context, PGPByte group, pgpDHExchangeRef *exchange )  {
	PGPError err;
	PGPByte *secretX = NULL;
	PGPSize secretXsize;
	DHExchange *e;

	(void)group;

	*exchange = NULL;

	e = PGPNewData( PGPPeekContextMemoryMgr(context), sizeof(DHExchange), 0 );
	if( !e )
		return kPGPError_OutOfMemory;
	pgpClearMemory( e, sizeof(*e) );

	e->ctx =	context;
	e->mem = PGPPeekContextMemoryMgr(context);

	e->is_initiator = TRUE;

	err = PGPNewBigNum( e->ctx, TRUE /* secure */, &e->x );

	if( IsntPGPError(err) )
		err = PGPNewBigNum( e->ctx, FALSE, &e->t );
	if( IsntPGPError(err) )
		err = PGPNewBigNum( e->ctx, FALSE, &e->gy );
	if( IsntPGPError(err) )
		err = PGPNewBigNum( e->ctx, FALSE, &e->p );
	if( IsntPGPError(err) )
		err = PGPNewBigNum( e->ctx, FALSE, &e->g );

	/* initialize generator to 2 */
	if( IsntPGPError(err) )
		err = PGPBigNumSetQ( e->g, 2 );

	if( IsntPGPError(err) )
		err = PGPBigNumInsertBigEndianBytes( e->p, mod1024bytesG0, 0, sizeof(mod1024bytesG0) );

	if( IsntPGPError(err) )  {
		e->sizeInTransit = sizeof(mod1024bytesG0);
		e->buf = PGPNewData( e->mem, e->sizeInTransit, 0 );
	}

	if( IsPGPError(err) || e->x==NULL || e->g==NULL || e->p==NULL || 
		e->t==NULL || e->gy==NULL )
	{
		dhFree( e );
		return kPGPError_OutOfMemory;
	}

	(void)PGPDiscreteLogExponentBits( e->sizeInTransit, (PGPUInt32 *)&secretXsize );
	secretXsize = (secretXsize*3/2 + 7) / 8;

	secretX = PGPNewSecureData( e->mem, secretXsize, 0 );
	if( IsNull( secretX ) )	 {
		dhFree(e);
		return kPGPError_OutOfMemory;
	}
	
	(void)PGPGlobalRandomPoolAddSystemState();
	err = PGPContextGetRandomBytes( e->ctx, secretX, secretXsize ); 

	if( IsntPGPError(err) )
		err = PGPBigNumInsertBigEndianBytes( e->x, secretX, 0, secretXsize );

	if( secretX )
		(void)PGPFreeData( secretX ); 

	if( IsntPGPError(err) )
		*exchange = e;

	return err;
}


static PGPError dhImportGY(pgpDHExchangeRef exchange, const PGPByte *gy, PGPSize size )  {
	PGPError err;

	if( !gy || !size)
		return kPGPError_CorruptData;

	while( gy[0] == 0 && size )  {
		gy++;
		size--;
	}

	if( size > exchange->sizeInTransit )
		return kPGPError_CorruptData;
	
	err = PGPBigNumInsertBigEndianBytes( exchange->gy, gy, 0, size );
	if( IsPGPError(err) )
		return err;

	if( PGPBigNumCompareQ( exchange->gy, 0xffff ) < 0 )
		return kPGPError_BadParams;

	// TODO: small group check

	return err;
}

/* Perform DH step 1 with internal structures, i.e. g (op) x, 
 */
static PGPError
dhGenerate1( pgpDHExchangeRef exchange )  
{
	PGPError		err = kPGPError_NoErr;

	err = PGPBigNumExpMod(exchange->g, exchange->x, exchange->p, exchange->t ); 
	if( IsntPGPError(err) ) 
		err = PGPBigNumExtractBigEndianBytes( exchange->t, exchange->buf, 0, exchange->sizeInTransit );

	if( IsntPGPError(err) )
		exchange->bufSize = exchange->sizeInTransit;
		
	return err;
}

/* Perform DH step 2 with internal structures, i.e. (g (op) y) (op) x,
 */
static PGPError
dhGenerate2( pgpDHExchangeRef exchange )  
{
	PGPError		err = kPGPError_NoErr;

	err = PGPBigNumExpMod(exchange->gy, exchange->x, exchange->p, exchange->t ); 
	if( IsntPGPError(err) ) 
		err = PGPBigNumExtractBigEndianBytes( exchange->t, exchange->buf, 0, exchange->sizeInTransit );

	if( IsntPGPError(err) )
		exchange->bufSize = exchange->sizeInTransit;
		
	return err;
}

static PGPByte *dhBuf( pgpDHExchangeRef exchange )  {
	if( !exchange )
		return NULL;
	return exchange->buf;
}
static PGPSize dhBufSize( pgpDHExchangeRef exchange )  {
	if( !exchange )
		return 0;
	return exchange->bufSize;
}

/* add PKCS 5 padding to the null-terminated string */
static PGPInt32 pkcs5Padding( PGPByte *s, PGPSize blockSize )  {
	const PGPSize len = strlen( s )+1;	/* include '\0' */
	PGPByte paddingSize = (PGPByte)(blockSize - len % blockSize);
	PGPUInt32 i = paddingSize;

	s += len;

	while(i--)  
		*(s++) = paddingSize;

	*s = '\0';
	return paddingSize;
}


static PGPError parseHttpUrl( const PGPByte *url, 
	PGPInternetAddress *addr_out, PGPUInt16 *port_out, 
	PGPByte host_port_buf[kPGPLANet_ReceiveSize],
	PGPByte resource_buf[kPGPLANet_ReceiveSize]	)
{
	const static PGPUInt16 defPortInt = 80;
	PGPByte http[] = "http://";
	PGPByte url_buf[256];
	PGPError err = kPGPError_NoErr;
	PGPByte *resource;
	PGPByte *port;
	PGPByte root[] = "/";
	PGPUInt16 portInt;
	PGPHostEntry *host;

	pgpClearMemory( addr_out, sizeof(*addr_out) );
	*port_out = 0;
	resource_buf[0] = '\0';
	host_port_buf[0] = '\0';

	strncpy( url_buf, url, sizeof(url_buf) );
	url_buf[sizeof(url_buf)-1] = '\0';
	url = url_buf;

	if( strnicmp( url, http, sizeof(http)-1 ) != 0 )
		return kPGPError_FeatureNotAvailable;

	url += sizeof(http)-1;

	resource = strchr( url, '/' );

	if( !resource )  
		strcpy( resource_buf, root );
	else
		strcpy( resource_buf, resource );

	*resource = '\0';	/* cut resource part */

	/* cut and initialize the port number, if present */
	port = strchr( url, ':' );
	if( port == NULL )
		portInt = defPortInt;
	else  {
		portInt = atoi( port+1 );
		*port = '\0';
	}

	host = PGPGetHostByName( url );
	sprintf( host_port_buf, "%s:%d", url, portInt );

	if( host==NULL ) {
		resource_buf[0] = '\0';
		return PGPGetLastSocketsError();
	}

	*addr_out = *((PGPInternetAddress*) 
				host->h_addr_list[0]);
	*port_out = portInt;

	return err;
}

static PGPError createSocket( PGPInternetAddress ipAddress, PGPUInt16 port, PGPSocketRef *socket )
{
	PGPSocketRef s;
	PGPSocketAddressInternet address;
	PGPInt32 socketResult;

	*socket = kInvalidPGPSocketRef;

	s = PGPOpenSocket(	kPGPAddressFamilyInternet,
								kPGPSocketTypeStream,
								kPGPTCPProtocol);

	if(s == kInvalidPGPSocketRef) 
		return PGPGetLastSocketsError();

	address.sin_family = kPGPAddressFamilyInternet;
	address.sin_port = PGPHostToNetShort(port);
	address.sin_addr = ipAddress;

	socketResult = PGPConnect( s, (PGPSocketAddress *) &address, sizeof(address));
	if (socketResult == kPGPSockets_Error) {
		PGPCloseSocket(s);
		return PGPGetLastSocketsError();
	}

	*socket = s;

	return kPGPError_NoErr;
}

/*
XXX
static PGPError addParam( PGPByte *buffer, PGPSize totalSize,
						 PGPByte *name, PGPByte *value )
{
	PGPSize l = strlen( buffer );
	PGPSize avail = totalSize - l;

	if( totalSize < l )
		return kPGPError_CorruptData;

	if( strlen(name)+strlen(value)+3 > avail )
		return kPGPError_BufferTooSmall;

	sprintf( buffer+l, "&%s=%s", name, value );

	return kPGPError_NoErr;
}
*/

static const char *kContentLength = "Content-Length: ";
static const char *kWarning = "Warning: ";

// look for HTTP optional header
//    Warning: warn-code SP warn-agent SP warn-text
// and parse it, if it was generated by us. 
static PGPError parseHttpWarningHeader( const PGPByte *header, PGPError *remoteError )  {
	PGPByte *p;

	*remoteError = kPGPError_NoErr;
	
	p = strstr( header, kWarning );
	if( p == NULL )  
		return kPGPError_NoErr;
		
	if( p > header+2 && *(p-1)=='\n' && *(p-2)=='\r' ) {
		p += strlen(kWarning);
		
		// HTTP Warning: warn-code SP warn-agent SP warn-text
		// p now points on warn-code
		
		if( atoi(p) == 99 )  {	// we pass PGPError as this "Miscellaneous warning"
			p = strchr( p, ' ' );
			
			if( p )
				p++;	// p points on warn-agent
			else
				return kPGPError_CorruptData;	// HTTP violation
			
			p = strchr( p, ' ' );	// skip warn-agent
			
			if( !p )
				return kPGPError_CorruptData;	// HTTP violation
			
			{
				PGPByte warn_text[40];
				const static PGPByte pgperror[] = "PGPError=";
				
				p++;	// p points on warn-text, this is quoted PGPError=errcode
				
				if( *p == '"' )  {
					strncpy( warn_text, p+1, sizeof(warn_text) );
					warn_text[sizeof(warn_text)-1] = '\0';
					
					p = strchr( warn_text, '"' );	// check for another quote
					if( p )  
						*p = '\0';
					else
						return kPGPError_CorruptData;	// HTTP violation
					
					p = warn_text;
				}
				
				if( strncmp( p, pgperror, sizeof(pgperror)-1 )==0 )  {
					p += sizeof(pgperror)-1;
					*remoteError = atoi(p);	// set output parameter
				}
			}
		}
	}
		
	return kPGPError_NoErr;
}

/* we assume that the first line in HTTP response is in the buffer
   other fields can be absent
 */
static PGPError parseHttpHeader( PGPByte header[kPGPLANet_ReceiveSize], 
	PGPSize *bodySize, PGPSize *headerSize, PGPUInt32 *statusCode )  
{
	PGPByte *p, *p2;
	PGPUInt32 i;
	int curBodySize = 0;
	int curHeaderSize = 0;
	int curStatusCode = 0;

	*bodySize = 0;
	*headerSize = 0;
	*statusCode = 0;

	/* find where body begins and eliminate it. It may not be present */
	p = strstr( header, "\r\n\r\n" );
	if( p != NULL )  {
		p[2] = '\0';
		curHeaderSize = p-header+4;
	}

	/* check for content lenght */
	p = strstr( header, kContentLength );
	if( p != NULL )  {
		if( p > header+2 && *(p-1)=='\n' && *(p-2)=='\r' ) {
			p += strlen(kContentLength);

			i = atoi(p);

			/* check content length that we could generate */
			if( i > 160*2/8 && i < 4*1024 )  {
				curBodySize = i;
			}
		}
	}

	/* now we are done with sizes - parse first line to get return code */
	p2 = strchr( header, '\r' );	
	if( !p2 )  {	/* we expect to get at least one line */
		return kPGPError_CorruptData;
	}

	// header =?= HTTP-Version SP Status-Code SP Reason-Phrase CRLF 

	if( strncmp( header, "HTTP/", 5 )!=0 )
		return kPGPError_CorruptData;

	p = strchr( header, ' ' );
	if( !p )
		return kPGPError_CorruptData;

	p++;
	i = atoi(p);

	if( i<100 )
		return kPGPError_CorruptData;

	curStatusCode = i;

	p = strchr(p, ' ');
	if( !p )
		return kPGPError_CorruptData;

	p++;
	
	/* p is reason */
	if( strchr(p, '\r') != (char *) p2 )
		return kPGPError_CorruptData;	// this wasn't on the first line

	*bodySize = curBodySize;
	*headerSize = curHeaderSize;
	*statusCode = curStatusCode;

	return kPGPError_NoErr;
}

static void sleepSeconds( PGPUInt32 sec )  {
#if PGP_UNIX || PGP_OSX
	usleep(sec * 1000000);
#elif PGP_WIN32
	Sleep(sec * 1000);
#else
	sleep( sec );
#endif
}

static PGPError roundTrip( PGPSocketRef socket, 
			const PGPByte *host_port, const PGPByte *resource, 
			PGPBoolean last, 
			const PGPByte *inData, 
			PGPByte bufferRead[kPGPLANet_ReceiveSize], 
			PGPUInt32 *httpError, PGPError *remotePgpError ) 
{
	static const char *kPost = "POST ";
	static const char *kHTTPVersion_term = " HTTP/1.1\r\n";
	static const char *kContentType = "Content-type: application/x-www-form-urlencoded\r\n";
	static const char *kConnectionKeepAlive = "Connection: Keep-Alive\r\n";
	static const char *kConnectionClose = "Connection: close\r\n";
	static const char *kHost = "Host: ";

	PGPByte bufferSend[kPGPLANet_SendSize];

#if PGP_MACINTOSH
	#define OS " (Macintosh)"
#elif PGP_WIN32
	#define OS " (Windows)"
#elif PGP_UNIX
	#define OS " (Unix)"
#endif
	
	static const char *	kUserAgent = "User-Agent: PGP" OS "\r\n";
	#undef OS

	PGPInt32 socketResult=0;
	const PGPSize inDataLength = inData != NULL ? strlen( inData ) : 0;

 	bufferSend[0] = '\0';
	bufferRead[0] = '\0';

	if( httpError )
		*httpError = 0;

	if( remotePgpError )
		*remotePgpError = kPGPError_NoErr;

	if( inData!=NULL && inData[0]=='\0' )
		inData = NULL;

	/* send first line */
	strcat( bufferSend, kPost );
	strcat( bufferSend, resource);
	strcat( bufferSend, kHTTPVersion_term);

	strcat( bufferSend, kUserAgent);
	strcat( bufferSend, kContentType);

	/* sending Host field */
	strcat( bufferSend, kHost );
	strcat( bufferSend, host_port );
	strcat( bufferSend, "\r\n" );

	strcat( bufferSend, last ? kConnectionClose : kConnectionKeepAlive );

	/* forth line - content length */
	{
		char s[20];
		strcat(bufferSend, kContentLength );
		sprintf(s, "%u\r\n", (unsigned) inDataLength);
		strcat(bufferSend, s );
	}

	/* insert empty line */
	strcat(bufferSend, "\r\n");

#if PGP_DEBUG_PRINT
	fprintf( stderr,"-----begin sending packet-----\n");
	fprintf( stderr,"%s%s\n", bufferSend, inData==NULL ? (const PGPByte*)" " : inData);
	fprintf( stderr,"-----end sending packet-----\n");
#endif

	PGPSend( socket, bufferSend, strlen(bufferSend), kPGPSendFlagNone );

	/* finally, send data */
	if( inData != NULL )
		PGPSend( socket, inData, inDataLength, kPGPSendFlagNone );

	/* now receive data */
	{
		PGPUInt32 bytesRead		= 0;
		PGPInt32 avail;
		PGPSize headerSize, bodySize; 
		PGPUInt32 httpStatusCode; 
		PGPError err;
		PGPError remote_err;
		PGPInt32 max_wait=5;	/* how long receiver waits before declaring timeout */

		do {
			avail = PGPReceive( socket, bufferRead, kPGPLANet_ReceiveSize, MSG_PEEK );
			if( avail == 0 )  {
				sleepSeconds( 1 );
#if PGP_DEBUG_PRINT
				fprintf( stderr,"\twaited one sec to receive data from the socket...\n");
#endif
			}
		} while(max_wait-- > 0 && avail==0);
		
		if( avail==0 ) { /* timed out */
#if PGP_DEBUG_PRINT
			fprintf( stderr,"timed out: no data in the socket\n");
#endif
			return kPGPError_ServerTimedOut;
		}
		if( avail < 0 )
			return PGPGetLastSocketsError();
		
		err = parseHttpHeader( bufferRead, &bodySize, &headerSize, &httpStatusCode );
		if( IsPGPError(err) )
			return err;

		if( httpError )
			*httpError = httpStatusCode;

		/* check for remote PGPError code returned as a warn-text in Warning:  */
		err = parseHttpWarningHeader( bufferRead, &remote_err );
		if( IsPGPError(err) )
			return err;

		if( remotePgpError )
			*remotePgpError = remote_err;

		if( bodySize && headerSize && httpStatusCode==200 && bodySize + headerSize == avail )  {	/* already read */
			bytesRead = avail;
			PGPReceive( socket, bufferSend, avail, kPGPReceiveFlagNone ); 
				// reuse bufferSend for cleaning the socket
		}
		else  {
			avail = bodySize + headerSize;
			if( avail > kPGPLANet_ReceiveSize  )
				avail = kPGPLANet_ReceiveSize;
			while( bytesRead < avail )
			{
				socketResult = PGPReceive( socket, ( (PGPByte *) bufferRead ) + bytesRead, 
							avail-bytesRead, kPGPReceiveFlagNone );
			
				if( socketResult <= 0 )
					break;
				
				bytesRead += socketResult;
			}
		}

		if( socketResult < 0 )  {
			PGPError err = PGPGetLastSocketsError();
			return err;
		}

		if( headerSize )  {
			bufferRead[headerSize-2] = '\0';
			bufferRead[headerSize-1] = '\0';
			bufferRead[bodySize + headerSize] = '\0';
		}

#if PGP_DEBUG_PRINT
		fprintf( stderr,"-----begin received packet-----\n" );
		fprintf( stderr, "%s\n", bufferRead );
		fprintf( stderr, "-----begin body-----\n" );
		fprintf( stderr, "%s\n", bufferRead+headerSize );
		fprintf( stderr,"-----end received packet-----\n");
#endif	/* PGP_DEBUG_PRINT */

		if( bodySize )
			memmove( bufferRead, bufferRead+headerSize, bodySize+1 );
		else
			bufferRead[0] = '\0';
	}

	return kPGPError_NoErr;
}

/* appends application/x-www-form-urlencoded parameter and value,
   encodes all illegal characters as %XX
   See section 2.2 from RFC 1738
  */
static void addFormUrlEncodedParam( const PGPByte *name, const PGPByte *value, 
										PGPByte *buffer )  
{
	PGPByte v[kPGPLANet_SendSize];
	PGPSize len = strlen( buffer );
	PGPUInt32 w_value_len=0;
	PGPByte *p;
	static const PGPByte illegal[] = "&=+% \t\n\r<>\"#{}|\\^[]~`";
	PGPUInt16 wv[kPGPLANet_SendSize];
	PGPUInt32 i;

	if( name == NULL || value == NULL )
		return;

	/* name expect to be OK */	
	pgpAssert( strchr(name, '=')==NULL );

	pgpUTF8StringToUCS2 (
		value, kPGPUnicodeNullTerminated, wv, sizeof(wv)/sizeof(wv[0]),	&w_value_len);

	if( w_value_len==0 )
		return;	/* ignore the pair */

	p = v;

	for( i=0; i<w_value_len; i++ )  {
		PGPByte utf8[8];
		PGPUInt32 utf8_len=0;
		
		pgpUCS2StringToUTF8( wv+i, 1, utf8, sizeof(utf8), &utf8_len );
		if( utf8_len==0 )
			return;	/* ignore the pair */
		utf8[sizeof(utf8)-1] = '\0';

		if( utf8_len==1 && strchr( illegal, wv[i] )==NULL && wv[i] > 0x1f && wv[i] < 0x80 )
			*(p++) = (PGPByte)wv[i];	/* unchacnged  */
		else  {
			/* encode all bytes for the Unicode character */
			PGPUInt32 j;
			PGPByte e[8];
			for( j=0; j<utf8_len; j++ )  {
				sprintf( e, "%%%02X", utf8[j] );
				strcpy( p, e );
				p += 3;
			}
		}
	}

	*p = '\0';

	if( !len )
		sprintf( buffer, "%s=%s", name, v );
	else  {
		buffer += len;
		sprintf( buffer, "&%s=%s", name, v );
	}
}

/* parse parametes in the form of application/x-www-form-urlencoded */
typedef struct formUrlEncodedParam {
	const PGPByte *tok;
	PGPByte *value;
} formUrlEncodedParam;
static PGPError parseParams( const PGPByte *in, 
	PGPByte *ticket, PGPByte *scookie, PGPByte *ln, PGPByte *companyName, PGPByte *customerName, 
	PGPByte *la, PGPByte *groupid )
{
	PGPByte buf[kPGPLANet_ReceiveSize];
	PGPByte *p;
	const PGPByte *s;
	PGPUInt32 i;
	PGPUInt32 j;
	PGPSize len;
	PGPBoolean more;
	
	formUrlEncodedParam params[] = {
		{ tok_ticket, ticket },
		{ tok_scookie, scookie }, 
		{ tok_ln, ln }, 
		{ tok_companyName, companyName }, 
		{ tok_customerName, customerName },
		{ "la", la },
		{ "group_id", groupid }
	};

	for( i=0; i<sizeof(params)/sizeof(params[0]); i++ )  {
		if( params[i].value )
			params[i].value[0] = '\0';
	}

	/* process parameters */
	s = in;
	more = TRUE;
	while( more )  {
		PGPInt32 k;
		p = strchr( s, '&' );
		if( p != NULL )  {
			more = TRUE;
			*p = '\0';
		}
		else 
			more = FALSE;

		/* s points to the pair, decode it into buf */
		{
			/* remove line breaks and remove encoding */
			i=j=0;
			while( s[i]=='\r' )
				i++;
			len = strlen(s);
			if( len > sizeof(buf) )
				len = sizeof(buf);
			for( ; i<len; i++ )  {
				if( s[i]=='\r' || s[i]=='\n' || s[i]==' ')  
					continue;
				if( s[i]=='%' )  {
					if( len-i >= 3 )  {
						PGPUInt32 bv;
						PGPByte b[] = { s[i+1], s[i+2], '\0' };
						if( sscanf( b, "%02X", &bv ) )
							buf[j++] = (PGPByte)bv;
						i += 2;	/* two extra */
						continue;	/* one more i */
					}
					else	/* remove it */
						break;
				}
				if( s[i]=='+' )  {
					buf[j++] = ' ';
					continue;
				}
				buf[j++] = s[i];
			}
			buf[j] = '\0';
		}

		for( k=0; k<sizeof(params)/sizeof(params[0]); k++ )  {
			if( params[k].value!=NULL && strncmp( buf, params[k].tok, strlen(params[k].tok))==0 &&
				buf[strlen(params[k].tok)]=='=' )  
			{
				strcpy( params[k].value, buf+strlen(params[k].tok)+1 );
			}
		}

		if( more )
			s += strlen(s)+1;
	}

	return kPGPError_NoErr;
}

/* make authorization request, that will be sent to the website as 
   the main payload */
static PGPError makeAuthReq( PGPContextRef context, const PGPLicenseNumberRef ln,
	const PGPByte *customerName,
	const PGPByte *companyName,
	PGPByte buf[kPGPLANet_ReceiveSize] )  
{
	PGPError err = kPGPError_NoErr;
	PGPByte ln_b32[kPGPLicenseNumberMaxSize] = {'\0'};
	PGPLNExportBase32( context, ln, NULL, ln_b32 );

	buf[0] = '\0';
	if( ln_b32[0] != '\0' )  {
		PGPByte *hash_b64 = NULL;
		addFormUrlEncodedParam( tok_ln, ln_b32, buf );
		addFormUrlEncodedParam( tok_customerName, customerName, buf );
		addFormUrlEncodedParam( tok_companyName, companyName, buf );

		err = PGPLAGetNameHash( context, customerName, companyName, &hash_b64 );
		if( IsntPGPError( err ) )
			addFormUrlEncodedParam( tok_nameHash, hash_b64, buf );
			
		if( hash_b64 != NULL )
			PGPFreeData( hash_b64 );
	}
	else  {
		return kPGPError_CorruptData;
	}

	return err;
}

/* creates objects from authorization request */
PGPError pgpLAParseAuthenticationRequest(PGPContextRef context,
	const PGPByte *authReq, PGPLicenseNumberRef *ln, 
	PGPByte *custName, PGPByte *compName )
{
	PGPError err;
	PGPByte ln_b32[kPGPLANet_ReceiveSize];

	err = parseParams( authReq, 
		NULL, NULL, ln_b32, compName, custName, NULL, NULL );
	if( IsntPGPError(err) )  {
		err = PGPLNImportBase32( context, ln_b32, ln );
	}

	return err;
}

/* sends request for license authorization */
PGPError pgpLACreateNet(PGPContextRef context,
	const PGPByte *url, 
	const PGPLicenseNumberRef licenseNumber,
	const PGPByte *customerName,
	const PGPByte *companyName,
	PGPLicenseAuthorizationRef *buffer )
{
	PGPError err;
	PGPInternetAddress addr;
	PGPUInt16 port;
	PGPByte host_port[kPGPLANet_UrlMax];
	PGPByte resource[kPGPLANet_UrlMax];
	PGPByte out[kPGPLANet_ReceiveSize];
	PGPUInt32 httpError=0;
	PGPError remotePgpError=kPGPError_NoErr;
	PGPSocketRef socket = kInvalidPGPSocketRef;
	PGPByte ln_cust_comp_b64[kPGPLANet_SendSize];

	PGPByte ticket_b64[kPGPLANet_ReceiveSize];
	PGPByte scookie_b64[kPGPLANet_ReceiveSize];
	PGPByte IV[32];		/* for CBC IV */

	PGPByte *ticket_stub_b64 = NULL;
	PGPByte *ln_cust_comp_b64_encr_b64 = NULL;

	PGPByte postParams[kPGPLANet_ReceiveSize];

	pgpDHExchangeRef dhExchange = NULL;
	PGPSize tsize;

	PGPSize cipherBlockSize=0;
	PGPSymmetricCipherContextRef symmRef = kInvalidPGPSymmetricCipherContextRef;
	PGPCBCContextRef cbcContext = NULL;

	err = dhInitDL( context, 0 /*group*/, &dhExchange );
	if( IsPGPError(err) )
		return err;

	err = dhGenerate1( dhExchange );
	if( IsPGPError(err) )  {
		dhFree(dhExchange);
		return err;
	}

#if PGP_DEBUG_PRINT
	fprintf(stderr,"DH=");
	dhPrintBuffer( dhExchange );
	fprintf(stderr,"\n");
#endif

	tc_encode_base64( PGPPeekContextMemoryMgr(context), 
		&ticket_stub_b64, &tsize, 
		dhBuf(dhExchange), dhBufSize(dhExchange) );

	makeAuthReq( context, licenseNumber, customerName, companyName, 
		ln_cust_comp_b64 );

	err = parseHttpUrl( url, &addr, &port, host_port, resource );
	if( IsPGPError(err) )  {
		dhFree(dhExchange);
		return err;
	}

	if( IsntPGPError(err) )
		err = createSocket( addr, port, &socket );

	postParams[0] = '\0';
	addFormUrlEncodedParam( "r", tok_ticket, postParams );
	addFormUrlEncodedParam( "group", "0", postParams );

	if( IsntPGPError(err) )
		err = roundTrip( socket, host_port, resource, FALSE, postParams, out, &httpError, &remotePgpError );

	if( httpError == 200 && IsntPGPError(err) && IsntPGPError(remotePgpError) )  {
		err = parseParams( out, ticket_b64, scookie_b64, NULL, NULL, NULL, NULL, NULL );
	}
	else {
		ticket_b64[0] = scookie_b64[0] = '\0';
		if( httpError )  {
#if PGP_DEBUG_PRINT
			fprintf( stderr, "HTTP communication with %s returned %d error code in round 1\n", url, httpError );
#endif
			if( IsPGPError(remotePgpError) )
				err = remotePgpError;
			else
				err = kPGPError_NetworkOpFailed;
		}
	}
	
	if( IsntPGPError(err) ) {
		if( *ticket_b64 == '\0' || *scookie_b64 == '\0' )  
			err = kPGPError_BadParams;
	}

#if PGP_DEBUG_PRINT
	if( IsntPGPError(err) )  {
		fprintf( stderr, "Received (Phase 1):\n" );
		fprintf( stderr, "ticket=%s\n", ticket_b64 );
		fprintf( stderr, "scookie=%s\n", scookie_b64 );
	}
#endif

	if( IsntPGPError(err) )  {
		PGPByte *ticket_bin = NULL;
		PGPSize ticket_bin_size;
		PGPByte *scookie_bin_p=NULL;
		PGPInt32 paddingSize;

		PGPByte ln_cust_comp_b64_encr[kPGPLANet_SendSize];

		PGPUInt32 i;

		tc_decode_base64( PGPPeekContextMemoryMgr(context), 
			&ticket_bin, &ticket_bin_size, ticket_b64 );

		if( ticket_bin )
			err = dhImportGY( dhExchange, ticket_bin, ticket_bin_size );

		if( ticket_bin )
			PGPFreeData(ticket_bin);

		/* prepare secure ln */
		dhGenerate2( dhExchange );

		PGPNewSymmetricCipherContext( context, kPGPCipherAlgorithm_3DES, &symmRef );
		err = PGPGetSymmetricCipherSizes( symmRef, NULL, &cipherBlockSize );
		pgpAssert( cipherBlockSize <= sizeof(IV) );

		/* get (part of) binary secure cookie and put it in IV */
		tc_decode_base64( PGPPeekContextMemoryMgr(context), 
			&scookie_bin_p, &tsize, scookie_b64 );
		if( scookie_bin_p )  {
			pgpCopyMemory( scookie_bin_p, IV, cipherBlockSize );
			PGPFreeData(scookie_bin_p);
		}

		err = PGPNewCBCContext( symmRef, &cbcContext );
		symmRef = kInvalidPGPSymmetricCipherContextRef;
		

		err = PGPInitCBC( cbcContext, dhBuf(dhExchange), IV );

		paddingSize = pkcs5Padding( ln_cust_comp_b64, cipherBlockSize );

		/* do an encryption, check the result */
		for( i = 0; i < strlen(ln_cust_comp_b64)+1 + paddingSize; i += cipherBlockSize )
		{
			err = PGPCBCEncrypt( cbcContext, ln_cust_comp_b64 + i, cipherBlockSize, ln_cust_comp_b64_encr + i );

			if( IsPGPError( err ) )
				break;
		}

		tc_encode_base64( PGPPeekContextMemoryMgr(context), 
			&ln_cust_comp_b64_encr_b64, &tsize, 
			ln_cust_comp_b64_encr, i );	/* multiple of block size */

	}

	postParams[0] = '\0';
	/// XXX check parameters size -- what if scookie was huge??? XXX
	addFormUrlEncodedParam( "r", "activate", postParams );
	addFormUrlEncodedParam( "protected", "yes", postParams );
	addFormUrlEncodedParam( tok_scookie, scookie_b64, postParams );
	addFormUrlEncodedParam( tok_ticket_stub, ticket_stub_b64, postParams );
	
	addFormUrlEncodedParam( "authorize_request", ln_cust_comp_b64_encr_b64, postParams );

	if( ticket_stub_b64 != NULL )
		PGPFreeData( ticket_stub_b64 );

	if( ln_cust_comp_b64_encr_b64  != NULL )
		PGPFreeData( ln_cust_comp_b64_encr_b64  );

	if( IsntPGPError(err) )
		err = roundTrip( socket, host_port, resource, TRUE, postParams, out, &httpError, &remotePgpError );

	if( IsntPGPError(err) )  {
		if( IsPGPError(remotePgpError) )
			err = remotePgpError;

		if( httpError<200 || httpError>=300)  {
#if PGP_DEBUG_PRINT
			fprintf( stderr, 
				"HTTP communication with %s returned %d HTTP code and %d remote PGPError in round 2\n", 
				url, httpError, remotePgpError );
#endif
			if( IsntPGPError(err) )
				err = kPGPError_NetworkOpFailed;
		}
		else  if( httpError==204 ) {
#if PGP_DEBUG_PRINT
			fprintf( stderr, "HTTP LA server said that there is no information to return in round 2. "
				"This means that the server software refused LA (remote PGPError %d)\n", remotePgpError );
#endif
			if( IsntPGPError(err) )
				err = kPGPError_NetLARefused;
		}
	}

	if( IsntPGPError(err)  )  {
		PGPByte la[kPGPLANet_ReceiveSize];
		PGPByte *la_encr = NULL;
		PGPSize la_encr_size;
		PGPUInt32 i;

		parseParams( out, NULL, NULL, NULL, NULL, NULL, la, NULL );

		if( la[0] != '\0' )  {
			tc_decode_base64( PGPPeekContextMemoryMgr(context), 
				&la_encr, &la_encr_size, la );

			err = PGPInitCBC( cbcContext, dhBuf(dhExchange), IV );

			for( i = 0; i < la_encr_size; i += cipherBlockSize )
			{
				err = PGPCBCDecrypt( cbcContext, la_encr + i, cipherBlockSize, la + i );
				if( IsPGPError( err ) )
					break;
			}

			// TODO: check PKCS5 padding 

			err = PGPLAImportBase64( context, la, buffer );
		}
		else {
			err = kPGPError_NetLARefused;	// We got HTTP OK, but could not parse data
											// this is an unidentified server error
		}
	}

	if( cbcContext != NULL )
		(void) PGPFreeCBCContext( cbcContext );

	if( PGPSymmetricCipherContextRefIsValid( symmRef ) )
		(void)PGPFreeSymmetricCipherContext( symmRef );

	if( dhExchange )
		dhFree( dhExchange );

	if( socket != kInvalidPGPSocketRef )
		PGPCloseSocket(socket);

	return err;
}
